UNIT AcsPro;

INTERFACE

Uses Windows, SysUtils, AclApi2, AccCtrl2, TypInfo, Dialogs,
     BinIO, ItemPro, StopDlg;

Type TAcsKind = (dir_node, file_node, acs_node);

     TAcsItem = class (TItem)
        kind: TAcsKind;
        user: string;
        value: string;

        desc: pointer;
        list: pointer;

        constructor CreateItem (p_kind: TAcsKind;
                                p_name, p_user, p_value: string);

        procedure ShowItem (var txt: TColumns); override;
        function  GetHeader: TBrwHeader; override;

        procedure Dupl (item: TItem); override;
        function  IsEqual (item: TItem): boolean; override;

        procedure Load (f: TBinRead; v: integer); override;
        procedure Save (f: TBinWrite; v: integer); override;
     end;

Type EAccess = class (Exception);

Function ScanPermissions (path: string): TAcsItem;

IMPLEMENTATION

(**************************** SCAN PERMISSIONS ****************************)

Procedure Error (ans: longword);
Begin
     raise EAccess.Create ('AcsPro error ' +
                           IntToStr (ans) + ': ' +
                           SysErrorMessage (ans));
End;

{--------------------------------------------------------------------------}

Function TrusteeToStr (const trustee: PTrustee): string;
Begin
     result := '???';
     if trustee <> nil then
        if trustee.TrusteeForm = TRUSTEE_IS_NAME then
           if trustee.ptstrName <> nil then
              result := trustee.ptstrName;
End;

Procedure ScanAccess (top: TAcsItem; path: string);

Const attr: array [0..4] of string =
      ('SUB_OBJECTS',        {0x01}
       'SUB_CONTAINERS',     {0x02}
       'NO_PROPAGATE',       {0x04}
       'SUB_ONLY',           {0x08}
       'COMPOSED'            {0x10} );

Var  err: longword;

     owner: P_TRUSTEE_A;
     group: P_TRUSTEE_A;
     desc: PSECURITY_DESCRIPTOR;

     count, i: longword;
     list, elem: PEXPLICIT_ACCESS_A;

     bits: longword;
     mask: longword;
     k: integer;

     n_txt, u_txt, v_txt: string;
     item: TAcsItem;

Begin
     desc := nil;
     owner := nil;
     group := nil;
     list := nil;
     try

        (*
        DWORD GetNamedSecurityInfo (
           LPTSTR pObjectName, // name of the object
           SE_OBJECT_TYPE ObjectType, // type of object
           SECURITY_INFORMATION SecurityInfo, // type of security information to retrieve
           PSID *ppsidOwner, // receives a pointer to the owner SID
           PSID *ppsidGroup, // receives a pointer to the primary group SID
           PACL *ppDacl, // receives a pointer to the DACL
           PACL *ppSacl, // receives a pointer to the SACL
           PSECURITY_DESCRIPTOR *ppSecurityDescriptor); // receives a pointer to the security descriptor

        function GetNamedSecurityInfo (
           pObjectName: PAnsiChar;
           ObjectType: SE_OBJECT_TYPE;
           SecurityInfo: SECURITY_INFORMATION; ppsidOwner,
           ppsidGroup: PPSID;
           ppDacl, ppSacl: PPACL; { !! PPACL }
           var ppSecurityDescriptor: PSECURITY_DESCRIPTOR):
           DWORD; stdcall; {use localfree to release ppSecurityDescriptor}
        *)

        err := GetNamedSecurityInfo (

           PChar (path),   // name of the object
           SE_FILE_OBJECT,	// type of object

           0 * OWNER_SECURITY_INFORMATION or
           0 * GROUP_SECURITY_INFORMATION or
           1 * DACL_SECURITY_INFORMATION or
           0 * SACL_SECURITY_INFORMATION,     // type of security information to retrieve

           nil,     // receives a pointer to the owner SID
           nil,     // receives a pointer to the primary group SID
           nil,     // receives a pointer to the DACL
           nil,     // receives a pointer to the SACL
           desc);   // receives a pointer to the security descriptor

        if err <> ERROR_SUCCESS then begin
           top.Add (TAcsItem.CreateItem (acs_node, 'Error', SysErrorMessage (err), ''));
           exit;
           Error (err);
        end;

        {--------------------------------------------------------------------}

        (*
        DWORD LookupSecurityDescriptorParts (
           PTRUSTEE *pOwner, // receives the owner SID from the security descriptor
           PTRUSTEE *pGroup, // receives the group SID from the security descriptor
           PULONG cCountOfAccessEntries, // receives number of access-control entries
           PEXPLICIT_ACCESS *pListOfAccessEntries, // receives an array of DACL access-control entries
           PULONG cCountOfAuditEntries,	// receives number of audit-control entries
           PEXPLICIT_ACCESS *pListOfAuditEntries, // receives an array of SACL audit-control entries
           PSECURITY_DESCRIPTOR pSD); // pointer that receives the new security descriptor

        function LookupSecurityDescriptorParts (
           pOwner, pGroup: P_PTRUSTEE_; { !! TRUSTEE_ -> P_TRUSTEE_ }
           cCountOfAccessEntries: PULONG;
           pListOfAccessEntries: P_PEXPLICIT_ACCESS_; { !! PEXPLICIT_ACCESS_ -> P_PEXPLICIT_ACCESS_ }
           cCountOfAuditEntries: PULONG;
           pListOfAuditEntries: P_PEXPLICIT_ACCESS_; { !! PEXPLICIT_ACCESS_ -> P_PEXPLICIT_ACCESS_ }
           pSD: PSECURITY_DESCRIPTOR {!? var SECURITY_DESCRIPTOR -> PSECURITY_DESCRIPTOR}):
           DWORD; stdcall;
        *)

        err := LookupSecurityDescriptorParts (
           nil, // @ owner,   // receives the owner SID from the security descriptor
           nil, // @ group,   // receives the group SID from the security descriptor
           @ count,   // receives number of access-control entries
           @ list,    // receives an array of DACL access-control entries
           nil,       // receives number of audit-control entries
           nil,       // receives an array of SACL audit-control entries
           desc);     // pointer that receives the new security descriptor

        if err <> ERROR_SUCCESS then begin
           top.Add (TAcsItem.CreateItem (acs_node, 'Error', SysErrorMessage (err), ''));
           exit;
           Error (err);
        end;

        {--------------------------------------------------------------------}

        top.desc := desc;
        top.list := list;

        top.Add (TAcsItem.CreateItem (acs_node, 'owner', TrusteeToStr (owner), ''));
        top.Add (TAcsItem.CreateItem (acs_node, 'group', TrusteeToStr (group), ''));

        elem := list;
        for i := 0 to count-1 do begin
           if elem.grfAccessMode <= High (ACCESS_MODE) then
              n_txt := GetEnumName (typeinfo (ACCESS_MODE), ord (elem.grfAccessMode))
           else
              n_txt := '???';
           u_txt := TrusteeToStr (@elem.Trustee);
           v_txt := IntToHex (elem.grfAccessPermissions, 8);


           bits := elem.grfInheritance;
           if bits = 11 then
              v_txt := v_txt + ' SUB'
           else
              for k := low (attr) to high (attr) do begin
                 mask := 1 shl k;
                 if (bits and mask) <> 0 then
                    v_txt := v_txt + ' ' + attr [k];
              end;

           item := TAcsItem.CreateItem (acs_node, n_txt, u_txt, v_txt);
           top.Add (item);

           inc (elem); // next list element
        end;

     finally
        {
        if owner <> nil then
           Win32Check (LocalFree (longword (owner)) = 0);
        }

        {
        if group <> nil then
           Win32Check (LocalFree (longword (group)) = 0);
        }

        {
        if list <> nil then
           Win32Check (LocalFree (longword (list)) = 0);

        if desc <> nil then
           Win32Check (LocalFree (longword (desc)) = 0);
        }
     end;
End;

Procedure ScanBranch (top: TAcsItem; path: string);
Var  rec:  TSearchRec;
     item: TAcsItem;
     len:  integer;
     ans:  longword;
     kind: TAcsKind;
Begin
     StopInfo (path);
     if StopFlag then exit;

     len := length (path);
     if (len>0) and (path[len]='\') then path := copy (path, 1, len-1);

     ans := FindFirst (path + '\*.*', faAnyFile, rec);
     try
        while ans = 0 do begin
           if (rec.name <> '.') and (rec.name <> '..') then begin
              if (rec.attr and faDirectory) <> 0
                 then kind := dir_node
                 else kind := file_node;

              item := TAcsItem.CreateItem (kind, rec.name, '', '');
              top.Add (item);
           end;
           ans := FindNext (rec);
        end;
     finally
        FindClose (rec);
     end;

     if ans <> ERROR_NO_MORE_FILES then Error (ans);

     { scan permissions }
     item := top.sub.first as TAcsItem;
     while (item <> nil) and not StopFlag do begin
        if item.kind <> acs_node then
           ScanAccess (item, path + '\' + item.name);
        item := item.next as TAcsItem;
     end;

     { scan subdirectories }
     item := top.sub.first as TAcsItem;
     while (item <> nil) and not StopFlag do begin
        if item.kind = dir_node
           then ScanBranch (item, path + '\' + item.name);
        item := item.next as TAcsItem;
     end;
End;

Function ScanPermissions (path: string): TAcsItem;
Var  ans:  longword;
     rec:  TSearchRec;
     kind: TAcsKind;
Begin
     result := nil;
     StopOpen ('Scanning ' + path);

     try
        try

           ans := FindFirst (path, faAnyFile, rec);
           try
              if (ans <> 0) and (ans <> ERROR_FILE_NOT_FOUND) then
                 Error (ans);


              if ans = 0 then begin
                 if (rec.attr and faDirectory) <> 0
                    then kind := dir_node
                    else kind := file_node
              end
              else kind := dir_node;

              result := TAcsItem.CreateItem (kind, path, '', '');
           finally
              FindClose (rec);
           end;

           if result <> nil then ScanBranch (result, path);

        finally
           StopClose;
        end;

     except
        result.Free;
        raise;
     end;

     if StopFlag then begin
        result.Free;
        result := nil;
     end;
End;

(******************************** ACS ITEM ********************************)

Constructor TAcsItem.CreateItem (p_kind:  TAcsKind;
                                 p_name:  string;
                                 p_user: string;
                                 p_value: string);
Begin
     inherited Create;

     kind  := p_kind;
     name  := p_name;
     user  := p_user;
     value := p_value;

     folder := (kind <> acs_node);
     if kind = file_node then pict := page_pict;
End;

(******************************* SHOW ITEMS *******************************)

Procedure TAcsItem.ShowItem (var txt: TColumns);
Begin
     txt[1] := name;
     txt[2] := user;
     txt[3] := value;
End;

(********************************* HEADER *********************************)

Var AcsHeader: TBrwHeader = nil;

Procedure CreateAcsHeader;
Begin
     AcsHeader := TBrwHeader.Create;
     with AcsHeader do begin
        Add ('Name', 100, ByName);
        Add ('User', 200, ByName);
        Add ('Value', 200, ByValue);
     end;
End;

Function TAcsItem.GetHeader: TBrwHeader;
Begin
     if AcsHeader = nil then CreateAcsHeader;
     result := AcsHeader;
End;

(******************************** COMPARE *********************************)

Procedure TAcsItem.Dupl (item: TItem);
Var  p: TAcsItem;
Begin
     inherited;
     p := item as TAcsItem;
     user := p.user;
     value := p.value;
End;

Function TAcsItem.IsEqual (item: TItem): boolean;
Var  p: TAcsItem;
Begin
     p := item as TAcsItem;
     result := (user = p.user) and (value = p.value);
End;

(****************************** LOAD / SAVE *******************************)

Procedure TAcsItem.Load (f: TBinRead; v: integer);
Begin
     inherited;
End;

Procedure TAcsItem.Save (f: TBinWrite; v: integer);
Begin
     inherited;
End;

END.
