UNIT CompDlg;

INTERFACE

Uses Windows, Messages, SysUtils, Classes, Graphics,
     Controls, Forms, Dialogs, StdCtrls, ExtCtrls,
     Menus, ComCtrls;

{--------------------------------------------------------------------------}

Type TTimeOpt    = (exact_time, no_ms_time, no_sec_time, one_hour_time);
     THistoryOpt = (full_hist, min_hist, no_hist);

     TCompareOpt = record
        contents: boolean;

        size: boolean;
        time: boolean;
        attr: boolean;
        crc:  boolean;

        resolution: TTimeOpt;

        incl_new:   boolean;
        incl_old:   boolean;
        incl_modif: boolean;
        incl_equal: boolean;

        hist: THistoryOpt;
     end;

Function CompareHistoryConfig: TCompareOpt;
Function CompareContentsConfig: TCompareOpt;
Function DifferenceConfig: TCompareOpt;

{--------------------------------------------------------------------------}

Type
  TCompareDialog = class (TForm)
    CompareGroupBox: TGroupBox;
    ContentsCheckBox: TCheckBox;
    SizeCheckBox: TCheckBox;
    TimeCheckBox: TCheckBox;
    TimeComboBox: TComboBox;
    AttrCheckBox: TCheckBox;
    CrcCheckBox: TCheckBox;

    IncludeGroupBox: TGroupBox;
    NewCheckBox: TCheckBox;
    OldCheckBox: TCheckBox;
    ModifCheckBox: TCheckBox;
    EqualCheckBox: TCheckBox;

    HistoryRadioGroup: TRadioGroup;

    CancelButton: TButton;
    OkButton: TButton;
    { checkpoint }

    procedure TimeCheckBoxClick (Sender: TObject);
    { checkpoint }

  public
    function Execute (var opt: TCompareOpt): boolean;

  private
    procedure UpdateFields;
    procedure SetFields (const opt: TCompareOpt);
    procedure GetFields (out opt: TCompareOpt);
  end;

Var CompareDialog: TCompareDialog;

IMPLEMENTATION

{$R *.DFM}

(************************* DEFAULT CONFIGURATIONS *************************)

Function CompareHistoryConfig: TCompareOpt;
Begin
     with result do begin
        contents := false;

        size := true;
        time := true;
        attr := false;
        crc  := false;

        resolution := no_ms_time;
        hist := full_hist;

        incl_new   := true;
        incl_old   := true;
        incl_modif := true;
        incl_equal := true;
     end;
End;

Function CompareContentsConfig: TCompareOpt;
Begin
     with result do begin
        contents := true;

        size := true;
        time := false;
        attr := false;
        crc  := false;

        resolution := no_ms_time;
        hist := no_hist;

        incl_new   := true;
        incl_old   := true;
        incl_modif := true;
        incl_equal := true;
     end;
End;

Function DifferenceConfig: TCompareOpt;
Begin
     with result do begin
        contents := false;
        
        size := true;
        time := true;
        attr := false;
        crc  := false;

        resolution := no_ms_time;
        hist := no_hist;

        incl_new := true;
        incl_old := true;
        incl_modif := true;
        incl_equal := false;
     end;
End;

(**************************** SET / GET FIELDS ****************************)

Procedure TCompareDialog.UpdateFields;
Var  b: boolean;
Begin
     b := TimeCheckBox.Checked;
     TimeComboBox.Enabled := b;
End;

Procedure TCompareDialog.SetFields (const opt: TCompareOpt);
Begin
     ContentsCheckBox.Checked := opt.contents;
     SizeCheckBox.Checked := opt.size;
     TimeCheckBox.Checked := opt.time;
     AttrCheckBox.Checked := opt.attr;
     CrcCheckBox.Checked  := opt.crc;

     TimeComboBox.ItemIndex := ord (opt.resolution);

     NewCheckBox.Checked   := opt.incl_new;
     OldCheckBox.Checked   := opt.incl_old;
     ModifCheckBox.Checked := opt.incl_modif;
     EqualCheckBox.Checked := opt.incl_equal;

     HistoryRadioGroup.ItemIndex := ord (opt.hist);

     UpdateFields;
End;

Procedure TCompareDialog.GetFields (out opt: TCompareOpt);
Begin
     opt.contents := ContentsCheckBox.Checked ;

     opt.size := SizeCheckBox.Checked;
     opt.time := TimeCheckBox.Checked;
     opt.attr := AttrCheckBox.Checked;
     opt.crc  := CrcCheckBox.Checked;

     opt.resolution := TTimeOpt (TimeComboBox.ItemIndex);

     opt.incl_new   := NewCheckBox.Checked;
     opt.incl_old   := OldCheckBox.Checked;
     opt.incl_modif := ModifCheckBox.Checked;
     opt.incl_equal := EqualCheckBox.Checked;

     opt.hist := THistoryOpt (HistoryRadioGroup.ItemIndex);
End;

(***************************** NOTIFICATIONS ******************************)

Procedure TCompareDialog.TimeCheckBoxClick (Sender: TObject);
Begin
     UpdateFields;
End;

(******************************** EXECUTE *********************************)

Function TCompareDialog.Execute (var opt: TCompareOpt): boolean;
Var  form: TForm;
Begin
     form := Application.MainForm;
     Top  := form.Top + (form.Height - Height) div 2;
     Left := form.Left + (form.Width - Width) div 2;

     SetFields (opt);
     result := (ShowModal = mrOK);
     if result then GetFields (opt);
End;


END.

