UNIT CompPro;

INTERFACE

Uses Windows, SysUtils, Classes,
     ItemPro, FilePro, SortPro, StopDlg, CompDlg;

Function Compare (src, dst: TItem; const cfg: TCompareOpt): TItem;
{ Porovnani stromu, vytvari se novy strom }

Procedure Evaluate (src, dst: TItem; const cfg: TCompareOpt);
{ Nalezni prvky zdrojoveho stromu v cilovem stromu }

Procedure Lookup (src, dst: TItem; cfg: TCompareOpt);
{ Nalezni prvky zdrojoveho stromu v cilovem stromu bez ohledu na adresare }

Function FindDuplications (src: TItem; cfg: TCompareOpt): TItem;
{ Nalezni soubory se stejnym jmenem, vytvor seznam }

Function Duplicate (item: TItem): TItem;

IMPLEMENTATION

(******************************** OPTIONS *********************************)

Var opt: TCompareOpt; { global options }

(********************************* ERROR **********************************)

Procedure Error (s: string);
Begin
     raise Exception.Create (s);
End;

Procedure CheckFolder (item: TItem);
Begin
     if not item.folder then Error ('Folder expected');
End;

(**************************** DUPLICATE ITEMS *****************************)

Function Duplicate (item: TItem): TItem;
Var  p: TItem;
Begin
     p := TItemClass (Item.ClassType). Create;
     p.Dupl (item);
     result := p;
End;

(***************************** COMPARE ITEMS ******************************)

Function CompareContents (src, dst: TFileItem): boolean;
Const BufMax = 16*1024;
Var   f, g: file;
      save: byte;
      f_len, g_len, blk, i: integer;
      f_buf, g_buf: array [1..BufMax] of byte;
Begin
     StopInfo2 (src.FullName, dst.FullName);
     result := true;
     try
        save := FileMode;
        try
           FileMode := 0;
           Assign (f, src.FullName);
           Reset (f, 1);
           try
              Assign (g, dst.FullName);
              Reset (g, 1);
              try
                 blk := 0;
                 repeat
                    inc (blk);
                    if blk mod 10 = 0 then StopTest;

                    BlockRead (f, f_buf, sizeof (f_buf), f_len);
                    BlockRead (g, g_buf, sizeof (g_buf), g_len);

                    if (f_len <> g_len) then
                       result := false
                    else begin
                       i := 1;
                       while (i <= f_len) and (f_buf[i] = g_buf[i]) do
                          inc (i);
                       if i <= f_len then result := false;
                    end;
                 until not result or eof (f) or eof (g) or StopFlag;

                 if not eof (f) or not eof (g) then result := false;
              finally
                 Close (g);
              end;
           finally
              Close (f);
           end;
        finally
           FileMode := save;
        end;
     except
        result := false;
     end;
End;

{--------------------------------------------------------------------------}

Function CompInt (a, b: integer): integer;
Begin
     if a<b then result := -1
     else if a=b then result := 0
     else result := 1;
End;

Function CompProperties (src, dst: TFileItem): integer;
Const sec_mask = $FFFFFFE0;
      one_hour = $0000F800;
      AttrMask = faReadOnly + faArchive + faHidden + faSysFile;
Begin
     result := 0;

     { size }
     if (result=0) and opt.size then
        result := src.size - dst.size;

     { time }
     if (result=0) and opt.time then
        case opt.resolution of
           no_sec_time:
              result := (src.time and sec_mask) - (dst.time and sec_mask);

           one_hour_time:
              begin
                 result := src.time - dst.time;
                 if (result = one_hour) or (result = -one_hour) then
                    result := 0;
              end;

           else
              result := src.time - dst.time;
        end;

     { attr }
     if (result=0) and opt.attr then
        result := (src.attr and AttrMask) - (dst.attr and AttrMask);

     { crc }
     if (result=0) and opt.crc then begin
        result := ord (src.crc_ready) - ord (dst.crc_ready);
        if result = 0 then result := CompInt (src.crc, dst.crc);
     end;
End;

{--------------------------------------------------------------------------}

Function CompItems (s_ptr, d_ptr: TItem): integer;
Var  s_fil, d_fil: TFileItem;
Begin
     result := CompByName (s_ptr, d_ptr);

     if (result = 0) and (s_ptr is TFileItem) and (d_ptr is TFileItem) then begin
        s_fil := TFileItem (s_ptr);
        d_fil := TFileItem (d_ptr);
        result := CompProperties (s_fil, d_fil);
     end;
End;

{--------------------------------------------------------------------------}

Function EqualItems (s_ptr, d_ptr: TItem): boolean;
Var  s_fil, d_fil: TFileItem;
Begin
     if (s_ptr is TFileItem) and (d_ptr is TFileItem) then begin
        s_fil := TFileItem (s_ptr);
        d_fil := TFileItem (d_ptr);

        result := (CompProperties (s_fil, d_fil) = 0);

        { contents }
        if result and opt.contents then
           result := CompareContents (s_fil, d_fil);
     end
     else begin
        result := s_ptr.IsEqual (d_ptr);
     end;
End;

(********************************** CMP ***********************************)

Type TCmp = class
        { Obecne porovnani }

        incl: TModeSet; { zobrazovane druhy souboru }
        hist: TModeSet; { druhy souboru, pro ktere se vytvari history }

        procedure OldLeaf  (s_ptr: TItem); virtual; abstract;
        procedure NewLeaf  (d_ptr: TItem); virtual; abstract;
        procedure TwoLeaves (s_ptr, d_ptr: TItem); virtual; abstract;

        procedure OldFolder  (s_ptr: TItem); virtual; abstract;
        procedure NewFolder  (d_ptr: TItem); virtual; abstract;
        procedure TwoFolders (s_ptr, d_ptr: TItem); virtual; abstract;

        procedure OldItem  (s_ptr: TItem);
        procedure NewItem  (d_ptr: TItem);
        procedure TwoItems (s_ptr, d_ptr: TItem);

        procedure CompareTables (src, dst: TItem);
        procedure CompareLists (src, dst: TItem);
        procedure CompareFolders (src, dst: TItem);

        procedure SetOptions (const cfg: TCompareOpt);
     end;

{--------------------------------------------------------------------------}

Const cmp : CompFce = CompByName;

Procedure TCmp.CompareTables (src, dst: TItem);
Var  s_tab, d_tab: TableArr;
     s_cnt, d_cnt: integer;
     s_inx, d_inx: integer;
     s_ptr, d_ptr: TItem;
     rel: integer;
Begin
     CheckFolder (src);
     CheckFolder (dst);

     AllocTable (src, AcceptAll.Filter, s_tab);
     AllocTable (dst, AcceptAll.Filter, d_tab);

     SortTable (cmp, s_tab);
     SortTable (cmp, d_tab);

     s_inx := 0;
     d_inx := 0;

     s_cnt := length (s_tab);
     d_cnt := length (d_tab);

     while (s_inx < s_cnt) or (d_inx < d_cnt) do begin
        if StopFlag then exit;

        if d_inx >= d_cnt then begin
           OldItem (s_tab [s_inx]);
           inc (s_inx);
        end
        else if s_inx >= s_cnt then begin
           NewItem (d_tab [d_inx]);
           inc (d_inx);
        end
        else begin
           s_ptr := s_tab [s_inx];
           d_ptr := d_tab [d_inx];
           rel := cmp (s_ptr, d_ptr);

           if rel < 0 then begin
              OldItem (s_ptr);
              inc (s_inx);
           end
           else if rel > 0 then begin
              NewItem (d_ptr);
              inc (d_inx);
           end
           else begin
              TwoItems (s_ptr, d_ptr);
              inc (s_inx);
              inc (d_inx);
           end;
        end;
     end; {while}

     FreeTable (s_tab);
     FreeTable (d_tab);
End;

Procedure TCmp.CompareLists (src, dst: TItem);
Var  s_ptr, d_ptr: TItem;
     rel: integer;
Begin
     { compare sorted lists }

     CheckFolder (src);
     CheckFolder (dst);

     s_ptr := src.sub.first;
     d_ptr := dst.sub.first;

     while (s_ptr <> nil) or (d_ptr <> nil) do begin
        if StopFlag then exit;

        if d_ptr = nil then begin
           OldItem (s_ptr);
           s_ptr := s_ptr.next;
        end
        else if s_ptr = nil then begin
           NewItem (d_ptr);
           d_ptr := d_ptr.next;
        end
        else begin
           rel := cmp (s_ptr, d_ptr);

           if rel < 0 then begin
              OldItem (s_ptr);
              s_ptr := s_ptr.next;
           end
           else if rel > 0 then begin
              NewItem (d_ptr);
              d_ptr := d_ptr.next;
           end
           else begin
              TwoItems (s_ptr, d_ptr);
              s_ptr := s_ptr.next;
              d_ptr := d_ptr.next;
           end;
        end;
     end; {while}
End;

Procedure TCmp.CompareFolders (src, dst: TItem);
Begin
     StopInfo2 (src.SectionName, dst.SectionName);
     if StopFlag then exit;
     CompareLists (src, dst);
End;

{--------------------------------------------------------------------------}

Procedure TCmp.OldItem (s_ptr: TItem);
Begin
     if s_ptr.folder then OldFolder (s_ptr)
                     else OldLeaf (s_ptr);
End;

Procedure TCmp.NewItem (d_ptr: TItem);
Begin
     if d_ptr.folder then NewFolder (d_ptr)
                     else NewLeaf (d_ptr);
End;

Procedure TCmp.TwoItems (s_ptr, d_ptr: TItem);
Begin
     if s_ptr.folder then begin
        Assert (d_ptr.folder);
        TwoFolders (s_ptr, d_ptr);
     end
     else begin
        Assert (not d_ptr.folder);
        TwoLeaves (s_ptr, d_ptr);
     end;
End;

(**************************** SET FOLDER MODE *****************************)

Procedure SetFolderMode (item: TItem);
Var  p: TItem;
     ans: TMode;
Begin
     ans := equal;
     p := item.sub.first;

     while p <> nil do begin
        if p.mode <> equal then ans := modified;
        p := p.next;
     end;

     item.mode := ans;
End;

(********************************** COMP **********************************)

Type TComp = class (TCmp)
        { Porovnani dvou stromu, vytvari se novy strom }

        res:  TItem;       { vysledny strom }

        procedure PutItem  (item: TItem; val: TMode);
        procedure PutTwoLeaves (s_ptr, d_ptr: TItem; val: TMode);

        procedure OldLeaf  (s_ptr: TItem); override;
        procedure NewLeaf  (d_ptr: TItem); override;
        procedure TwoLeaves (s_ptr, d_ptr: TItem); override;

        procedure OldFolder  (s_ptr: TItem); override;
        procedure NewFolder  (d_ptr: TItem); override;
        procedure TwoFolders (s_ptr, d_ptr: TItem); override;
     end;

{--------------------------------------------------------------------------}

Procedure TComp.PutItem (item: TItem; val: TMode);
Var  tmp, t, p: TItem;
Begin
     { kopie polozky }
     tmp := Duplicate (item);
     tmp.mode := val;
     res.Add (tmp);

     { history }
     if val in hist then begin
        t := Duplicate (item);
        t.mode := val;
        tmp.AddHist (t);
     end;

     { vnorene polozky }
     if item.folder then begin
        res := tmp;

        p := item.sub.first;
        while p <> nil do begin
           PutItem (p, val);
           p := p.next;
        end;

        res := res.up;
     end;
End;

Procedure TComp.PutTwoLeaves (s_ptr, d_ptr: TItem; val: TMode);
Var  tmp, t: TItem;
Begin
     { kopie polozky }
     tmp := Duplicate (d_ptr);
     tmp.mode := val;
     res.Add (tmp);

     { history }
     if val in hist then begin
        t := Duplicate (s_ptr);
        t.mode := val;
        tmp.AddHist (t);

        t := Duplicate (d_ptr);
        t.mode := val;
        tmp.AddHist (t);
     end;
End;

{--------------------------------------------------------------------------}

Procedure TComp.OldLeaf (s_ptr: TItem);
Begin
     if old in incl then
        PutItem (s_ptr, old);
End;

Procedure TComp.OldFolder (s_ptr: TItem);
Begin
     if old in incl then
        PutItem (s_ptr, old);
End;

Procedure TComp.NewLeaf (d_ptr: TItem);
Begin
     if new in incl then
        PutItem (d_ptr, new);
End;

Procedure TComp.NewFolder (d_ptr: TItem);
Begin
     if new in incl then
        PutItem (d_ptr, new);
End;

Procedure TComp.TwoLeaves (s_ptr, d_ptr: TItem);
Var  val: TMode;
Begin
     if EqualItems (s_ptr, d_ptr) then val := equal
                                  else val := modified;
     if val in incl then
        PutTwoLeaves (s_ptr, d_ptr, val);
End;

Procedure TComp.TwoFolders (s_ptr, d_ptr: TItem);
Var  save, tmp: TItem;
Begin
     tmp := Duplicate (d_ptr);

     save := res;
     res := tmp;

     CompareFolders (s_ptr, d_ptr);
     SetFolderMode (tmp);

     res := save;

     if res = nil then res := tmp
     else if tmp.mode in incl then res.Add (tmp)
     else tmp.Free;
End;

(******************************** COMPARE *********************************)

Procedure TCmp.SetOptions (const cfg: TCompareOpt);
Begin
     opt := cfg;

     incl := [];
     if opt.incl_new   then Include (incl, new);
     if opt.incl_old   then Include (incl, old);
     if opt.incl_modif then Include (incl, modif);
     if opt.incl_equal then Include (incl, equal);

     case opt.hist of
        full_hist: hist := [old, new, modified, equal];
        min_hist:  hist := [modified];
        no_hist:   hist := [];
     end;

     hist := hist * incl;
End;

Function Compare (src, dst: TItem; const cfg: TCompareOpt): TItem;
{ Kompletni porovnani stromu }
Var  tmp: TComp;
Begin
     tmp := TComp.Create;
     tmp.SetOptions (cfg);
     tmp.TwoFolders (src, dst);
     result := tmp.res;
     tmp.Free;

     if StopFlag then begin
        result.Free;
        result := nil;
     end;
End;

(******************************** EVALUATE ********************************)

Type TEval = class (TCmp)
        { Porovnani dvou stromu, nevytvari se novy strom }
        { Nalezni prvky zdrojoveho stromu v cilovem stromu }

        procedure OldLeaf   (s_ptr: TItem); override;
        procedure NewLeaf   (d_ptr: TItem); override;
        procedure TwoLeaves (s_ptr, d_ptr: TItem); override;

        procedure OldFolder  (s_ptr: TItem); override;
        procedure NewFolder  (d_ptr: TItem); override;
        procedure TwoFolders (s_ptr, d_ptr: TItem); override;
     end;

Procedure TEval.OldLeaf (s_ptr: TItem);
Begin
     { nothing }
End;

Procedure TEval.NewLeaf (d_ptr: TItem);
Begin
     { pokud prvek neni oznaceny -> oznac jako novy }
     if d_ptr.mode = normal then
        d_ptr.mode := new;
End;

Procedure TEval.TwoLeaves (s_ptr, d_ptr: TItem);
Begin
     { pokud jiz nenastala rovnost, porovnej prvky }
     if d_ptr.mode <> equal then
        if EqualItems (s_ptr, d_ptr) then d_ptr.mode := equal
                                     else d_ptr.mode := modified;
End;

Procedure TEval.OldFolder (s_ptr: TItem);
Begin
     { nothing }
End;

Procedure TEval.NewFolder (d_ptr: TItem);
Var  p: TItem;
Begin
     { pokud vetev neni oznacena -> oznac jako novou }
     if d_ptr.mode = normal then begin
        d_ptr.mode := new;

        p := d_ptr.sub.first;
        while p <> nil do begin
           NewItem (p);
           p := p.next;
        end;
     end;
End;

Procedure TEval.TwoFolders (s_ptr, d_ptr: TItem);
Begin
     CompareFolders (s_ptr, d_ptr);
     SetFolderMode (d_ptr);
End;

Procedure Evaluate (src, dst: TItem; const cfg: TCompareOpt);
{ Nalezni prvky zdrojoveho stromu v cilovem stromu }
Var  tmp: TEval;
Begin
     tmp := TEval.Create;
     tmp.SetOptions (cfg);
     tmp.TwoFolders (src, dst);
     tmp.Free;
End;

(********************************** MIX ***********************************)

Procedure MixBranch (src, tmp: TItem);
{ Shrnuti souboru ze vsech podadresaru do jednoho adresare }
Var  p: TItem;
Begin
     p := src.sub.first;
     while p <> nil do begin
        if p.folder then MixBranch (p, tmp)
                    else tmp.Add (Duplicate (p));
        p := p.next;
     end;
End;

Function MixFolder (src: TItem; comp: CompFce): TItem;
{ Vytvor setrideny seznam souboru ze vsech podadresaru }
Var  tmp: TItem;
Begin
     tmp := Duplicate (src);
     MixBranch (src, tmp);
     SortFolder (tmp, comp, false);
     result := tmp;
End;

(********************************* LOOKUP *********************************)

Procedure LookupBranch (src: TItem; dst: TItem);
Var  s_ptr, d_ptr, p: TItem;
     rel: integer;
Begin
     s_ptr := src.sub.first;
     d_ptr := dst.sub.first;

     while d_ptr <> nil do begin

        if not d_ptr.folder and (d_ptr.mode <> equal) then begin
           { porovnej se vzorovym prvkem }
           rel := 0; 
           if s_ptr <> nil then
              rel := CompItems (s_ptr, d_ptr);

           { preskoc mensi vzorove prvky }
           while (s_ptr <> nil) and (rel < 0) do begin
              s_ptr := s_ptr.next;
              if s_ptr <> nil then
                 rel := CompItems (s_ptr, d_ptr);
           end;

           if (s_ptr <> nil) and (rel = 0)
              then d_ptr.mode := {equal} s_ptr.mode
              else d_ptr.mode := new ;
        end;

        d_ptr := d_ptr.next;
     end;

     { podadresare }
     p := dst.sub.first;
     while p <> nil do begin
        if p.folder then LookupBranch (src, p);
        p := p.next;
     end;

     SetFolderMode (dst);
End;

Procedure Lookup (src, dst: TItem; cfg: TCompareOpt);
{ Nalezni prvky zdrojoveho stromu v cilovem stromu bez ohledu na adresare }
Var  tmp, p, n: TItem;
Begin
     opt := cfg;
     tmp := MixFolder (src, CompItems); { seznam prvku z leveho stromu }

     { oznac opakujici se prvky }
     p := tmp.sub.first;
     while p <> nil do begin

        p.mode := equal;
        n := p.next;

        while (n <> nil) and (CompItems (p,n) = 0) do begin
           p.mode := old; { opakujici se prvky - cervena znacka }
           n.mode := old;
           n := n.next;
        end;

        p := n;
     end;

     LookupBranch (tmp, dst);
     tmp.Free;
End;

(*************************** FIND DUPLICATIONS ****************************)

Function FindDuplications (src: TItem; cfg: TCompareOpt): TItem;
{ Nalezni soubory se stejnym jmenem }
Var   tmp, p, n: TItem;
      ans, t, u: TItem;
Begin
     opt := cfg;

     tmp := MixFolder (src, CompItems); { seznam prvku z leveho stromu }

     ans := Duplicate (src); { vysledny seznam }
     ans.mode := modif;

     { vyjledej opakujici se prvky }
     p := tmp.sub.first;
     while p <> nil do begin
        n := p.next;
        while (n <> nil) and (CompItems (p,n) = 0) do begin

           {
           if ctrl.contents and (p is TFileItem) and (n is TFileItem)then
              if EqualContents (TFileItem (p), TFileItem (n))
                 then n.mode := equal
                 else n.mode := modified;
           }

           n := n.next;
        end;

        if n <> p.next then begin
           { vytvor prvek ve vyslednem seznamu }
           t := Duplicate (p);
           t.mode := modif;
           ans.Add (t);

           { pridej opakujici se prvky do 'history' }
           u := p;
           while u <> n do begin
              t.AddHist (Duplicate (u));
              u := u.next;
           end;
        end;

        p := n;
     end;

     tmp.Free; { vymaz seznam prku z leveho stromu }
     result := ans;
End;

END.
