UNIT DetPro;

INTERFACE

Uses Classes,  {TNotifyEvent}
     Menus,    {TMenuItem}
     Controls, {TAlign}
     ExtCtrls, {TSplitter}
     ComCtrls, {TListView}
     ItemPro,  {TItem}
     FilePro,  {TFileItem}
     VerPro,   {TVerAttr}
     ViewPro;  {TView}

Type TDetailStyle = (info_style,
                     history_style,
                     permissions_style,
                     version_style,
                     none_style);

Type TDetail = class
     public
        View:           TView;
        DetailSplitter: TSplitter;
        DetailView:     TListView;
        Images:         TImageList;

        MenuViewInfo: TMenuItem;
        MenuViewHistory: TMenuItem;
        MenuViewPermissions: TMenuItem;
        MenuViewVersion: TMenuItem;
        MenuViewNoDetails: TMenuItem;

        procedure Setup;
        procedure SetStyle (value: TDetailStyle);
        procedure UpdateViewMenu;

     private
        style:   TDetailStyle;
        current: TItem; { aktualni prvek zobrazovany v detailnim seznamu }
        header:  TBrwHeader;

        LastListViewWidth: integer;

        procedure ShowImages (b: boolean);
        procedure ShowHeader (hdr: TBrwHeader);
        procedure BeginUpdate;
        procedure EndUpdate;
        procedure ShowList (start: TItem);

        procedure Add (p_name, p_value: string);

        procedure ShowContents (item: TItem);
        procedure ShowHistory (item: TItem);
        procedure ShowPermissions (item: TItem);
        procedure ShowVersion (item: TItem);

        procedure ShowItem (item: TItem);

        procedure SelectItem (item: TItem);

        procedure ViewInfo        (Sender: TObject);
        procedure ViewHistory     (Sender: TObject);
        procedure ViewPermissions (Sender: TObject);
        procedure ViewVersion     (Sender: TObject);
        procedure MenuNoDetails        (Sender: TObject);
     end;

IMPLEMENTATION

(********************************** SHOW **********************************)

Procedure TDetail.ShowImages (b: boolean);
Begin
     if b then
        DetailView.SmallImages := Images
     else begin
        DetailView.SmallImages := nil;
        DetailView.ShowColumnHeaders := false; { posunout text na okraj okenka }
        DetailView.ShowColumnHeaders := true;
     end;
End;

Procedure TDetail.ShowHeader (hdr: TBrwHeader);
Var  i: integer;
     cc: TListColumn;
Begin
     if hdr <> header then begin
        header := hdr;

        DetailView.Columns.BeginUpdate;
        DetailView.Columns.Clear;

        for i := 1 to header.cnt do
           with header.tab[i] do begin
              cc := DetailView.Columns.Add;
              cc.Caption := c_title;
              cc.Width := c_width;
              cc.AutoSize := c_autosize;
           end;

        DetailView.Columns.EndUpdate;
     end;
End;

Procedure TDetail.BeginUpdate;
Begin
     // DetailView.Items.BeginUpdate;
     DetailView.Items.Clear;
End;

Procedure TDetail.EndUpdate;
Begin
     // DetailView.Items.EndUpdate;
End;

Procedure TDetail.ShowList (start: TItem);
Var  item: TItem;
     node: TListItem;
     i: integer;
     txt: TColumns;
Begin
     item := start;
     while item <> nil do begin

        node := DetailView.Items.Add;
        node.Caption := item.name;
        node.Data := item;
        node.ImageIndex := item.GetImageIndex (false);
        node.StateIndex := item.GetStateIndex;

        for i := 1 to Header.cnt do txt [i] := '';
        item.ShowItem (txt);
        for i := 2 to Header.cnt do node.SubItems.Add (txt [i]);

        item := item.next;
     end;
End;

(****************************** SIMPLE LIST *******************************)

Var SimpleHeader: TBrwHeader = nil;
    EmptyHeader: TBrwHeader = nil;

Procedure CreateSimpleHeader;
Begin
     SimpleHeader := TBrwHeader.Create;
     SimpleHeader.Add ('Name', 100, Unsorted);
     SimpleHeader.Add ('Value', 100, Unsorted);
End;

Procedure CreateEmptyHeader;
Begin
     EmptyHeader := TBrwHeader.Create;
End;

Procedure TDetail.Add (p_name, p_value: string);
Var  node: TListItem;
Begin
     node := DetailView.Items.Add;
     node.Caption := p_name;
     node.SubItems.Add (p_value);
     // node.ImageIndex := -1;
End;

(******************************* SHOW ITEM ********************************)

Procedure TDetail.ShowContents (item: TItem);
Begin
     item.GetInfo (Add);
End;

Procedure TDetail.ShowHistory (item: TItem);
Begin
     ShowList (item.history.first);
End;

Procedure TDetail.ShowPermissions (item: TItem);
Begin
     ShowList (item.access.first);
End;

Procedure TDetail.ShowVersion (item: TItem);
Var  file_item: TFileItem;
     atr: TVerAttr;
     info: TVerInfo;
Begin
     if item is TFileItem then begin
        file_item := item as TFileItem;
        file_item.GetVer;
        info := file_item.ver;

        if info <> nil then
           for atr := low (TVerAttr) to high (TVerAttr) do
              if info.value [atr] <> '' then
                 Add (AttrName [atr], info.value [atr]);
     end;
End;

{--------------------------------------------------------------------------}

Procedure TDetail.ShowItem (item: TItem);
Begin
     current := item;
     if style <> none_style then begin

        if (style <> history_style)
           then ShowHeader (SimpleHeader)
           else if item = nil then ShowHeader (EmptyHeader)
                              else ShowHeader (item.GetHeader);

        BeginUpdate;

        if item <> nil then
           case Style of
              info_style:        ShowContents (item);
              history_style:     ShowHistory (item);
              permissions_style: ShowPermissions (item);
              version_style:     ShowVersion (item);
           end;

        EndUpdate;
     end;
End;

(******************************* LIST VIEW ********************************)

Procedure TDetail.SelectItem (item: TItem);
Begin
     if item <> current then
        ShowItem (item);
End;

(********************************** MENU **********************************)

Procedure TDetail.UpdateViewMenu;
Var  menu: TMenuItem;
Begin
     case Style of
        info_style:        menu := MenuViewInfo;
        history_style:     menu := MenuViewHistory;
        permissions_style: menu := MenuViewPermissions;
        version_style:     menu := MenuViewVersion;
        none_style:        menu := MenuViewNoDetails;
        else               menu := nil; {compiler}
     end;

     menu.Checked := true;
End;

Procedure TDetail.SetStyle (value: TDetailStyle);
Begin
     if (value <> none_style) and (style = none_style) then begin
        { show detail view }
        view.ListView.Align := alLeft;
        view.ListView.Width := LastListViewWidth;
        DetailView.Visible := true;
        DetailSplitter.Visible := true;

        if DetailSplitter.Left > DetailView.Left then
           DetailSplitter.Left := DetailView.Left;
     end
     else if (value = none_style) and (style <> none_style) then begin
        { hide detail view }
        LastListViewWidth := view.ListView.Width;
        DetailView.Visible := false;
        DetailSplitter.Visible := false;
        view.ListView.Align := alClient;
     end;

     style := value;
     ShowImages (style = history_style);
     ShowItem (current);
End;

Procedure TDetail.ViewInfo (Sender: TObject);
Begin
     SetStyle (info_style);
End;

Procedure TDetail.ViewHistory (Sender: TObject);
Begin
     SetStyle (history_style);
End;

Procedure TDetail.ViewPermissions (Sender: TObject);
Begin
     SetStyle (permissions_style);
End;

Procedure TDetail.ViewVersion (Sender: TObject);
Begin
     SetStyle (version_style);
End;

Procedure TDetail.MenuNoDetails (Sender: TObject);
Begin
     SetStyle (none_style);
End;

(********************************* SETUP **********************************)

Procedure SetupMenu (MenuItem: TMenuItem; MenuProc: TNotifyEvent);
Begin
     MenuItem.GroupIndex := 12; {!?}
     MenuItem.RadioItem := true;
     MenuItem.OnClick := MenuProc;
End;

Procedure TDetail.Setup;
Begin
     if SimpleHeader = nil then CreateSimpleHeader;
     if EmptyHeader = nil then CreateEmptyHeader;

     style := none_style;
     LastListViewWidth := 120;

     with DetailView do begin
        Visible := false;
        Align := alClient;
        ViewStyle := vsReport;
     end;

     with DetailSplitter do begin
        Visible := false;
        Align := alLeft;
     end;

     SetupMenu (MenuViewInfo,        ViewInfo);
     SetupMenu (MenuViewHistory,     ViewHistory);
     SetupMenu (MenuViewPermissions, ViewPermissions);
     SetupMenu (MenuViewVersion,     ViewVersion);
     SetupMenu (MenuViewNoDetails,   MenuNoDetails);

     view.OnSelect := SelectItem;
End;

END.

