UNIT FilePro;

INTERFACE

Uses Windows, SysUtils, Classes, Menus,
     BinIO, CRC, ItemPro, VerPro, StopDlg;

{--------------------------------------------------------------------------}

Type TFileItem = class (TItem)
        time:  integer;   { datum a cas posledni zmeny }
        size:  integer;   { delka }
        attr:  integer;   { atributy }

        ext:   string;    { pripona }
        path:  string;    { uplna cesta - pouze pro adresare }

        crc:       integer;   { 32-bit CRC }
        crc_ready: boolean;   { crc je platne }

        ver:       TVerInfo;  { version info }
        ver_ready: boolean;   { true => version info bylo nacteno }

        CreationTime:   TFileTime;
        LastAccessTime: TFileTime;
        LastWriteTime:  TFileTime;

        constructor CreateSimp (p_name: string;
                                p_time, p_size, p_attr: integer);

        constructor CreateItem (p_name: string;
                                p_time, p_size, p_attr: integer;
                                p_create, p_access, p_write: TFileTime);
        procedure Update;

        function  FullName: string;
        procedure Dupl (item: TItem); override;
        function  IsEqual (item: TItem): boolean; override;


        procedure ShowItem (var txt: TColumns); override;
        function  GetSortSet: TSortSet; override;
        function  GetHeader: TBrwHeader; override;
        function  GetSumHeader: TBrwHeader; override;

        function DirName: string; {!?}
        function FileName: string; {!?}


        procedure GetSum (var rec: TCalcRec); override;
        procedure GetCRC; override;
        procedure GetInfo (proc: TShowProc); override;
        procedure GetVer;

        procedure Load (f: TBinRead; v: integer); override;
        procedure Save (f: TBinWrite; v: integer); override;
     end;

Function ScanDir (path: string): TFileItem;

IMPLEMENTATION

(******************************* FILE ITEM ********************************)

Procedure TFileItem.Update;
Var  len, i: integer;
Begin
     { folder }
     folder := (attr and faDirectory) <> 0;

     { extension }
     len := length (name);
     i := len;
     while (i>0) and (name[i]<>'.') do dec (i);
     if i>0 then ext := copy (name, i+1, len-i);
End;

Constructor TFileItem.CreateSimp (p_name: string;
                                  p_time, p_size, p_attr: integer);
Begin
     inherited Create;

     name := p_name;
     time := p_time;
     size := p_size;
     attr := p_attr;

     Update;
End;

Constructor TFileItem.CreateItem (p_name: string;
                                  p_time, p_size, p_attr: integer;
                                  p_create, p_access, p_write: TFileTime);
Begin
     inherited Create;

     name := p_name;
     time := p_time;
     size := p_size;
     attr := p_attr;

     CreationTime := p_create;
     LastAccessTime := p_access;
     LastWriteTime := p_write;

     Update;
End;

Function NewFileItem (var rec: TSearchRec): TFileItem;
Begin
     result := TFileItem.CreateItem
                   (rec.name,
                    rec.time, rec.size, rec.attr,
                    rec.FindData.ftCreationTime,
                    rec.FindData.ftLastAccessTime,
                    rec.FindData.ftLastWriteTime);
End;

(******************************* SCAN FILES *******************************)

Type EFile = class (Exception);

Procedure Error (ans: longword);
Begin
     raise EFile.Create ('File error ' +
                          IntToStr (ans) + ': ' +
                          SysErrorMessage (ans));
End;

Procedure ScanBranch (top: TFileItem; path: string);
Var  rec:  TSearchRec;
     item: TFileItem;
     len:  integer;
     ans:  longword;
Begin
     StopInfo (path);
     if StopFlag then exit;

     top.path := path;
     len := length (path);
     if (len>0) and (path[len]='\') then path := copy (path, 1, len-1);

     ans := FindFirst (path + '\*.*', faAnyFile, rec);
     try
        while ans = 0 do begin
           if (rec.name <> '.') and (rec.name <> '..') then begin
              item := NewFileItem (rec);
              top.Add (item);
           end;
           ans := FindNext (rec);
        end;
     finally
        FindClose (rec);
     end;

     if ans <> ERROR_NO_MORE_FILES then Error (ans);

     { scan subdirectories }
     item := top.sub.first as TFileItem;
     while (item <> nil) and not StopFlag do begin
        if item.attr and faDirectory <> 0 then
           ScanBranch (item, path + '\' + item.name);
        item := item.next as TFileItem;
     end;
End;

Function ScanDir (path: string): TFileItem;
Var  ans:  longword;
     ok:   boolean;
     rec:  TSearchRec;
Begin
     result := nil;
     StopOpen ('Scanning ' + path);

     try
        try

           ans := FindFirst (path, faAnyFile, rec);
           try
              if (ans <> 0) and (ans <> ERROR_FILE_NOT_FOUND) then
                 Error (ans);

              ok := (ans = 0) and
                    (rec.attr and faDirectory <> 0) and
                    (rec.name <> '.') and
                    (rec.name <> '..');

              if ok then result := NewFileItem (rec)
                    else result := TFileItem.CreateSimp (path, 0, 0, faDirectory); {!?}
           finally
              FindClose (rec);
           end;

           if result <> nil then ScanBranch (result, path);

        finally
           StopClose;
        end;

     except
        result.Free;
        raise;
     end;

     if StopFlag then begin
        result.Free;
        result := nil;
     end;
End;

(******************************* FULL NAME ********************************)

Function TFileItem. FullName: string;
Var  p: TItem;
     f: TFileItem;
     cont: boolean;
     c: char;
Begin
     p := self;
     result := '';
     cont := true;

     while cont and (p <> nil) and (p is TFileItem) do begin
        f := TFileItem (p);
        if f.path <> '' then begin
           c := f.path [length (f.path)];
           if (result <> '') and (c in [':', '\'])
              then result := f.path + result
              else result := f.path + '\' + result;
           cont := false;
        end
        else begin
           if result = '' then result := f.name
                          else result := f.name + '\' + result;
        end;
        p := p.up;
     end;
End;

(******************************** COMPARE *********************************)

Procedure TFileItem.Dupl (item: TItem);
Var  p: TFileItem;
Begin
     inherited;
     p := item as TFileItem;

     time := p.time;
     size := p.size;
     attr := p.attr;
     ext  := p.ext;

     crc  := p.crc;
     crc_ready := p.crc_ready;

     { !? CreationTime ... }
End;

Function TFileItem.IsEqual (item: TItem): boolean;
Var   p: TFileItem;
Begin
     p := item as TFileItem;
     result := (size = p.size) and (time = p.time);
End;

(******************************* SHOW FILES *******************************)

Procedure TFileItem.ShowItem (var txt: TColumns);
Var  tmp: string;
     t:   string [4];
Begin
     txt[1] := name;
     txt[2] := ext;

     if not folder then
        txt[3] := IntToStr (size);

     if time <> 0 then
        txt[4] := DateTimeToStr (FileDateToDateTime (time));

     t := '----';
     if attr and faReadOnly <> 0 then t[1] := 'r';
     if attr and faArchive  <> 0 then t[2] := 'a';
     if attr and faHidden   <> 0 then t[3] := 'h';
     if attr and faSysFile  <> 0 then t[4] := 's';
     txt[5] := t;

     if crc_ready then
        txt[6] := IntToHex (crc, 8);

     case mode of
        new:      tmp := 'New';
        old:      tmp := 'Old';
        modified: tmp := 'Modified';
        equal:    tmp := 'Equal';
        else      tmp := '';
     end;
     txt[7] := tmp;
End;

(******************************* SORT FILES *******************************)

Function TFileItem.GetSortSet: TSortSet;
Begin
     result := [ByExt, BySize, ByTime];
End;

(********************************* HEADER *********************************)

Var FileHeader: TBrwHeader = nil;
    FileSumHeader: TBrwHeader = nil;

Procedure CreateFileHeaders;
Begin
     FileHeader := TBrwHeader.Create;
     with FileHeader do begin
        Add ('Name',  110, ByName);
        Add ('Ext',    38, ByExt);
        Add ('Size',   60, BySize);
        Add ('Time',  100, ByTime);
        Add ('Attr',   32, Unsorted);
        Add ('CRC',    72, Unsorted);
        Add ('State',  60, ByColor);
     end;

     FileSumHeader := TBrwHeader.Create;
     with FileSumHeader do begin
        Add ('Name',        110, ByName);
        Add ('Files',        60, ByLeafs);
        Add ('Directories',  60, ByFolders);
        Add ('Bytes',       100, ByBytes);
        Add ('Used',        100, ByUsed);
     end;
End;

Function TFileItem.GetHeader: TBrwHeader;
Begin
     if FileHeader = nil then CreateFileHeaders;
     result := FileHeader;
End;

Function TFileItem.GetSumHeader: TBrwHeader;
Begin
     if FileSumHeader = nil then CreateFileHeaders;
     result := FileSumHeader;
End;

(****************************** DETAIL INFO *******************************)

Function FileTimeToStr (ft: TFileTime): string;
Var  st: TSystemTime;
     dt: TDateTime;
Begin
     FileTimeToSystemTime (ft, st);
     dt := SystemTimeToDateTime (st);
     result := DateTimeToStr (dt);
End;

Procedure TFileItem.GetInfo (proc: TShowProc);
Begin
     proc ('Created',  FileTimeToStr (CreationTime));
     proc ('Modified', FileTimeToStr (LastWriteTime));
     proc ('Accesed',  FileTimeToStr (LastAccessTime));
End;

(****************************** VERSION INFO *******************************)

Function TFileItem.DirName: string;
Begin
     if up is TFileItem
        then result := (up as TFileItem).path
        else result := '';
End;

Function TFileItem.FileName: string;
Begin
     result := DirName + '\' + name;
End;

Procedure TFileItem.GetVer;
Begin
     if (ver=nil) and not ver_ready then begin
        ver := ScanVerInfo (FileName);
        ver_ready := true;
     end;
End;

(********************************** CALC **********************************)

Procedure TFileItem.GetSum (var rec: TCalcRec);
Begin
     if folder then begin
        inc (rec.fil_dirs);
     end
     else begin
        inc (rec.fil_files);
        inc (rec.fil_bytes, size);
     end;
End;

(********************************** CRC ***********************************)

Procedure TFileItem.GetCRC;
Const size = 32 * 1024;
Var   f: file;
      len, blk, save: integer;
      err: boolean;
      buf: array [1..size] of byte;
Begin
     if not folder and not crc_ready then begin
        StopInfo (FileName);

        crc := -1;
        save := FileMode;
        // err := false;

        {$I-}
        try
           FileMode := 0; { read only }
           Assign (f, FileName);
           Reset (f, 1);
           err := IoResult <> 0;

           if not err then
              try
                 blk := 0;
                 while not Eof (f) and not err do begin
                    inc (blk);
                    if blk mod 10 = 0 then StopTest;
                    BlockRead (f, buf, size, len);
                    err :=  (IoResult <> 0);
                    if not err then UpdateCRC (crc, @buf, len);
                 end;
              finally
                 Close (f);
              end;

        finally
           FileMode := save;
        end;
        {$I+}

        if not err then begin
           crc := not crc;
           crc_ready := true;
        end;
     end;
End;

(****************************** LOAD / SAVE *******************************)

Procedure TFileItem.Load (f: TBinRead; v: integer);
Begin
     inherited;

     f.ReadInt (Time);
     f.ReadInt (Size);
     f.ReadInt (Attr);
     f.ReadStr (Ext);
     f.ReadStr (Path);

     if v = 2 then begin
        f.ReadBool (crc_ready);
        f.ReadInt (crc);
     end;

     if v >= 3 then begin
        f.ReadBool (crc_ready);
        if crc_ready then f.ReadInt (crc);
     end;

     { !? CreationTime ... }
End;

Procedure TFileItem.Save (f: TBinWrite; v: integer);
Begin
     inherited;

     f.WriteInt (Time);
     f.WriteInt (Size);
     f.WriteInt (Attr);
     f.WriteStr (Ext);
     f.WriteStr (Path);

     if v = 2 then begin
        f.WriteBool (crc_ready);
        f.WriteInt (crc);
     end;

     if v >= 3 then begin
        f.WriteBool (crc_ready);
        if crc_ready then f.WriteInt (crc);
     end;
End;

END.

