UNIT FiltDlg;

INTERFACE

Uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, ExtCtrls, Buttons,
  ItemPro, ToolWin, Menus;

Type
  TFiltDialog = class (TForm)
    ToolBar: TToolBar;

    BtnNormal: TToolButton;
    BtnNew: TToolButton;
    BtnOld: TToolButton;
    BtnModified: TToolButton;
    BtnEqual: TToolButton;
    BtnSep1: TToolButton;

    BtnChecked: TToolButton;
    BtnUnchecked: TToolButton;
    BtnPartial: TToolButton;
    BtnSep2: TToolButton;

    BtnThisDir: TToolButton;
    BtnSubDir: TToolButton;
    { checkpoint }

    procedure ModeMouseUp  (Sender: TObject; Button: TMouseButton;
                            Shift: TShiftState; X, Y: Integer);
    procedure StateMouseUp (Sender: TObject; Button: TMouseButton;
                            Shift: TShiftState; X, Y: Integer);

    procedure BtnThisDirMouseUp (Sender: TObject; Button: TMouseButton;
                                 Shift: TShiftState; X, Y: Integer);
    procedure BtnSubDirMouseUp  (Sender: TObject; Button: TMouseButton;
                                 Shift: TShiftState; X, Y: Integer);

    procedure FormCreate (Sender: TObject);
    procedure FormShow   (Sender: TObject);
    { checkpoint }

  public
    modes:   TModeSet;
    states:  TStateSet;
    ThisDir: boolean;
    SubDir:  boolean;

  protected
    procedure UpdateButtons;
    procedure Recalc; virtual;
    procedure ChoosePage; virtual;

  private
    procedure SwitchMode (val: TMode);
    procedure SetModes (val: TModeSet);

    procedure SwitchState (val: TState);
    procedure SetStates (val: TStateSet);

    procedure SwitchFlag (var flag: boolean);
    procedure SetFlags (val_this, val_sub: boolean);
  end;

Var
  FiltDialog: TFiltDialog;

IMPLEMENTATION

Uses Brw; { Browser }

{$R *.DFM}

{--------------------------------------------------------------------------}

Procedure TFiltDialog.UpdateButtons;
Begin
     BtnNormal.Down   := normal in modes;
     BtnNew.Down      := new in modes;
     BtnOld.Down      := old in modes;
     BtnModified.Down := modified in modes;
     BtnEqual.Down    := equal in modes;

     BtnChecked.Down   := checked in states;
     BtnPartial.Down   := partial in states;
     BtnUnchecked.Down := unchecked in states;

     BtnSubDir.Down  := SubDir;
     BtnThisDir.Down := ThisDir;
End;

Procedure TFiltDialog.Recalc;
Begin
     { nothing }
End;

Procedure TFiltDialog.ChoosePage;
Begin
     { nothing }
End;

{--------------------------------------------------------------------------}

Procedure TFiltDialog.SwitchMode (val: TMode);
Begin
     if val in modes then Exclude (modes, val)
                     else Include (modes, val);
     Recalc;
End;

Procedure TFiltDialog.SetModes (val: TModeSet);
Begin
     if modes <> val then begin
        modes := val;
        Recalc;
     end;
End;

Procedure TFiltDialog.SwitchState (val: TState);
Begin
     if val in states then Exclude (states, val)
                      else Include (states, val);
     Recalc;
End;

Procedure TFiltDialog.SetStates (val: TStateSet);
Begin
     if states <> val then begin
        states := val;
        Recalc;
     end;
End;

Procedure TFiltDialog.SwitchFlag (var flag: boolean);
Begin
     flag := not flag;
     Recalc;
End;

Procedure TFiltDialog.SetFlags (val_this, val_sub: boolean);
Begin
     if (ThisDir <> val_this) or (SubDir <> val_sub) then begin
        ThisDir := val_this;
        SubDir := val_sub;
        Recalc;
     end;
End;

{--------------------------------------------------------------------------}

Procedure TFiltDialog.ModeMouseUp (Sender: TObject;
                                     Button: TMouseButton;
                                     Shift: TShiftState;
                                     X, Y: Integer);
Const all = [normal, new, old, modified, equal];
Var   val: TMode;
Begin
     if Sender = BtnNew then val := new
     else if Sender = BtnOld then val := old
     else if Sender = BtnModified then val := modified
     else if Sender = BtnEqual then val := equal
     else val := normal;

     if Button = mbLeft then begin
        if Shift = [] then SwitchMode (val);
        if Shift = [ssShift] then SetModes ([val]);
        if Shift = [ssCtrl] then SetModes (all-[val]);
        if Shift = [ssAlt] then SetModes (all);
        UpdateButtons;
     end;
End;

Procedure TFiltDialog.StateMouseUp (Sender: TObject;
                                      Button: TMouseButton;
                                      Shift: TShiftState;
                                      X, Y: Integer);
Const all = [checked, partial, unchecked];
Var   val: TState;
Begin
     if Sender = BtnChecked then val := checked
     else if Sender = BtnPartial then val := partial
     else val := unchecked;

     if Button = mbLeft then begin
        if Shift = [] then SwitchState (val);
        if Shift = [ssShift] then SetStates ([val]);
        if Shift = [ssCtrl] then SetStates (all-[val]);
        if Shift = [ssAlt] then SetStates (all);
        UpdateButtons;
     end;
End;

Procedure TFiltDialog.BtnThisDirMouseUp (Sender: TObject;
                                           Button: TMouseButton;
                                           Shift: TShiftState;
                                           X, Y: Integer);
Begin
     if Button = mbLeft then begin
        if Shift = [] then SwitchFlag (ThisDir);
        if Shift = [ssShift] then SetFlags (true, false);
        if Shift = [ssCtrl] then SetFlags (false, true);
        if Shift = [ssAlt] then SetFlags (true, true);
        UpdateButtons;
     end;
End;

Procedure TFiltDialog.BtnSubDirMouseUp (Sender: TObject;
                                          Button: TMouseButton;
                                          Shift: TShiftState;
                                          X, Y: Integer);
Begin
     if Button = mbLeft then begin
        if Shift = [] then SwitchFlag (SubDir);
        if Shift = [ssShift] then SetFlags (false, true);
        if Shift = [ssCtrl] then SetFlags (true, false);
        if Shift = [ssAlt] then SetFlags (true, true);
        UpdateButtons;
     end;
End;

{--------------------------------------------------------------------------}

Procedure TFiltDialog.FormCreate (Sender: TObject);
Begin
     modes   := [normal, new, old, modified, equal];
     states  := [unchecked, checked];
     SubDir  := true;
     ThisDir := true;
End;

Procedure TFiltDialog.FormShow (Sender: TObject);
Begin
     modes := Browser.view.modes;                     
     states := Browser.view.states;
     ThisDir := true;
     SubDir := true;
     UpdateButtons;
     Recalc;
     ChoosePage;
End;

END.
