UNIT GrpPro;

INTERFACE

Uses Windows, SysUtils, Classes, Menus, DDEMan,
     BinIO, ItemPro, TextPro, StopDlg;

{--------------------------------------------------------------------------}

Type TGroupAttr = (group_file,
                   num_of_items,
                   unknown1);

     TProgAttr = (command_line,
                  default_directory,
                  icon_file,
                  icon_index,
                  position,
                  prog_shortcut,
                  minimize,
                  unknown2);

Const GroupAttrName: array [TGroupAttr] of string =
      ('Group file',
       'Number of items',
       'Unknown');

      ProgAttrName: array [TProgAttr] of string =
      ('Command line',
       'Default directory',
       'Icon file',
       'Icon index',
       'Position',
       'Shortcut',
       'Minimize',
       'Unknown');

{--------------------------------------------------------------------------}

Type TGrpItem = class (TItem)
        info: array [TGroupAttr] of string;

        constructor CreateFolder (p_name: string);

        procedure GetInfo (proc: TShowProc); override;

        procedure Load (f: TBinRead; v: integer); override;
        procedure Save (f: TBinWrite; v: integer); override;
     end;

Type TPrgItem = class (TItem)
        info:  array [TProgAttr] of string;

        constructor CreateItem (p_name: string);

        procedure GetInfo (proc: TShowProc); override;

        procedure Load (f: TBinRead; v: integer); override;
        procedure Save (f: TBinWrite; v: integer); override;
     end;

Function ScanGrp: TGrpItem;

IMPLEMENTATION

Uses Brw; { Browser }

(************************** SCAN PROGRAM GROUPS ***************************)

Type EGrp = class (Exception);

Procedure Error (ans: integer);
Begin
     raise EGrp.Create ('Grp error ' + IntToStr (ans));
End;

Procedure ErrorMsg (msg: string);
Begin
     raise EGrp.Create (msg);
End;

Procedure ScanItem (top: TGrpItem; txt: string);
Var  item: TPrgItem;
     list: TStringList;
     i: integer;
     a: TProgAttr;
Begin
     list := TStringList.Create;
     list.CommaText := txt;

     item := TPrgItem.CreateItem (list [0]);
     top.Add (item);

     i := 1;
     for a := Low (TProgAttr) to High (TProgAttr) do begin
        item.info[a] := list[i];
        inc (i);
     end;

     list.Free;
End;

Procedure ScanGroup (top: TGrpItem; client: TDDEClientConv; name: string);
Var  item: TGrpItem;
     p: PChar;
     list, note: TStringList;
     i: integer;
     a: TGroupAttr;
Begin
     StopTest;
     item := TGrpItem.CreateFolder (name);
     top.Add (item);

     p := client.RequestData (name);
     list := TStringList.Create;
     list.Text := p;
     StrDispose (p);
     if list.Count = 0 then ErrorMsg ('Unexpected text');

     note := TStringList.Create;
     note.CommaText := list [0];

     i := 1;
     for a := Low (TGroupAttr) to High (TGroupAttr) do begin
        item.info[a] := note[i];
        inc (i);
     end;

     for i := 1 to list.Count-1 do
        ScanItem (item, list[i]);

     list.Free;
     note.Free;
End;

Function ScanGrp: TGrpItem;
Var  client: TDDEClientConv;
     p: PChar;
     list: TStringList;
     i: integer;
Begin
     result := TGrpItem.CreateFolder ('Program Groups');
     StopOpen ('Groups');

     try
        try

           client := TDDEClientConv.Create (Browser);
           if not client.SetLink ('Progman', 'Progman')
              then ErrorMsg ('Cannot estabilish DDE link');

           p := client.RequestData ('Groups');
           list := TStringList.Create;
           list.Text := p;
           StrDispose (p);

           try
              for i := 0 to list.Count-1 do
                 ScanGroup (result, client, list[i]);
           finally
              client.Free;
              list.Free;
           end;

        finally
           StopClose;
        end;

     except
        result.Free;
        raise;
     end;

     if StopFlag then begin
        result.Free;
        result := nil;
     end;
End;

(********************************** INIT *********************************)

Constructor TGrpItem.CreateFolder (p_name: string);
Begin
     inherited Create;
     name   := p_name;
     folder := true;
End;

Constructor TPrgItem.CreateItem (p_name: string);
Begin
     inherited Create;
     name  := p_name;
End;

(********************************** INFO **********************************)

Const quo = '"';

Procedure TGrpItem.GetInfo (proc: TShowProc);
Var  a: TGroupAttr;
Begin
     proc ('Group ', name);

     for a := Low (TGroupAttr) to High (TGroupAttr) do
        proc (GroupAttrName[a], info[a]);
End;

Procedure TPrgItem.GetInfo (proc: TShowProc);
Var  a: TProgAttr;
Begin
     proc ('Program ' , name);

     for a := Low (TProgAttr) to High (TProgAttr) do
        proc (ProgAttrName[a], info[a]);
End;

(****************************** LOAD / SAVE *******************************)

Procedure TGrpItem.Load (f: TBinRead; v: integer);
Var  a: TGroupAttr;
Begin
     inherited;
     for a := Low (TGroupAttr) to High (TGroupAttr) do
        f.ReadStr (info [a]);
End;

Procedure TGrpItem.Save (f: TBinWrite; v: integer);
Var  a: TGroupAttr;
Begin
     inherited;
     for a := Low (TGroupAttr) to High (TGroupAttr) do
        f.WriteStr (info [a]);
End;

Procedure TPrgItem.Load (f: TBinRead; v: integer);
Var  a: TProgAttr;
Begin
     inherited;
     for a := Low (TProgAttr) to High (TProgAttr) do
        f.ReadStr (info [a]);
End;

Procedure TPrgItem.Save (f: TBinWrite; v: integer);
Var  a: TProgAttr;
Begin
     inherited;
     for a := Low (TProgAttr) to High (TProgAttr) do
        f.WriteStr (info [a]);
End;

END.

