UNIT IniPro;

INTERFACE

Uses Windows, SysUtils, Classes,
     ItemPro, TextPro, StopDlg;

Type TIniItem = TTextItem;

Function ScanIni (path: string): TIniItem;

IMPLEMENTATION

(***************************** SCAN INI FILES *****************************)

Type EIni = class (Exception);

Procedure Error (ans: integer);
Begin
     raise EIni.Create ('Ini file error ' + IntToStr (ans));
End;

Procedure ErrorMsg (msg: string);
Begin
     raise EIni.Create (msg);
End;

Procedure ScanIniSection (top: TIniItem; path, section: string);
Const size = 30*1024;
Var   item, tmp: TIniItem;
      ans, len, inx:  integer;
      buf, ref:  PChar;
      line, name, value: string;
Begin
     item := TIniItem.CreateFolder ('['+section+']');
     top.Add (item);
     GetMem (buf, size);

     ans := GetPrivateProfileSection
            (PChar (section), // section name
             buf,             // address of destination buffer
             size,            // size of destination buffer
             PChar (path));   // initialization filename

     if ans >= size-2 then ErrorMsg ('Buffer too small');

     ref := buf;
     while ref^ <> chr (0) do begin
        line := ref;
        inx := pos ('=', line);

        if inx = 0 then begin
           name := line;
           value := '';
        end
        else begin
           name := copy (line, 1, inx-1);
           value := copy (line, inx+1, length (line)-inx);
        end;

        tmp := TIniItem.CreateItem (name, value);
        item.Add (tmp);

        len := StrLen (ref);
        inc (ref, len+1);
     end;

     FreeMem (buf);
End;

Procedure ScanIniFile (top: TIniItem; path, name: string);
Const size = 8*1024;
Var   item: TIniItem;
      ans, len: integer;
      buf, ref: PChar;
Begin
     path := path + '\' + name;
     StopInfo ('INI file ' + path);
     if StopFlag then exit;

     item := TIniItem.CreateFolder (name);
     top.Add (item);
     GetMem (buf, size);

     ans := GetPrivateProfileString
            (nil,            // section name
             nil,            // key name
             nil,            // default string
             buf,            // address of destination buffer
             size,           // size of destination buffer
             PChar (path));  // initialization filename

     if ans >= size-2 then ErrorMsg ('Buffer too small');

     ref := buf;
     while ref^ <> chr (0) do begin
        ScanIniSection (item, path, ref);
        len := StrLen (ref);
        inc (ref, len+1);
     end;

     FreeMem (buf);
End;

Function ScanIni (path: string): TIniItem;
Var  ans: integer;
     rec: TSearchRec;
Begin
     result := TIniItem.CreateFolder ('INI files');
     StopOpen ('INI files');

     try
        try

           ans := FindFirst (path + '\*.INI', faAnyFile, rec);
           try

              while ans = 0 do begin
                 ScanIniFile (result, path, rec.name);
                 ans := FindNext (rec);
              end;

           finally
              FindClose (rec);
              if ans <> ERROR_NO_MORE_FILES then Error (ans);
           end;

        finally
           StopClose;
        end;

     except
        result.Free;
        raise;
     end;

     if StopFlag then begin
        result.Free;
        result := nil;
     end;
End;

END.

