UNIT ItemPro;

INTERFACE

Uses SysUtils, Classes, Menus, ComCtrls,
     BinIO, StopDlg;

(********************************** SORT **********************************)

Type TSort = (ByName,
              ByExt,
              ByTime,
              BySize,
              ByValue,
              ByColor,
              Unsorted,

              ByLeafs,
              ByFolders,
              ByBytes,
              ByUsed);

Type TSortSet = set of TSort;

(********************************* STATE **********************************)

Type TState = (unchecked,
               checked,
               partial);

Type TStateSet = set of TState;

(********************************** MODE **********************************)

Type TMode = (normal,
              equal,
              modified,
              old,
              new);

Const modif = modified;

Type TModeSet = set of TMode;

(******************************** CATEGORY ********************************)

Type TCategory = (file_category,
                  dir_category);

     TCategorySet = set of TCategory;

(********************************* IMAGE **********************************)

Type TPict = (std_pict,
              page_pict,
              box_pict,
              tree_pict);

(********************************* HEADER *********************************)

Const MaxColumn = 7;

Type TColumns = array [1..MaxColumn] of string;

     TBrwColumn = record
        c_title:    string;
        c_width:    integer;
        c_sort:     TSort;
        c_autosize: boolean;
        c_enabled:  boolean;
     end;

     TBrwHeader = class
     public
        cnt: integer;
        tab: array [1..MaxColumn] of TBrwColumn;

        procedure Add (p_title:    string;
                       p_width:    integer;
                       p_sort:     TSort;
                       p_autosize: boolean = false);
     end;

(********************************** CALC **********************************)

Type TCalcRec = record
        fil_dirs:    integer;  { pocet adresaru }
        fil_files:   integer;  { pocet souboru }
        fil_bytes:   integer;  { pocet bytu v souborech }
        fil_used:    integer;  { spotrebovane misto }

        reg_keys:    integer;  { pocet klicu v registru }
        reg_values:  integer;  { pocet hodnot v registru }
        reg_bytes:   integer;  { pocet bytu pro hodnoty }
        reg_used:    integer;  { spotrebovane misto (klice a hodnoty) }

        all_folders: integer;  { pocet adresarovych polozek }
        all_nodes:   integer;  { poccet koncovych polozek }
        all_bytes:   integer;  { pocet bytu }
        all_used:    integer;  { spotrebovane misto }
     end;

(********************************** ITEM **********************************)

Type TItem = class;

     TQueue = record
        first: TItem;
        last:  TItem;
     end;

     TShowProc = procedure (name, value: string) of object;

     TItem = class
        prev:   TItem;     { predchazejici prvek }
        next:   TItem;     { nasledujici prvek }

        up:     TItem;     { nadrazeny prvek }
        sub:    TQueue;    { vnorene prvky }

        ref:    TTreeNode; { odkaz na prvek TreeView }
        lst:    integer;   { pozice uvnitr ListView (od 0) }

        access:  TQueue;   { pristupova prava }
        history: TQueue;   { vnorene prvky pro porovnavani }

        { nahravane polozky }

        index:  integer;   { puvodni poradove cislo }
        folder: boolean;   { adresar }
        pict:   TPict;     { obrazek }

        mode:   TMode;     { vysledek porovnani }
        state:  TState;    { vybrana polozka }

        name:   string;    { jmeno }

        { pomocne polozky }

        sum_leafs:   integer;
        sum_folders: integer;
        sum_bytes:   integer;
        sum_used:    integer;

        { init / cleanup }
        constructor Create;
        destructor Destroy; override;

        { pridavani / odstranovani prvku }
        procedure Add (item: TItem);
        procedure AddFirst (item: TItem);
        procedure AddLast (item: TItem);
        procedure AddPrev (item: TItem);
        procedure AddNext (item: TItem);
        procedure Unlink;

        procedure AddHist (item: TItem);

        { zobrazeni polozky v tabulce }
        function GetHeader: TBrwHeader; virtual;
        function GetSumHeader: TBrwHeader; virtual;

        procedure ShowItem (var txt: TColumns); virtual;
        procedure ShowSumItem (var txt: TColumns); virtual;

        function GetImageIndex (sel: boolean): integer;
        function GetStateIndex: integer;
        function GetSortSet: TSortSet; virtual;

        function SectionName: string; {!?}

        { kopirovani a porovnavani }
        procedure Dupl (item: TItem); virtual;
        function  IsEqual (item: TItem): boolean; virtual;

        { soucty }
        procedure GetSum (var rec: TCalcRec); virtual;

        { CRC }
        procedure GetCRC; virtual;

        { detailni informace }
        procedure GetInfo (proc: TShowProc); virtual;

        { load / save }
        procedure Load (f: TBinRead; v: integer); virtual;
        procedure Save (f: TBinWrite; v: integer); virtual;
     end;

(******************************* ALLOCATION *******************************)

Var ItemCount: integer = 0;
    ItemSize:  integer = 0;

Function NewRoot: TItem;
Function NewBox (p_name: string): TItem;

(****************************** LOAD / SAVE *******************************)

Type  TItemClass = class of TItem;

Procedure RegisterItem (inx: integer; cls: TItemClass);

Function  LoadCtl (fname: string): TItem;
Procedure SaveCtl (fname: string; item: TItem);

IMPLEMENTATION

(********************************* ERROR **********************************)

Procedure Error (s: string);
Begin
     raise Exception.Create (s);
End;

(********************************* HEADER *********************************)

Procedure TBrwHeader.Add (p_title:    string;
                          p_width:    integer;
                          p_sort:     TSort;
                          p_autosize: boolean = false);

Begin
     inc (cnt);
     if cnt > MaxColumn then Error ('Too many columns');

     with tab [cnt] do begin
        c_title    := p_title;
        c_width    := p_width;
        c_sort     := p_sort;
        c_autosize := p_autosize;
        c_enabled  := true;
     end;
End;

(****************************** NEW FOLDER ********************************)

Function NewRoot: TItem;
Var  item: TItem;
Begin
     item := TItem.Create;
     item.name := 'Tree';
     item.folder := true;
     item.pict := tree_pict;
     result := item;
End;

Function NewBox (p_name: string): TItem;
Var  item: TItem;
Begin
     item := TItem.Create;
     item.name := p_name;
     item.folder := true;
     item.pict := box_pict;
     result := item;
End;

(**************************** CREATE / DESTROY ****************************)

Constructor TItem.Create;
Begin
     inc (ItemCount);
     inc (ItemSize, InstanceSize);
     lst := -1;
End;

Destructor TItem.Destroy;
Var  p, t: TItem;
Begin
     if ref <> nil then Error ('Visible item');

     dec (ItemCount);
     dec (ItemSize, InstanceSize);

     p := history.first;
     while p <> nil do begin
        t := p;
        p := p.next;
        t.Free;
     end;

     p := sub.first;
     while p <> nil do begin
        t := p;
        p := p.next;
        t.Free;
     end;
End;

(******************************** ADD ITEM ********************************)

Procedure TItem.Add (item: TItem);
Begin
     item.prev := sub.last;
     item.next := nil;
     item.up   := self;

     if sub.first = nil then begin
        sub.first := item;
        item.index := 1;
     end
     else begin
        sub.last.next := item;
        item.index := sub.last.index + 1;
     end;

     sub.last := item;

     { pouze Add nastavuje 'index'
       AddFirst, ..., Unlink polozku 'index' neovlivnuji }
End;

(****************************** INSERT ITEMS ******************************)

Procedure Link (root: TItem; item: TItem; bef, aft: TItem);
{ Zacleneni prvku 'item' do seznamu 'root' }
{  parametry 'bef' a 'aft' udavaji predchozi a nasledujici prvky }
Begin
     if (root=nil) or (item=nil) then Error ('Cannot insert');

     { test, zda prvek neni zaclenen do nejakeho seznamu }
     if (item.prev<>nil) or (item.next<>nil) then Error ('Already in list');

     item.prev := bef;
     item.next := aft;
     item.up   := root;

     if item.prev=nil then root.sub.first := item
                      else item.prev.next := item;

     if item.next=nil then root.sub.last := item
                      else item.next.prev := item;
End;

Procedure TItem.AddFirst (item: TItem);
{ Vlozeni prvku 'item' na zacatek seznamu vnorenych prvku }
Begin
     Link (self, item, nil, sub.first);
End;

Procedure TItem.AddLast (item: TItem);
{ Vlozeni prvku 'item' na konec seznamu vnorenych prvku }
Begin
     Link (self, item, sub.last, nil);
End;

Procedure TItem.AddPrev (item: TItem);
{ Vlozeni prvku 'item' pred prvek 'self' }
Begin
     Link (up, item, prev, self);
End;

Procedure TItem.AddNext (item: TItem);
{ Vlozeni prvku 'item' do seznamu za prvek 'self' }
Begin
     Link (up, item, self, next);
End;

(****************************** UNLINK ITEM *******************************)

Procedure TItem. Unlink;
{ Odpojeni prvku ze seznamu }
Begin
     if up=nil then Error ('Cannot unlink top level item');

     { vynech prvek ze seznamu }
     if prev=nil then up.sub.first := next
                 else prev.next := next;

     if next=nil then up.sub.last := prev
                 else next.prev := prev;

     { oznac prvek jako volny }
     prev := nil;
     next := nil;
End;

(************************ ADD ITEM TO HISTORY LIST ************************)

Procedure TItem.AddHist (item: TItem);
Begin
     item.prev := history.last;
     item.next := nil;
     item.up   := self;

     if history.first = nil then begin
        history.first := item;
        item.index := 1;
     end
     else begin
        history.last.next := item;
        item.index := history.last.index + 1;
     end;

     history.last := item;
End;

(******************************* SHOW ITEM ********************************)

Var ItemHeader: TBrwHeader = nil;

Procedure CreateItemHeader;
Begin
     ItemHeader := TBrwHeader.Create;
     with ItemHeader do begin
        Add ('Name', 100, ByName);
     end;
End;

Function TItem.GetHeader: TBrwHeader;
Begin
     result := ItemHeader;
End;

Function TItem.GetSumHeader: TBrwHeader;
Begin
     result := ItemHeader; { dalsi sloupce se obecne nezobrazuji }
End;

{--------------------------------------------------------------------------}

Procedure TItem.ShowItem (var txt: TColumns);
Begin
     txt[1] := name;
End;

Function Conv10 (n: integer): string;
Var  s: string;
Begin
     str (n:10, s);
     result := copy (s, 1, 1) + ' ' +
               copy (s, 2, 3) + ' ' +
               copy (s, 5, 3) + ' ' +
               copy (s, 8, 3);
End;

Function Conv6 (n: integer): string;
Var  s: string;
Begin
     if n < 1000000 then begin
        str (n:6, s);
        result := copy (s, 1, 3) + ' ' + copy (s, 4, 3);
     end
     else result := Conv10 (n);
End;

Procedure TItem.ShowSumItem (var txt: TColumns);
Begin
     txt[1] := name;
     txt[2] := Conv6  (sum_leafs);
     txt[3] := Conv6  (sum_folders);
     txt[4] := Conv10 (sum_bytes);
     txt[5] := Conv10 (sum_used);
End;

{--------------------------------------------------------------------------}

Function TItem.GetImageIndex (sel: boolean): integer;
Const Box           = 0;
      ClosedFolder  = 1;
      OpenedFolder  = 2;
      Page          = 3;

      norm_inx      =  0;  { box, closed folder, opened folder, page }
      red_inx       =  4;
      green_inx     =  8;
      blue_inx      = 12;
      yellow_inx    = 16;
      gray_inx      = 20;

      tree_inx      = 24; { single picture }

Var   tmp: integer;
Begin
     case mode of
        equal:    tmp := gray_inx;
        modified: tmp := yellow_inx;
        new:      tmp := green_inx;
        old:      tmp := red_inx;
        else      tmp := norm_inx;
     end;

     case pict of
        tree_pict:   result := tree_inx;
        box_pict:    result := tmp + Box;
        page_pict:   result := tmp + Page;
        else         if not folder then result := tmp + Page
                     else if sel then result := tmp + OpenedFolder
                                 else result := tmp + ClosedFolder;
     end;
End;

Function TItem.GetStateIndex: integer;
Const unchecked_inx = 1;
      checked_inx   = 2;
      partial_inx   = 3;
Begin
     case state of
        unchecked: result := unchecked_inx;
        checked:   result := checked_inx;
        partial:   result := partial_inx;
        else       result := 0;
     end;
End;

{--------------------------------------------------------------------------}

Function TItem.GetSortSet: TSortSet;
Begin
     result := [];
End;

Function TItem.SectionName: string;
Var  p: TItem;
Begin
     p := self;
     while (p <> nil) and (p.pict<>box_pict) do
        p := p.up;

     if p = nil then result := ''
                else result := p.name;
End;

(************************** DUPLICATE / COMPARE ***************************)

Procedure TItem.Dupl (item: TItem);
{ Kopirovani hodnot z 'item' do 'self' }
Begin
     index := item.index;
     folder := item.folder;
     pict := item.pict;
     name := item.name;
     { 'mode' a 'state' se nekopiruji }
End;

Function TItem.IsEqual (item: TItem): boolean;
{ Porovnani prvku 'self' a 'item' }
Begin
     result := true;
End;

(********************************** INFO **********************************)

Procedure TItem.GetSum (var rec: TCalcRec);
Begin
     { nothing }
End;

Procedure TItem.GetCRC;
Begin
     { nothing }
End;

Procedure TItem.GetInfo (proc: TShowProc);
Begin
     { nothing }
End;

(****************************** LOAD / SAVE *******************************)

Const VerMin = 1;
      VerMax = 3;
      FileMark: longword = $ABCD1234;
      ItemMark: longword = $ABCD0000;

      { ver 2 - TFileItem - crc }

      { ver 3 - komprese pomoci zlib
              - TFileItem - usporneji crc }

Const RegMax  = 8;
Var   RegTab: array [1..RegMax] of TItemClass;

Procedure RegisterItem (inx: integer; cls: TItemClass);
Begin
     if (inx < 1) or (inx > RegMax) or (RegTab [inx] <> nil) then
        Error ('Class table full');
     RegTab [inx] := cls;
End;

Procedure InitRegTab;
Begin
     FillChar (RegTab, sizeof (RegTab), 0);
End;

{--------------------------------------------------------------------------}

Procedure TItem.Load (f: TBinRead; v: integer);
{ Precti jeden prvek }
Begin
     f.ReadCode (Index);
     f.ReadBool (Folder);
     f.ReadByte (byte (Pict));
     f.ReadByte (byte (Mode));
     f.ReadByte (byte (State));
     f.ReadStr  (Name);
End;

Procedure TItem.Save (f: TBinWrite; v: integer);
{ Uloz jeden prvek }
Begin
     f.WriteCode (Index);
     f.WriteBool (Folder);
     f.WriteByte (byte (Pict));
     f.WriteByte (byte (Mode));
     f.WriteByte (byte (State));
     f.WriteStr  (Name);
End;

{--------------------------------------------------------------------------}

Function LoadBranch (f: TBinRead; v: integer): TItem;
{ Precti zaznam a vnorene zaznamy }
Var  inx: longword;
     cnt, i: integer;
     cls: TItemClass;
     item, tmp: TItem;
Begin
     { znacka a typ zaznamu }
     f.ReadOrd (inx);
     dec (inx, ItemMark);
     if (inx < 1) or (inx > RegMax) or (RegTab [inx] = nil) then
        Error ('Unknown class');
     cls := RegTab [inx];

     { vlastni zaznam }
     item := cls.Create;
     item.Load (f, v);

     { history }
     f.ReadCode (cnt);
     for i := 1 to cnt do begin
        tmp := LoadBranch (f, v);
        item.AddHist (tmp);
     end;

     { vnorene zaznamy }
     f.ReadCode (cnt);
     for i := 1 to cnt do begin
         tmp := LoadBranch (f, v);
         item.Add (tmp);
     end;

     result := item;
End;

Procedure SaveBranch (f: TBinWrite; item: TItem; v: integer);
{ Uloz zaznam a vnorene zaznamy }
Var  cls: TItemClass;
     inx: longword;
     cnt: integer;
     p:   TItem;
Begin
     { znacka a typ zaznamu }
     cls := TItemClass (item.ClassType);
     inx := 1;
     while (inx <= RegMax) and (RegTab[inx] <> cls) do inc (inx);
     if inx > RegMax then Error ('Unknown class');
     inc (inx, ItemMark);

     f.WriteOrd (inx);
     item.Save (f, v);

     { history }
     cnt := 0;
     p := item.history.first;
     while p <> nil do begin
        inc (cnt);
        p := p.next;
     end;
     f.WriteCode (cnt);

     p := item.history.first;
     while p <> nil do begin
        SaveBranch (f, p, v);
        p := p.next;
     end;

     { sub-items }
     cnt := 0;
     p := item.sub.first;
     while p <> nil do begin
        inc (cnt);
        p := p.next;
     end;
     f.WriteCode (cnt);

     p := item.sub.first;
     while p <> nil do begin
        SaveBranch (f, p, v);
        p := p.next;
     end;
End;

{--------------------------------------------------------------------------}

Function LoadCtl (fname: string): TItem;
{ Precti data ze souboru }
Var  f: TBinRead;
     n: longword;
     v: integer;
Begin
     BeginWait ('Loading ' + fname);
     try

        f := TBinRead.Create;
        try
           f.Open (fname);

           { znacka souboru }
           f.ReadOrd (n);
           if n <> FileMark then Error ('Invalid file data');

           { verze }
           f.ReadInt (v);
           if (v < VerMin) or (v > VerMax) then Error ('Invalid file version');

           if v >= 3 then f.StartDecompression
                     else f.StartBuffering;

           result := LoadBranch (f, v);

        finally
           f.Close;
           f.Free;
        end;

     finally
        EndWait;
     end;
End;

Procedure SaveCtl (fname: string; item: TItem);
{ Uloz data do souboru }
Const v = VerMax;
Var   f: TBinWrite;
Begin
     BeginWait ('Saving ' + fname);
     try

        f := TBinWrite.Create;
        try
           f.Open (fname);
           f.WriteOrd (FileMark);
           f.WriteInt (v);

           if v >= 3 then f.StartCompression
                     else f.StartBuffering;

           SaveBranch (f, item, v);

        finally
           f.Close;
           f.Free;
        end;

     finally
        EndWait;
     end;
End;

(**************************************************************************)

BEGIN
     InitRegTab;
     CreateItemHeader;
END.

