UNIT PackPro;

INTERFACE

Uses Windows, SysUtils, ShellAPI, Forms,
     TxtIO, ItemPro, FilePro, StopDlg;

Function ScanIComp (fname: string): TItem;
Function ScanCab (fname: string): TItem;
Function ScanPak (fname: string): TItem;

Function ScanArc (fname: string): TItem;
Function ScanArj (fname: string): TItem;
Function ScanLzh (fname: string): TItem;
Function ScanZip (fname: string): TItem;

Function LoadDir (FileName: string): TItem;

IMPLEMENTATION

(********************************* ERROR **********************************)

Procedure Error (msg: string);
Begin
     raise Exception.Create (msg);
End;

(******************************** EXECUTE *********************************)

Var TempName: string;

Procedure Exec (cmd, param: string);

Var  fname, path, tmp: array [0..MAX_PATH-1] of char;
     Attr: TSecurityAttributes;
     hFile: THandle;
     StartupInfo: TStartupInfo;
     ProcessInfo: TProcessInformation;
     ok: boolean;
     ans: integer;
     txt: array [0..255] of char;
     f: text;
Begin
     ans := FindExecutable (PChar (cmd), nil, fname);
     if ans <= 32 then
        Error ('File not found: ' + cmd);

     param := string (fname) + ' ' + param;

     GetTempPath (sizeof (path), path);
     GetTempFileName (path, 'elm', 0, tmp);
     TempName := tmp;

     Attr.nLength := sizeof (TSecurityAttributes);
     Attr.bInheritHandle := TRUE;
     Attr.lpSecurityDescriptor := nil;

     hFile := CreateFile (PChar (TempName),
                          GENERIC_READ or GENERIC_WRITE,
                          0, { no sharing }
                          @ Attr,
                          CREATE_ALWAYS,
                          FILE_ATTRIBUTE_NORMAL,
                          0); { no template }

     if hFile = INVALID_HANDLE_VALUE then
        Error ('File creation failed');

     FillChar (StartupInfo, sizeof (StartupInfo), 0);
     with StartupInfo do begin
        cb := sizeof (StartupInfo);
        dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
        wShowWindow := SW_HIDE;
        hStdInput := GetStdHandle (STD_INPUT_HANDLE);
        hStdOutput := hFile;
        hStdError := GetStdHandle (STD_ERROR_HANDLE);;
     end;

     ok := CreateProcess
        (fname,         { application name }
         PChar (param), { command line }
         nil,           { process security attributes }
         nil,           { thread security attributes }
         true,          { inherit handles }
         0,             { flags }
         nil,           { environment }
         nil,           { current directory }
         StartupInfo,
         ProcessInfo);

     if not ok then begin
        FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM,
                       nil, { message source }
                       GetLastError (),
                       0, { language }
                       txt,
                       sizeof (txt),
                       nil);

        Error ('CreateProcess failed (' + string (txt) + ')');
     end;

     with ProcessInfo do begin
        ans := WaitForSingleObject (hProcess, 60000);
        if ans = WAIT_TIMEOUT then
           Error ('Wait for process timeout');

        if not CloseHandle (hThread) then
           Error ('Error closing tread handle');

        if not CloseHandle (hProcess) then
           Error ('Error closing process handle');
     end;

     if not CloseHandle (hFile) then
        Error ('Closing handle failed');

     Assign (f, TempName);
     {$I-}
     Reset (f);
     {$I+}
     if IoResult <> 0 then
        Error ('Cannot open redirection file');
     if FileSize (f) = 0 then
        Error ('Empty redirection file');
     Close (f);
End;

Procedure Shell (cmd, param: string);
Var  txt: array [0..MAX_PATH-1] of char;
Begin
     GetEnvironmentVariable ('comspec', txt, sizeof (txt));
     Exec (txt, '/c ' + cmd + ' ' + param);
End;

(********************************* STORE **********************************)

Type PackRec = record
        top: TItem;
        cur: TItem;
        txt: string;
     end;

Function CreateDir (path: string): TFileItem;
Begin
     result := TFileItem.CreateSimp (path, 0, 0, faDirectory);
End;

Procedure Init (var p: PackRec; fname: string);
Begin
     fname := ExtractFileName (fname);
     p.top := CreateDir (fname);
     p.cur := p.top;
     p.txt := '';
End;

Procedure Clear (var p: PackRec);
Begin
     p.top.Free;
     p.top := nil;
     p.cur := nil;
End;

Procedure Lookup (var p: PackRec; path: string);
Var  i, k, len: integer;
     eq: boolean;
     tmp: TItem;
     part: string;
Begin
     len := length (p.txt);
     eq := (copy (path, 1, len) = p.txt) and
           (length (path) > len+1) and
           (path [len+1] = '\');

     if eq then
        i := len+2
     else begin
        p.cur := p.top;
        i := 1;
     end;

     len := length (path);
     while i <= len do begin
        { oddel jmeno directory }
        k := i;
        while (k<=len) and (path[k] <> '\') do inc (k);
        if k=i then Error ('Empty name');
        part := copy (path, i, k-i);
        if k <= len then i := k+1
                    else i := k;

        { hledej directory }
        tmp := p.cur.sub.first;
        while (tmp <> nil) and ((tmp.name <> part) or not tmp.folder) do
           tmp := tmp.next;
           { v registry je mozne pouzit stejne jmeno pro 'key' i 'value' }

        if tmp = nil then begin
           { directory nebylo nalezeno }
           tmp := CreateDir (part);
           p.cur.Add (tmp);
        end;
        p.cur := tmp;
     end;

     p.txt := path;
End;

Procedure StoreSpec (var p: PackRec;
                     path, name: string;
                     size, time, attr: integer;
                     crc_ready: boolean; crc: integer);
Var  item: TFileItem;
Begin
     if p.txt <> path then Lookup (p, path);

     item := TFileItem.CreateSimp (name, time, size, attr);
     item.crc_ready := crc_ready;
     item.crc := crc;
     p.cur.Add (item);
End;

Procedure StoreCRC (var p: PackRec;
                    name: string;
                    size, time, attr: integer;
                    crc_ready: boolean; crc: integer);
Var  len: integer;
Begin
     len := length (name);
     if name [len] = '\' then
        name := copy (name, 1, len-1);

     StoreSpec (p, ExtractFilePath (name),
                ExtractFileName (name), size, time, attr,
                crc_ready, crc);
End;

Procedure Store (var p: PackRec;
                 name: string;
                 size, time, attr: integer);
Begin
     StoreCRC (p, name, size, time, attr, false, 0);
End;

(********************************* ICOMP **********************************)

Function ScanIComp (fname: string): TItem;
Var  size, time, attr: integer;
     name: string;
     ans: PackRec;
Begin
     BeginWait ('Importing ' + fname);
     result := nil;
     try

        Exec ('icomp', '-l ' + fname);
        Init (ans, fname);
        try
           OpenInp (TempName);
           try
              FindLine  (' Date     Time  OrigSize Attr CompSize Name');
              CheckLine (' ======== ===== ======== ==== ======== ====');

              while PeekChar (2) <> ' ' do begin
                 time := GetStdDateTime; { 2 columns }
                 size := GetInt;
                 attr := GetAttr;
                 SkipInt; { comp size }
                 name := GetName;
                 CheckEOL;
                 GetLine;
                 Store (ans, name, size, time, attr);
              end;

              CheckLine ('                --------      -------- ---------');
              CheckLine ('                OrigSize      CompSize FileCount');
              CheckLine ('                --------      -------- ---------');

           finally
              CloseInp;
              EraseInp;
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

(********************************** CAB ***********************************)

Function ScanCab (fname: string): TItem;
Var  name: string;
     size, time, attr: integer;
     ans: PackRec;
Begin
     BeginWait ('Importing ' + fname);
     result := nil;
     try

        Exec ('extract', '-d ' + fname);
        Init (ans, fname);
        try
           OpenInp (TempName);
           try
              FindPartLine ('Microsoft (R) Diamond Extraction Tool ');
              while not InpEof and not (InpChar in ['0'..'9']) do GetLine;

              while PeekChar (1) <> ' ' do begin
                 time := GetStdDateTime; { 2 columns }
                 attr := GetAttr;
                 size := StrToInt (GetCommaName);
                 name := GetName;
                 CheckEOL;
                 GetLine;
                 Store (ans, name, size, time, attr);
              End;

              CheckPartLine ('');

           finally
              CloseInp;
              EraseInp;
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

(********************************** PAK ***********************************)

Function ScanPak (fname: string): TItem;
Var  name: string;
     size, time, attr: integer;
     ans: PackRec;
Begin
     BeginWait ('Importing ' + fname);;
     result := nil;
     try

        Exec ('unpaq', '-v ' + fname);
        Init (ans, fname);
        try
           OpenInp (TempName);
           try
              FindLine  (' Size     Date      Time      File name     Comment');
              CheckLine (' -------  --------  --------  ------------  -----------');

              while PeekChar (2) <> '=' do begin
                 size := GetInt;
                 time := GetStdDateTime;  { 2 columns }
                 name := GetName;
                 { skip comment }
                 GetLine;
                 attr := 0;
                 Store (ans, name, size, time, attr);
              end;

              CheckLine (' =======');
              
           finally
              CloseInp;
              EraseInp;
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

(********************************** ARC ***********************************)

Const bad_header  = 'Bad archive header';
      int_lim     = 'Internal buffer too small';
      old_version = 'Old archive version';

Procedure Open (var f: file; fname: string);
Var  save: byte;
Begin
     save := FileMode;
     FileMode := 0; { read only }
     try
        Assign (f, fname);
        Reset (f, 1);
     finally
        FileMode := save;
     end;
End;

Procedure ModifyPath (var path: string);
Var  i: integer;
Begin
     for i := 1 to length (path) do
        if path [i] = '/' then path [i] := '\';
End;

Function ScanArc (fname: string): TItem;

Type ArcHeader = packed record
        mark:      byte;                  { 1A hex }
        ver:       byte;                  { header version }
        name:      array [0..12] of char; { file name }
        size:      longint;               { size of compressed file }
        date:      word;                  { file date }
        time:      word;                  { file time }
        crc:       word;                  { cyclic redundancy check }
        orig_size: longint;               { size of uncompressed file }
     end;

Var   ans:  PackRec;
      f:    file;
      fp:   integer;
      stop: boolean;
      hdr:  ArcHeader;
      time: integer;
      len:  integer;

Begin
     BeginWait ('Importing ' + fname);;
     result := nil;
     try

        Init (ans, fname);
        try
           Open (f, fname);
           try

              fp := 0;
              repeat
                 Seek (f, fp);
                 BlockRead (f, hdr, sizeof (hdr), len);
                 if (len < 2) or (hdr.mark <> $1A) then Error (bad_header);

                 stop := (hdr.ver = 0);
                 if hdr.ver = 1 then Error (old_version);

                 if not stop then begin
                    if len <> sizeof (hdr) then Error (bad_header);

                    { next position }
                    inc (fp, sizeof (hdr)+hdr.size);


                    time := longint (hdr.date) shl 16 + hdr.time;
                    Store (ans, hdr.name, hdr.orig_size, time, 0);
                 end; {if}
              until stop;

           finally
              Close (f);
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

(********************************** ARJ ***********************************)

Function ScanArj (fname: string): TItem;

Const GARBLED_FLAG  = $01;
      RESERVED_FLAG = $02;
      VOLUME_FLAG   = $04;
      EXTFILE_FLAG  = $08;
      PATHSYM_FLAG  = $10;

Type ArjInfo = packed record
        id:          word; { header id (main and local file) = 0xEA60 or 60000U }
        header_size: word; { basic header size
                             (from 'first_hdr_size' thru 'comment' below)
                              = first_hdr_size + strlen(filename) + 1
                                               + strlen(comment) + 1
                              = 0 if end of archive }
     end;

Type ArjHeader = packed record
        first_hdr_size:    byte;    { size up to and including 'extra data' }
        arj_ver:           byte;    { archiver version number }
        arj_extr_ver:      byte;    { minimum archiver version to extract }
        host_os:           byte;    { 0 = MSDOS, 1 = PRIMOS, 2 = UNIX, 3 = AMIGA, 4 = MACDOS }
        arj_flags:         byte;    { (0x01 = GARBLED_FLAG) indicates passworded file
                                      (0x02 = RESERVED)
                                      (0x04 = VOLUME_FLAG)  indicates continued file to next volume
                                      (0x08 = EXTFILE_FLAG) indicates file starting position field
                                      (0x10 = PATHSYM_FLAG) indicates path translated ("\" changed to "/") }
        method:            byte;    { 0 = stored, 1 = compressed most ... 4 compressed fastest }
        file_type:         byte;    { 0 = binary, 1 = 7-bit text, 2 = comment header, 3 = directory }
        reserved:          byte;    { reserved }
        time:              longint; { date time modified }
        comp_size:         longint; { compressed size }
        orig_size:         longint; { original size (this will be different for text mode compression) }
        CRC:               longint; { original file's CRC }
        filespec_pos:      word;    { filespec position in filename }
        file_mode:         word;    { file access mode }
        host_data:         word;    { host data (currently not used) }
     end;

     {
        ?   extra data
            4 bytes for extended file position when used
            (this is present when EXTFILE_FLAG is set)

        ?   filename (null-terminated string)
        ?   comment  (null-terminated string)

        4   basic header CRC

        2   1st extended header size (0 if none)
        ?   1st extended header (not currently used)
        4   1st extended header's CRC
        ...
        ?   compressed file
     }

Const lim = 10; { file_start (4), CRC (4), ext_hdr_size (2) }

Var   ans:     PackRec;
      f:       file;
      fp:      integer;
      stop:    boolean;
      inf:     ArjInfo;
      hdr:     ArjHeader;
      name:    string;
      comment: string;

      i: integer;
      ext_hdr_size: word;
      blk: array [0..sizeof (ArjHeader)-1 + 10 + 512] of byte;

Begin
     BeginWait ('Importing ' + fname);;
     result := nil;
     try

        Init (ans, fname);
        try
           Open (f, fname);
           try

              fp := 0;
              repeat
                 Seek (f, fp);
                 BlockRead (f, inf, sizeof (inf));

                 if inf.id <> 60000 then Error (bad_header);
                 stop := (inf.header_size = 0);

                 if not stop then begin
                    if inf.header_size+lim > sizeof (blk) then Error (int_lim);
                    BlockRead (f, blk, inf.header_size+lim);

                    { header }
                    move (blk, hdr, sizeof (hdr));

                    { file name }
                    i := hdr.first_hdr_size;
                    name := PChar (@ blk[i]);
                    ModifyPath (name);
                    while blk[i] <> 0 do inc (i);
                    inc (i);

                    { comment }
                    comment := PChar (@ blk[i]);
                    while blk[i] <> 0 do inc (i);
                    inc (i);

                    if i <> inf.header_size then Error (bad_header);

                    { header CRC }
                    inc (i, 4);

                    { extended header }
                    ext_hdr_size := blk[i] + blk[i+1] * 256;
                    if ext_hdr_size <> 0 then inc (ext_hdr_size, 4); { CRC }

                    { next position }
                    inc (fp, 4 {ArjInfo} + inf.header_size +
                             4 {CRC} + 2 {size} +
                             ext_hdr_size + hdr.comp_size);


                    { store file }
                    if (hdr.file_type <> 2) and
                       (hdr.arj_flags and EXTFILE_FLAG = 0) then
                       StoreCRC (ans, name,
                                 hdr.orig_size, hdr.time, hdr.file_mode,
                                 true, hdr.CRC);
                 end;
              until stop;

           finally
              Close (f);
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

(********************************** LZH ***********************************)

Function ScanLzh (fname: string): TItem;

Const ExtFlag = $0100; { Attr and ExtFlag <> 0 => ext. header }

Type LzhHeader = packed record
        HeadSize : Byte;                {size of header}
        HeadChk  : Byte;                {checksum for header}
        HeadID   : array[1..5] of Char; {compression type tag}
        NewSize  : LongInt;             {compressed size}
        OrigSize : LongInt;             {original size}
        Time     : Word;                {packed time}
        Date     : Word;                {packed date}
        Attr     : Word;                {file attributes}
        FName    : string [255];        {filename (variable length)}
        { CRC    : Word; }              {16-bit CRC (immediately follows FName)}
     end;

     LzhExtHeader = packed record
        ExtSize: word;                  { size of ext. header }
        ExtId:   byte;                  { type of information, 2 => path }
        data:    array [1..80] of byte;
     end;

Var   f:     file;
      fp:    longint;
      hdr:   LzhHeader;
      ext:   LzhExtHeader;

      Buf:   array [0..2047] of char;
      i, BR: integer;
      found: boolean;
      len:   word;
      s:     string [80];

      ans: PackRec;
      stop: boolean;
      name: string;
      time: integer;
Begin
     BeginWait ('Importing ' + fname);;
     result := nil;
     try
        Init (ans, fname);
        try
           Open (f, fname);
           try
              BlockRead (f, Buf, sizeof (Buf), BR);
              i := 0;
              Dec (BR, 5);
              found := False;
              while (i < BR) and not found do begin
                found := (Buf[i] = '-') and (Buf[i+1] = 'l') and (Buf[i+4] = '-');
                Inc (i);
              end;

              if not found then Error (bad_header);
              fp := i-3;

              repeat
                 Seek (f, fp);
                 BlockRead (f, hdr, 1);
                 stop := (hdr.HeadSize = 0);

                 if not stop then begin
                    BlockRead (f, hdr.HeadChk, hdr.HeadSize-1);

                    if (hdr.HeadId[1] <> '-') or (hdr.HeadId[5] <> '-')
                       then Error (bad_header);

                    name := hdr.FName;
                    { CRC := PWord (@ hdr.Fname [length (hdr.Fname) + 1])^; }

                    if hdr.Attr and ExtFlag <> 0 then begin
                       { read length of first extended header }
                       BlockRead (f, ext.ExtSize, 2);

                       if (ext.ExtSize > 2) and (ext.ExtSize <= sizeof (ext)) then begin
                          { read extended header }
                          BlockRead (f, ext.ExtId, ext.ExtSize-2);

                          { read pathname }
                          if ext.ExtId = 2 then begin
                             len := ext.ExtSize-3;
                             if len < sizeof (s) then begin
                                move (ext.data, s[1], len);
                                s[0] := chr (len);
                                for i := 1 to len do
                                   if s[i] = chr (255) then s[i] := '\';
                                name := s + name;
                             end;
                          end;
                       end;
                    end;

                    inc (fp, hdr.HeadSize + 2 + hdr.NewSize);

                    time := longint (hdr.Date) shl 16 + hdr.Time;
                    Store (ans, name, hdr.OrigSize, time, hdr.Attr);
                 end; {if}
              until stop;

           finally
              Close (f);
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

(********************************** ZIP ***********************************)

Function ScanZip (fname: string): TItem;

Type LocalHeader = packed record
        {... signature ...}         {= LocalHeaderSig}
        VersionNeeded : Word;       {version needed to extract}
        BitFlag : Word;             {general purpose bit flag}
        Method : Word;              {compression method: 0-6}
        Time : Word;                {time file was last modified}
        Date : Word;                {date "}
        Crc : LongInt;              {32-bit CRC}
        NewSize : LongInt;          {compressed size of file}
        OrigSize : LongInt;         {uncompressed size of file}
        NameLength : Word;          {length of filename}
        ExtraLength : Word;         {length of extra field}
        {... filename (variable size)     ...}
        {... extra field (variable size)  ...}
        {... compressed data ...}
     end;

     CentralDirHead = packed record
        {... signature ...}         {= CentralDirHeadSig}
        VersionMade : Word;         {version created by}
        VersionNeeded : Word;       {version needed to extract}
        BitFlag : Word;             {general purpose bit flag}
        Method : Word;              {compression method: 0..6}
        Time : Word;                {time file was last modified}
        Date : Word;                {date "}
        Crc : LongInt;              {32-bit CRC}
        NewSize : LongInt;          {compressed size of file}
        OrigSize : LongInt;         {uncompressed size of file}
        NameLength : Word;          {length of filename}
        ExtraLength : Word;         {length of extra field}
        CommentLength : Word;       {length of comment for file}
        DiskNumberStart : Word;     {number of disk on which file begins}
        InternalAttrs : Word;       {internal file attributes, low bit = Text}
        ExternalAttrs : LongInt;    {external file attributes, DOS dir attr}
        LocalHeaderOfs : LongInt;   {relative offset of local header for file}
        {... filename (variable size)     ...}
        {... extra field (variable size)  ...}
        {... file comment (variable size) ...}
     end;

     CentralDirTail = packed record
        {... signature ...}         {= CentralDirTailSig}
        DiskNumber : Word;          {disk number for this disk}
        CentralDirDisk : Word;      {number of disk with start of central dir}
        EntriesThisDisk : Word;     {total entries in directory on this disk}
        TotalEntries : Word;        {total entries in central directory}
        CentralDirSize : LongInt;   {size of the central directory}
        CentralDirOfs : LongInt;    {offset of start of central dir with respect
                                     to starting disk number}
        CommentLength : Word;       {length of ZIP file comment}
        {... zip file comment (variable size) ...}
     end;

const LocalHeaderSig    = $04034B50; {signatures}
      CentralDirHeadSig = $02014B50;
      CentralDirTailSig = $06054B50;

Const IdSize    = 4;                 {SizeOf(LongInt)}
      BufSize   = 512;

Var   ans: PackRec;
      f: file;

      rem: longint;
      found: boolean;
      BlockPos, BlockSize: integer;
      BytesRead, BufPos: longint;

      stop: boolean;
      fp: longint;
      sig: longint;
      name: string;
      time: integer;

      Buf: array[1..BufSize+IdSize] of Byte;
      CDT: CentralDirTail;
      hdr: CentralDirHead;
Begin
     BeginWait ('Importing ' + fname);;
     result := nil;
     try

        Init (ans, fname);
        try
           rem := 0;
           found := false;

           Open (f, fname);
           try
              BlockPos := FileSize (f);
              BufPos := 0;

              while not found and (BlockPos > 0) do begin
                 if BlockPos >= BufSize then begin
                    BlockSize := BufSize;
                    BlockPos := BlockPos - BufSize;
                 end
                 else begin
                    BlockSize := BlockPos;
                    BlockPos := 0;
                 end;

                 { precti blok }
                 Seek (f, BlockPos);
                 BlockRead (f, Buf, BlockSize, BytesRead);

                 { pridej zbytek }
                 PInteger (@ Buf [BytesRead+1])^ := rem;

                 BufPos := BytesRead;
                 while (BufPos > 0) and
                       (PInteger (@ Buf [BufPos])^ <> CentralDirTailSig) do
                       dec (BufPos);

                 found := (BufPos > 0);

                 { uschovej zbytek }
                 rem := PInteger (@ Buf [1])^;
              end;

              if not found then Error (bad_header);
              fp := BlockPos + BufPos-1 + IdSize;

              Seek (f, fp);
              BlockRead (f, CDT, SizeOf (CDT));
              fp := CDT.CentralDirOfs;

              repeat
                 Seek (f, fp);
                 BlockRead (f, sig, 4);
                 stop := (sig = CentralDirTailSig);

                 if not stop then begin
                    if sig <> CentralDirHeadSig then Error (bad_header);

                    BlockRead (f, hdr, sizeof (hdr));
                    SetString (name, nil, hdr.NameLength);
                    BlockRead (f, name[1], hdr.NameLength);
                    ModifyPath (name);

                    inc (fp, 4 + sizeof (hdr) + hdr.NameLength +
                             hdr.ExtraLength + hdr.CommentLength);

                    time := longint (hdr.Date) shl 16 + hdr.Time;
                    StoreCRC (ans, name,
                              hdr.OrigSize, time, hdr.ExternalAttrs,
                              true, hdr.CRC);
                 end; {if}
              until stop;

           finally
              Close (f);
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

(*************************** DIRECTORY LISTING ****************************)

Function ConvOem (s: string): string;
Var  cnv, res: array [0..255] of char;
Begin
     StrLCopy (cnv, PChar (s), sizeof (cnv));
     OemToCharBuff (cnv, res, sizeof (res));
     result := res;
End;

Function LoadDir (FileName: string): TItem;
Const inv_format = 'Unknown directory format';
Var  i, start: integer;
     time, size, attr: integer;
     path: string;
     ans: PackRec;
     dos_style, nt_style: boolean;
     n: string [8];
     e: string [3];
     t: string [255];

     procedure UpdatePath;
     begin
          if start = 0 then begin
             start := length (path);
             if path [start] <> '\' then inc (start);
          end;
          path := copy (path, start+1, 255);
     end;

Begin
     BeginWait ('Importing ' + FileName);
     result := nil; { compiler }
     try

        Init (ans, FileName);
        try
           path := '';
           start := 0;
           dos_style := false;
           nt_style := false;

           OpenInp (FileName);
           try
              while not InpEof do begin
                 GetLine;

                 if copy (InpLine, 1, 12) = 'Directory of' then begin
                    if nt_style then Error (inv_format);
                    dos_style := true;

                    path := ConvOem (copy (InpLine, 14, 255));
                    UpdatePath;
                 end

                 else if copy (InpLine, 1, 14) = ' Directory of ' then begin
                    if dos_style then Error (inv_format);
                    nt_style := true;

                    path := ConvOem (copy (InpLine, 15, 255));
                    UpdatePath;
                 end

                 else if InpLine = 'Total files listed:' then
                    { nothing }

                 else if (InpLine<>'') and (InpLine[1] <> ' ') then begin

                    n := '';
                    e := '';
                    time := 0;

                    if dos_style then begin
                       { file name }
                       i := 0;
                       while (InpChar <> ' ') and (i < 8) do begin
                          n := n + InpChar;
                          inc (i);
                          GetChar;
                       end;

                       { file extension }
                       Space;
                       if InpPos <= 12 then begin
                          i := 0;
                          while (InpChar <> ' ') and (i < 3) do begin
                             e := e + InpChar;
                             inc (i);
                             GetChar;
                          end;
                       end;
                    end;

                    { file date and time}
                    if not dos_style then
                       time := GetFlexDateTime;

                    { file size }
                    t := GetCommaName;
                    if t = '<DIR>' then begin
                       attr := faDirectory;
                       size := 0;
                    end
                    else begin
                       attr := 0;
                       size := StrToInt (t);
                    end;

                    { file date and time}
                    if dos_style then
                       time := GetFlexDateTime;

                    { file name }
                    Space;
                    t := '';
                    while InpChar <> cr do begin
                       t := t + InpChar;
                       GetChar;
                    end;

                    if t = '' then begin
                       t := n;
                       if e <> '' then t := t + '.' + e;
                    end;

                    { DOS -> Windows character set }
                    t := ConvOem (t);

                    if (t <> '.') and (t <> '..') then
                       StoreSpec (ans, path, t, size, time, attr, false, 0);
                 end; {if}
              end; {while}

           finally
              CloseInp;
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

END.

