UNIT RegPro;

INTERFACE

Uses Windows, SysUtils, Classes,
     BinIO, TxtIO, ItemPro, StopDlg;

Type TRegItem = class (TItem)
        kind:  integer;  { value type }
        size:  integer;
        value: pointer;

        constructor CreateFolder (p_name: string);
        constructor CreateItem (p_name:  string;
                                p_kind:  integer;
                                p_value: pointer;
                                p_size:  integer);

        function  GetHeader: TBrwHeader; override;
        function  GetSumHeader: TBrwHeader; override;
        procedure ShowItem (var txt: TColumns); override;

        procedure Dupl (item: TItem); override;
        function  IsEqual (item: TItem): boolean; override;
        procedure GetSum (var rec: TCalcRec); override;

        procedure Load (f: TBinRead; v: integer); override;
        procedure Save (f: TBinWrite; v: integer); override;
     end;

Function ScanReg (computer: string): TRegItem;

Function  LoadReg (FileName: string): TItem; { load from .reg file }
Procedure SaveReg (FileName: string; item: TItem); { save to .reg file }

IMPLEMENTATION

(******************************** REG ITEM ********************************)

Constructor TRegItem.CreateFolder (p_name: string);
Begin
     inherited Create;

     name   := p_name;
     folder := true;
End;

Constructor TRegItem.CreateItem (p_name:  string;
                                 p_kind:  integer;
                                 p_value: pointer;
                                 p_size:  integer);
Begin
     inherited Create;
     name  := p_name;
     kind  := p_kind;
     value := p_value;
     size  := p_size;
End;

(********************************* HEADER *********************************)

Var RegHeader:    TBrwHeader = nil;
    RegSumHeader: TBrwHeader = nil;

Procedure CreateRegHeaders;
Begin
     RegHeader := TBrwHeader.Create;
     with RegHeader do begin
        Add ('Name',  100, ByName);
        Add ('Value', 200, ByValue);
     end;

     RegSumHeader := TBrwHeader.Create;
     with RegSumHeader do begin
        Add ('Name',   110, ByName);
        Add ('Values',  60, ByLeafs);
        Add ('Keys',    60, ByFolders);
        Add ('Bytes',  100, ByBytes);
        Add ('Used',   100, ByUsed);
     end;
End;

Function TRegItem.GetHeader: TBrwHeader;
Begin
     if RegHeader = nil then CreateRegHeaders;
     result := RegHeader;
End;

Function TRegItem.GetSumHeader: TBrwHeader;
Begin
     if RegSumHeader = nil then CreateRegHeaders;
     result := RegSumHeader;
End;

(******************************* SHOW ITEMS *******************************)

Procedure TRegItem.ShowItem (var txt: TColumns);
Type ByteArr = array [1..maxint] of byte;
Var  i, lim: integer;
     n: dword;
     s: string [255];
     t: ^ ByteArr;
Begin
     if value = nil then s := ''
     else if (kind = REG_SZ) or (kind = REG_EXPAND_SZ) then begin
        s := '"' + PChar (value) + '"';
     end
     else if kind = REG_DWORD then begin
        n := PDword (value)^;
        s := '0x' + IntToHex (n, 8) + ' (' + IntToStr (n) + ')';
     end
     else begin
        if size > 80 then lim := 80
                     else lim := size;
        s := '';
        t := value;
        for i := 1 to lim do begin
           s := s + IntToHex (t^[i], 2) + ' ';
        end;

        if lim < size then s := s + ' ...';
     end;

     txt[1] := name;
     txt[2] := s;
End;

(******************************** COMPARE *********************************)

Procedure TRegItem.Dupl (item: TItem);
Var  p: TRegItem;
Begin
     inherited;
     p := item as TRegItem;

     kind := p.kind;
     size := p.size;

     value := AllocMem (p.size);
     move (p.value^, value^, p.size);
End;

Function GetByte (adr: pointer; inx: integer): byte;
Begin
     inc (integer (adr), inx);
     result := PByte (adr)^;
End;

Function TRegItem.IsEqual (item: TItem): boolean;
Var  p: TRegItem;
     val1, val2: PByte;
     len1, len2, i, lim: integer;
Begin
     p := item as TRegItem;

     val1 := value;
     len1 := size;

     val2 := p.value;
     len2 := p.size;

     if kind <> p.kind then result := false
     else if (kind = REG_SZ) or (kind = REG_EXPAND_SZ) then begin

        i := 0;
        if len1 < len2 then lim := len1
                       else lim := len2;

        while (i<lim) and (val1^=val2^) and (val1^<>0) do begin
           inc (i);
           inc (val1);
           inc (val2);
        end;

        result := ((i=len1) or (val1^=0)) and ((i=len2) or (val2^=0));
     end
     else begin
        if len1 <> len2 then result := false
        else if len1 = 0 then result := true
        else result := CompareMem (val1, val2, len1);
     end;
End;

(********************************** CALC **********************************)

Procedure TRegItem.GetSum (var rec: TCalcRec);
Begin
     if folder then begin
        inc (rec.reg_keys);
        inc (rec.reg_used, length (name));
     end
     else begin
        inc (rec.reg_values);
        inc (rec.reg_bytes, size);
        inc (rec.reg_used, length (name) + size);
     end;
End;

(****************************** LOAD / SAVE *******************************)

Procedure TRegItem.Load (f: TBinRead; v: integer);
Begin
     inherited;
     f.ReadInt (kind);
     f.ReadBlock (value, size);
End;

Procedure TRegItem.Save (f: TBinWrite; v: integer);
Begin
     inherited;
     f.WriteInt (kind);
     f.WriteBlock (value, size);
End;

(***************************** SCAN REGISTRY ******************************)

Type ERegistry = class (Exception);

Type InfoRec = record
        NumSubKeys:   longword;
        MaxSubKeyLen: longword;
        MaxClassLen:  longword;
        NumValues:    longword;
        MaxValueLen:  longword;
        MaxDataLen:   longword;
        MaxSecLen:    longword;
        FileTime:     TFileTime;
     end;

Procedure ErrorCode (ans: longword); 
Begin
     raise ERegistry.Create ('Registry error ' +
                             IntToStr (ans) + ': ' +
                             SysErrorMessage (ans));
End;

Procedure ScanRegBranch (top: TRegItem; key: HKey; path, alias: string);

Var  item, tmp: TRegItem;
     ans: longword;
     inx: integer;
     name_buf: PChar;
     data_buf, new_buf, p: PByte;
     name_len, data_len, data_typ: longword;
     Info: InfoRec;

Begin
     if alias <> '' then StopInfo (alias)
                    else StopTest;
     if StopFlag then exit;
     item := nil;

     { open key }
     ans := RegOpenKeyEx
            (key,	    // handle of open key
             PChar (path),  // address of name of subkey to open
             0,	            // reserved
             KEY_READ,	    // security access mask
             key); 	    // address of handle of open key

     if (ans = 0) or (ans = ERROR_ACCESS_DENIED) then begin
        { %% je mozne ze klic je nepristupny, ale existuje }
        if path = '' then item := TRegItem.CreateFolder (alias)
                     else item := TRegItem.CreateFolder (path);
        top.Add (item);
     end;

     if (ans <> 0) and
        (ans <> ERROR_ACCESS_DENIED) and
        (ans <> ERROR_FILE_NOT_FOUND)
        then ErrorCode (ans);
        { %% chyba ERROR_FILE_NOT_FOUND ve vetvi HKLM\System\Clone }

     if ans = 0 then
        try
           { query info }
           FillChar (Info, sizeof (Info), 0);

           with Info do
              ans := RegQueryInfoKey
                     (key,             // handle of key to query
                      nil,             // buffer for class string
                      nil,             // size of class string buffer
                      nil,             // reserved
                      @ NumSubKeys,    // number of subkeys
                      @ MaxSubKeyLen,  // longest subkey name length
                      @ MaxClassLen,   // longest class string length
                      @ NumValues,     // number of value entries
                      @ MaxValueLen,   // longest value name length
                      @ MaxDataLen,    // longest value data length
                      @ MaxSecLen,     // security descriptor length
                      @ FileTime);     // last write time

           if ans = ERROR_INSUFFICIENT_BUFFER then ans := 0;
           { %% problem ve vetvi
             HKLM\Software\Microsoft\Windows NT\CurrentVersion\Perflib\009 }

           if (ans <> 0) and (ans <> ERROR_FILE_NOT_FOUND) then ErrorCode (ans);
           { %% problem ve vetvi
             HKLM\Software\Microsoft\Windows NT\CurrentVersion\Perflib\005 }

           if ans = 0 then begin

              { %% correct zero length }
              if Info.MaxSubKeyLen = 0 then Info.MaxSubKeyLen := 255;

              { %% correct small length }
              if Info.MaxDataLen < 255 then Info.MaxDataLen := 255;

              { values }
              if Info.NumValues <> 0 then begin
                 name_buf := nil;
                 data_buf := nil;
                 try
                    GetMem (name_buf, Info.MaxValueLen+1);
                    GetMem (data_buf, Info.MaxDataLen+1);

                    for inx := 0 to Info.NumValues-1 do begin
                       name_len := Info.MaxValueLen+1;
                       data_len := Info.MaxDataLen+1;

                       ans := RegEnumValue
                              (key,	    // handle of key to query
                               inx,	    // index of value to query
                               name_buf,    // address of buffer for value string
                               name_len,    // address for size of value buffer
                               nil,	    // reserved
                               @ data_typ,  // address of buffer for type code
                               data_buf,    // address of buffer for value data
                               @ data_len); // address for size of data buffer

                       if ans <> 0 then ErrorCode (ans);

                       { %% correct length }
                       if { (data_len = 0) and }
                          ((data_typ = REG_SZ) or (data_typ = REG_EXPAND_SZ))
                       then begin
                          p := data_buf;
                          while (data_len <= Info.MaxValueLen) and (p^ <> 0) do begin
                             inc (data_len);
                             inc (p);
                          end;
                       end;

                       new_buf := AllocMem (data_len);
                       move (data_buf^, new_buf^, data_len);

                       tmp := TRegItem.CreateItem (name_buf,
                                                   data_typ, new_buf, data_len);
                       item.Add (tmp);

                    end; {for}
                 finally
                    FreeMem (name_buf);
                    FreeMem (data_buf);
                 end;
              end; {if}

              { sub keys }
              if Info.NumSubKeys <> 0 then begin
                 GetMem (name_buf, Info.MaxSubKeyLen+1);
                 try
                    for inx := 0 to Info.NumSubKeys-1 do begin
                       name_len := Info.MaxSubKeyLen+1;

                       ans := RegEnumKeyEx
                              (key,  	         // handle of key to enumerate
                               inx,	         // index of subkey to enumerate
                               PChar (name_buf), // address of buffer for subkey name
                               name_len,         // address for size of subkey buffer
                               nil, 	         // reserved
                               nil,	         // address of buffer for class string
                               nil,	         // address for size of class buffer
                               nil);             // address for time key last written to

                       if ans <> 0 then ErrorCode (ans);
                       if ans = 0 then ScanRegBranch (item, key, name_buf,
                                                      alias + '\'+ name_buf);
                    end; {for}
                 finally
                    FreeMem (name_buf);
                 end;
              end; {if}
           end; {if}

        finally
           { close key }
           ans := RegCloseKey (key);
           if ans <> 0 then ErrorCode (ans);
        end;
End;

Procedure ScanRegTree (top: TRegItem; computer: string; key: HKey; alias: string);
Var  ans: longword;
Begin
     if computer <> '' then begin
        ans := RegConnectRegistry (PChar (computer), key, key);
        if ans <> ERROR_SUCCESS then ErrorCOde (ans);
        alias := computer + '\' + alias;
     end;

     ScanRegBranch (top, key, '', alias);
End;

Function ScanReg (computer: string): TRegItem;
Begin
     result := TRegItem.CreateFolder ('Registry');
     StopOpen ('Scanning registry');

     try
        try

           ScanRegTree (result, computer, HKEY_LOCAL_MACHINE, 'HKEY_LOCAL_MACHINE');
           ScanRegTree (result, computer, HKEY_USERS, 'HKEY_USERS');

        finally
           StopClose;
        end;

     except
        result.Free;
        raise;
     end;

     if StopFlag then begin
        result.Free;
        result := nil;
     end;
End;

(*************************** LOAD REGISTRY FILE ***************************)

Type StoreRec = record
        top: TItem;
        cur: TItem;
        txt: string;
     end;

Procedure ErrorMsg (msg: string);
Begin
     raise ERegistry.Create (msg);
End;

Procedure Init (var p: StoreRec; fname: string);
Begin
     fname := ExtractFileName (fname);
     p.top := TRegItem.CreateFolder (fname);
     p.cur := p.top;
     p.txt := '';
End;

Procedure Clear (var p: StoreRec);
Begin
     p.top.Free;
     p.top := nil;
     p.cur := nil;
End;

Procedure Lookup (var p: StoreRec; path: string);
Var  i, k, len: integer;
     eq: boolean;
     tmp: TItem;
     part: string;
Begin
     len := length (p.txt);
     eq := (copy (path, 1, len) = p.txt) and
           (length (path) > len+1) and
           (path [len+1] = '\');

     if eq then
        i := len+2
     else begin
        p.cur := p.top;
        i := 1;
     end;

     len := length (path);
     while i <= len do begin
        { oddel jmeno directory }
        k := i;
        while (k<=len) and (path[k] <> '\') do inc (k);
        if k=i then ErrorMsg ('Empty name');
        part := copy (path, i, k-i);
        if k <= len then i := k+1
                    else i := k;

        { hledej directory }
        tmp := p.cur.sub.first;
        while (tmp <> nil) and ((tmp.name <> part) or not tmp.folder) do
           tmp := tmp.next;
           { v registry je mozne pouzit stejne jmeno pro 'key' i 'value' }

        if tmp = nil then begin
           { directory nebylo nalezeno }
           tmp := TRegItem.CreateFolder (part);
           p.cur.Add (tmp);
        end;
        p.cur := tmp;
     end;

     p.txt := path;
End;

{--------------------------------------------------------------------------}

Const quote = '"'; { double quote }

Function GetRegKey: string;
Const max = 255;
Var   len, level: integer;
      s: string [max];
      txt: string;
Begin
     txt := '';
     len := 0;
     if InpChar <> '[' then ErrorMsg ('[ expected');
     GetChar; { skip [ }

     level := 0;
     while (InpChar <> ']') or (level > 0) do begin
        if InpChar = cr then ErrorMsg ('Registry key exceeds line');
        if len = max then begin
           s[0] := chr (len);
           txt := txt + s;
           len := 0;
        end;
        inc (len);
        s[len] := InpChar;
        if InpChar = '[' then inc (level);
        if InpChar = ']' then dec (level);
        GetChar;
     end;
     GetChar; { skip ] }

     s[0] := chr (len);
     result := txt + s;
End;

Function GetRegName: string;
Const max = 255;
Var   len: integer;
      s:   string [max];
Begin
     len := 0;
     if InpChar <> quote then ErrorMsg ('Quote expected');
     GetChar; { skip quote }

     while InpChar <> quote do begin

        if InpChar = '\' then begin
           GetChar;
           if InpChar = cr then ErrorMsg ('String exceeds line');
           if not (InpChar in [quote, '\']) then
              ErrorMsg ('Strange character');
        end;

        if len < max then begin
           inc (len);
           s[len] := InpChar;
        end
        else ErrorMsg ('String too long');

        if InpChar = cr then GetLine
                        else GetChar;
     end;

     GetChar; { skip ] }

     s[0] := chr (len);
     result := s;
End;

Procedure GetRegStr (var value: pointer; var size: integer);
Const max = 4*1024;
Var   len: integer;
      s:   array [1..max] of char;
Begin
     len := 0;
     if InpChar <> quote then ErrorMsg ('Quote expected');
     GetChar; { skip quote }

     while InpChar <> quote do begin

        if InpChar = '\' then begin
           GetChar;
           if InpChar = cr then ErrorMsg ('String exceeds line');
           if not (InpChar in [quote, '\']) then
              ErrorMsg ('Strange character');
        end;

        if len < max then begin
           inc (len);
           s[len] := InpChar;
        end
        else ErrorMsg ('String too long');

        if InpChar = cr then GetLine
                        else GetChar;
     end;

     GetChar; { skip ] }

     if len+1 > max then ErrorMsg ('String too long');
     s[len+1] := chr (0);

     size := len+1;
     value := AllocMem (size);
     move (s[1], value^, size);
End;

Function GetRegToken: string;
Const max = 255;
      lod = ['A'..'Z', 'a'..'z', '0'..'9'];
Var   len: integer;
      s: string [max];
Begin
     Space;
     if not (InpChar in lod) then ErrorMsg ('Token expected');
     len := 0;

     while InpChar in lod do begin
        inc (len);
        if len > max then ErrorMsg ('Token too long');
        s[len] := InpChar;
        GetChar;
     end;

     s[0] := chr (len);
     result := s;
End;

Procedure GetRegDword (var value: pointer; var size: integer);
Var n: DWord;
Begin
     n := DWord (StrToInt ('$'+GetRegToken));
     size := 4;
     value := AllocMem (size);
     PDWord (value)^ := n;
End;

Procedure GetRegBinary (var value: pointer; var size: integer);
Const max = 16*1024;
Var   n, i: integer;
      c: char;
      cont: boolean;
      len: integer;
      s: array [1..max] of byte;
Begin
     Space;

     if InpChar = cr then begin
        value := nil;
        size := 0;
     end
     else begin
        len := 0;
        cont := true;

        while cont do begin
           Space;

           if InpChar <> ',' then begin
              n := 0;
              for i := 1 to 2 do begin
                 n := n shl 4;
                 c := UpCase (InpChar);
                 if c in ['0'..'9'] then n := n + ord (c) - ord ('0')
                 else if c in ['A'..'F'] then n := n + ord (c) - ord ('A') + 10
                 else ErrorMsg ('Hex digit expected');
                 GetChar;
              end;

              if len < max then begin
                inc (len);
                s[len] := n;
              end
              else ErrorMsg ('Data too long');
           end;

           if InpChar = ',' then begin
              GetChar;
              if InpChar = '\' then GetLine;
           end
           else cont := false;
        end;

        size := len;
        value := AllocMem (size);
        move (s[1], value^, size);
     end;
End;

Procedure Check (c: char);
Begin
     Space;
     if InpChar <> c then ErrorMsg (c + ' expected');
     GetChar;
End;

Function GetRegValue: TRegItem;
Var  name, tag: string;
     kind: integer;
     value: pointer;
     size: integer;
Begin
     if InpChar = '@' then begin
        GetChar;
        name := '';
     end
     else name := GetRegName;

     Check ('=');

     Space;
     kind := 0;
     if InpChar = quote then begin
        kind := REG_SZ;
        GetRegStr (value, size);
     end
     else begin
        tag := GetRegToken;
        if tag = 'dword' then begin
           Check (':');
           kind := REG_DWORD;
           GetRegDword (value, size);
        end
        else if tag = 'hex' then begin
           kind := REG_BINARY;
           Space;
           if InpChar = '(' then begin
              GetChar;
              kind := StrToInt ('$'+GetRegToken);
              Check (')');
           end;
           Check (':');
           GetRegBinary (value, size);
        end
        else ErrorMsg ('Unknown value');
     end;

     CheckEol;
     result := TRegItem.CreateItem (name, kind, value, size);
End;

Function LoadReg (FileName: string): TItem;
Var  ans: StoreRec;
     key: string;
     item: TRegItem;
Begin
     BeginWait ('Importing ' + FileName);;
     result := nil;
     try

        Init (ans, FileName);
        try
           OpenInp (FileName);
           try

              CheckLine ('REGEDIT4');
              GetLine;

              while not InpEof do begin
                 key := GetRegKey;
                 Lookup (ans, key);
                 GetLine;

                 while InpLine <> '' do begin
                    item := GetRegValue;
                    ans.cur.Add (item);
                    GetLine;
                 end;

                 GetLine;
              end; {while}

           finally
              CloseInp;
           end;

           result := ans.top;

        except
           Clear (ans);
           raise;
        end;

     finally
        EndWait;
     end;
End;

(*************************** SAVE REGISTRY FILE ***************************)

Var OutFile: text;
    OutBuff: array [1..16*1024] of byte;

Const OutMax = 255;

Var OutTab: string [OutMax];
    OutPos: integer; { pocet znaku v OutTab }
    OutCol: integer; { cislo sloupce }

{--------------------------------------------------------------------------}

Procedure OpenOut (FileName: string);
Begin
     Assign (OutFile, FileName);
     SetTextBuf (OutFile, OutBuff);
     Rewrite (OutFile);

     OutPos := 0;
     OutCol := 0;
End;

Procedure SendTab;
Begin
     Assert ((OutPos >= 0) and (OutPos <= OutMax));

     if OutPos > 0 then begin
        OutTab [0] := chr (OutPos);
        Write (OutFile, OutTab);
     end;
     OutPos := 0;
End;

Procedure CloseOut;
Begin
     SendTab;
     Close (OutFile);
End;

{--------------------------------------------------------------------------}

Procedure PutChr (ch: char);
Begin
     if OutPos+1 >= OutMax then SendTab;

     inc (OutPos);
     OutTab [OutPos] := ch;

     inc (OutCol);
End;

Procedure PutQuoChr (ch: char);
Begin
     if ch in [quote, '\'] then PutChr ('\');
     PutChr (ch);
End;

Procedure PutStr (s: string);
Var  i: integer;
Begin
     for i := 1 to length (s) do
        PutChr (s[i]);
End;

Procedure PutQuoStr (s: string);
Var  i: integer;
Begin
     PutChr (quote);
     for i := 1 to length (s) do
        PutQuoChr (s[i]);
     PutChr (quote);
End;

Procedure PutHex (num: LongWord; digits: integer);
Const hex: array [0..15] of char = '0123456789abcdef';
Var  i, n, shift: integer;
Begin
     for i := 1 to digits do begin
        shift := (digits - i) * 4;
        n := (num shr shift) and 15;
        PutChr (hex[n]);
     end;
End;

Procedure PutEol;
Begin
     PutChr (cr);
     PutChr (lf);
     OutCol := 0;
End;

{--------------------------------------------------------------------------}

Procedure SaveValue (item: TRegItem);
Var  kind, size, i: integer;
     ref: PChar;
     c: char;
     d: DWord;
     p: PByte;
Begin
     if item.name = '' then PutChr ('@')
                       else PutQuoStr (item.name);

     PutChr ('=');

     kind := item.kind;
     size := item.size;

     if kind = REG_SZ then begin
        PutChr (quote);
        ref := item.value;
        for i := 0 to size-1 do begin
           c := ref[i];
           if c = chr (0) then break;
           PutQuoChr (c);
        end;
        PutChr (quote);
     end

     else if kind = REG_DWORD then begin
        PutStr ('dword:');
        d := PDWord (item.value)^;
        PutHex (d, 8);
     end

     else begin

        PutStr ('hex');
        if kind <> REG_BINARY then begin
           PutChr ('(');
           PutHex (kind, 1);
           PutChr (')');
        end;
        PutChr (':');

        p := item.value;
        for i := 1 to size do begin
           PutHex (p^, 2);
           if i < size then begin
              PutChr (',');
              if OutCol >= 77 then begin
                 PutChr ('\');
                 PutEol;
                 PutChr (' ');
                 PutChr (' ');
              end;
           end;
           inc (p);
        end;
     end;

     PutEol;
End;

Procedure SaveBranch (branch: TRegItem; path: string);
Var  item: TItem;
Begin
     if path <> '' then path := path + '\';
     path := path + branch.name;

     if path <> '' then begin {!?}
        PutStr ('[' + path + ']');
        PutEol;
     end;

     item := branch.sub.first;
     while item <> nil do begin
        if not item.folder then
           SaveValue (item as TRegItem);
        item := item.next;
     end;
               if path <> '' then {!?}

        PutEol; { empty line }

     item := branch.sub.first;
     while item <> nil do begin
        if item.folder then
           SaveBranch (item as TRegItem, path);
        item := item.next;
     end;
End;

Procedure SaveReg (FileName: string; item: TItem);
Begin
     if not (item is TRegItem) then
        ErrorMsg ('Registry branch required');

     if not item.folder then
        ErrorMsg ('Registry folder required');

     BeginWait ('Exporting ' + FileName);;
     try

        OpenOut (FileName);

        try
           PutStr ('REGEDIT4');
           PutEol;
           PutEol; { empty line }
           SaveBranch (item as TRegItem, '');
        finally
           CloseOut;
        end;

     finally
        EndWait;
     end;
End;

END.

