UNIT SortPro;

INTERFACE

Uses SysUtils,
     ItemPro, FilePro, TextPro;

(***************************** COMPARE ITEMS ******************************)

Function CompByName   (item1, item2: TItem): integer;
{
Function CompByExt  (item1, item2: TItem): integer;
Function CompBySize (item1, item2: TItem): integer;
Function CompByTime (item1, item2: TItem): integer;
Function CompByValue (item1, item2: TItem): integer;
Function CompByColor  (item1, item2: TItem): integer;
Function CompUnsorted (item1, item2: TItem): integer;
}

Function CompNameSize (item1, item2: TItem): integer;

(******************************* HEAP SORT ********************************)

Type TableArr = array of TItem;
     FiltFce  = function (item: TItem): boolean of object;
     CompFce  = function (item1, item2: TItem): integer;

Type TAcceptAll = class
     public
        function Filter (item: TItem): boolean;
     end;

Var  AcceptAll: TAcceptAll;

Function SortFunc (sort: TSort): CompFce;

Procedure AllocTable (item: TItem;
                      filt: FiltFce;
                      var tab: TableArr);

Procedure SortTable (comp: CompFce;
                     tab: TableArr);

Procedure SortFolder (folder: TItem;
                      comp: CompFce;
                      subdir: boolean);

Procedure FreeTable (var tab: TableArr);

IMPLEMENTATION

(***************************** COMPARE ITEMS ******************************)

Function CompByName (item1, item2: TItem): integer;
Begin
     result := ord (item2.folder) - ord (item1.folder);
     if result=0 then result := CompareText (item1.name, item2.name);
End;

Function CompByExt (item1, item2: TItem): integer;
Var  a, b: TFileItem;
Begin
     a := item1 as TFileItem;
     b := item2 as TFileItem;

     result := ord (b.folder) - ord (a.folder);
     if result=0 then result := CompareText (a.ext, b.ext);
     if result=0 then result := CompareText (a.name, b.name);
End;

Function CompBySize (item1, item2: TItem): integer;
Var  a, b: TFileItem;
Begin
     a := item1 as TFileItem;
     b := item2 as TFileItem;

     result := ord (b.folder) - ord (a.folder);
     if result=0 then result := b.size - a.size;
     if result=0 then result := CompareText (a.name, b.name);
End;

Function CompByTime (item1, item2: TItem): integer;
Var  a, b: TFileItem;
Begin
     a := item1 as TFileItem;
     b := item2 as TFileItem;

     result := ord (b.folder) - ord (a.folder);
     if result=0 then result := b.time - a.time;
     if result=0 then result := CompareText (a.name, b.name);
End;

Function CompByValue (item1, item2: TItem): integer;
Var  a, b: TTextItem;
Begin
     a := item1 as TTextItem;
     b := item2 as TTextItem;

     result := ord (b.folder) - ord (a.folder);
     if result=0 then result := CompareText (a.value, b.value);
     if result=0 then result := CompareText (a.name, b.name);
End;

Function CompByColor (item1, item2: TItem): integer;
Begin
     result := ord (item2.folder) - ord (item1.folder);
     if result=0 then result := ord (item2.mode) - ord (item1.mode);
     if result=0 then result := CompareText (item1.name, item2.name);
End;

Function CompUnsorted (item1, item2: TItem): integer;
Begin
     result := ord (item2.folder) - ord (item1.folder);
     if result=0 then result := item1.index - item2.index;
End;

{--------------------------------------------------------------------------}

Function CompByLeafs (item1, item2: TItem): integer;
Begin
     result := ord (item2.folder) - ord (item1.folder);
     if result=0 then result := item2.sum_leafs - item1.sum_leafs;
     if result=0 then result := CompareText (item1.name, item2.name);
End;

Function CompByFolders (item1, item2: TItem): integer;
Begin
     result := ord (item2.folder) - ord (item1.folder);
     if result=0 then result := item2.sum_folders - item1.sum_folders;
     if result=0 then result := CompareText (item1.name, item2.name);
End;

Function CompByBytes (item1, item2: TItem): integer;
Begin
     result := ord (item2.folder) - ord (item1.folder);
     if result=0 then result := item2.sum_bytes - item1.sum_bytes;
     if result=0 then result := CompareText (item1.name, item2.name);
End;

Function CompByUsed (item1, item2: TItem): integer;
Begin
     result := ord (item2.folder) - ord (item1.folder);
     if result=0 then result := item2.sum_used - item1.sum_used;
     if result=0 then result := CompareText (item1.name, item2.name);
End;

{--------------------------------------------------------------------------}

Function CompNameSize (item1, item2: TItem): integer;
Var  a, b: TFileItem;
Begin
     result := ord (item2.folder) - ord (item1.folder);
     if result=0 then result := CompareText (item1.name, item2.name);

     if (result=0) and (item1 is TFileItem) and (item2 is TFileItem) then begin
        a := item1 as TFileItem;
        b := item2 as TFileItem;
        result := b.size - a.size;
     end;
End;

(******************************* SORT FUNC ********************************)

Function SortFunc (sort: TSort): CompFce;
Begin
     result := CompUnsorted;
     case sort of
        ByName:   result := CompByName;
        ByExt:    result := CompByExt;
        BySize:   result := CompBySize;
        ByTime:   result := CompByTime;
        ByValue:  result := CompByValue;
        ByColor:  result := CompByColor;
        Unsorted: result := CompUnsorted;

        ByLeafs:   result := CompByLeafs;
        ByFolders: result := CompByFolders;
        ByBytes:   result := CompByBytes;
        ByUsed:    result := CompByUsed;
     end;
End;

(******************************* HEAP SORT ********************************)

Procedure Heapify (comp: CompFce; tab: TableArr; i, j: integer);
Label 1, 9;
Var   k, s, m: integer;
      t: TItem;
Begin
     k:=i;
   1:
     s := 2*k+1;
     if s>j then goto 9;

     m := s;
     if s+1<=j then
        if comp (tab[s], tab[s+1]) < 0 then
           m := s+1;

     if comp (tab[k], tab[m]) < 0 then begin
        t := tab[k];
        tab[k] := tab[m];
        tab[m] := t;

        k := m;
        goto 1;
     end;
   9:
End;

Procedure SortTable (comp: CompFce; tab: TableArr);
Var  i, cnt: integer;
     t: TItem;
Begin
     cnt := length (tab);

     for i := (cnt-2) div 2 downto 0 do
        Heapify (comp, tab, i, cnt-1);

     for i := cnt-1 downto 1 do begin
        t := tab[0];
        tab[0] := tab[i];
        tab[i] := t;
        Heapify (comp, tab, 0, i-1);
     end;
End;

(************************* ALLOCATE / FREE TABLE **************************)

Procedure AllocTable (item: TItem;
                      filt: FiltFce;
                      var tab: TableArr);
Var  p: TItem;
     n, i: integer;
     table: TableArr;
Begin
     { count items }
     p := item.sub.first;
     n := 0;
     while p <> nil do begin
        if filt (p) then inc (n);
        p := p.next;
     end;

     { allocate table }
     SetLength (table, n);

     { store pointers to table }
     p := item.sub.first;
     i := 0;
     while p <> nil do begin
        if filt (p) then begin
           table[i] := p;
           inc (i);
        end;
        p := p.next;
     end;

     tab := table;
End;

Procedure FreeTable (var tab: TableArr);
Begin
     tab := nil;
End;

Function TAcceptAll.Filter (item: TItem): boolean;
Begin
     result := true;
End;

(****************************** SORT FOLDER *******************************)

Procedure SortFolder (folder: TItem;
                      comp: CompFce;
                      subdir: boolean);
Var  tab: TableArr;
     cnt, i: integer;
     p, back, forw: TItem;
Begin
     { allocate }
     AllocTable (folder, AcceptAll.Filter, tab);

     { sort }
     SortTable (comp, tab);

     cnt := length (tab);
     if cnt <> 0 then begin
        { link sorted elements }
        folder.sub.first := tab [0];
        folder.sub.last := tab [cnt-1];

        back := nil;
        p := tab [0];
        for i := 0 to cnt-1 do begin

           if i=cnt-1 then forw := nil
                      else forw := tab [i+1];

           p.prev := back;
           p.next := forw;

           back := p;
           p := forw;
        end;

        { free table }
        FreeTable (tab);

        { sort next level }
        if subdir then begin
           p := folder.sub.first;
           while p <> nil do begin
              if p.folder then SortFolder (p, comp, true);
              p := p.next;
           end;
        end;
     end;
End;

BEGIN
     AcceptAll := TAcceptAll.Create;
END.

