UNIT TotalDlg;

INTERFACE

Uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, ComCtrls, ToolWin,
  ItemPro, FiltDlg;

Type
  TTotalDialog = class (TFiltDialog)
    PageControl: TPageControl;

    FileSheet: TTabSheet;
    RegistrySheet: TTabSheet;
    TotalSheet: TTabSheet;

    { directories }
    LabDirs: TLabel;
    LabFiles: TLabel;
    LabFileBytes: TLabel;
    LabFileUsed: TLabel;

    NumDirs: TEdit;
    NumFiles: TEdit;
    NumFileBytes: TEdit;
    NumFileUsed: TEdit;

    { registry }
    LabKeys: TLabel;
    LabValues: TLabel;
    LabRegSize: TLabel;
    LabRegUsed: TLabel;

    NumKeys: TEdit;
    NumValues: TEdit;
    NumRegBytes: TEdit;
    NumRegUsed: TEdit;

    { total }
    LabFolders: TLabel;
    LabNodes: TLabel;
    LabBytes: TLabel;
    LabUsed: TLabel;

    NumFolders: TEdit;
    NumNodes: TEdit;
    NumBytes: TEdit;
    NumUsed: TEdit;

    TopPanel: TPanel;
    BotomPanel: TPanel;
    CloseButton: TButton;
    { checkpoint }

    Procedure CloseButtonClick (Sender: TObject);
    { checkpoint }

  protected
    procedure Recalc; override;
    procedure ChoosePage; override;

  private
    rec: TCalcRec;
  end;

Var
  TotalDialog: TTotalDialog;

IMPLEMENTATION

Uses Brw; { Browser }

{$R *.DFM}

{--------------------------------------------------------------------------}

Procedure CalcFolder (item: TItem;
                      var rec: TCalcRec;
                      modes: TModeSet;
                      states: TStateSet);
Var  p: TItem;
Begin
     p := item.sub.first;
     while p <> nil do begin
        if not p.folder and (p.mode in modes) and (p.state in states) then begin
           inc (rec.all_nodes);
           p.GetSum (rec);
        end;
        p := p.next;
     end;
End;

Procedure CalcSubFolders (item: TItem;
                          var rec: TCalcRec;
                          modes: TModeSet;
                          states: TStateSet);
Var  p: TItem;
Begin
     p := item.sub.first;
     while p <> nil do begin
        if p.folder then begin
           inc (rec.all_folders);
           p.GetSum (rec);
           CalcFolder (p, rec, modes, states);
           CalcSubFolders (p, rec, modes, states);
        end;
        p := p.next;
     end;
End;

{--------------------------------------------------------------------------}

Function Conv (n: integer): string;
Var  s: string;
Begin
     str (n:10, s);
     result := ' ' + copy (s, 1, 1) +
               ' ' + copy (s, 2, 3) +
               ' ' + copy (s, 5, 3) +
               ' ' + copy (s, 8, 3);
End;

Procedure TTotalDialog.Recalc;
Var  item: TItem;
Begin
     FillChar (rec, sizeof (rec), 0);
     item := Browser.view.CurrentTreeItem;
     if item <> nil then begin
        if ThisDir then CalcFolder (item, rec, modes, states);
        if SubDir  then CalcSubFolders (item, rec, modes, states);
     end;

     rec.all_bytes := rec.fil_bytes + rec.reg_bytes;
     rec.all_used  := rec.fil_used + rec.reg_used;

     NumDirs.Text      := Conv (rec.fil_dirs);
     NumFiles.Text     := Conv (rec.fil_files);
     NumFileBytes.Text := Conv (rec.fil_bytes);
     NumFileUsed.Text  := Conv (rec.fil_used);

     NumKeys.Text      := Conv (rec.reg_keys);
     NumValues.Text    := Conv (rec.reg_values);
     NumRegBytes.Text  := Conv (rec.reg_bytes);
     NumRegUsed.Text   := Conv (rec.reg_used);

     NumFolders.Text   := Conv (rec.all_folders);
     NumNodes.Text     := Conv (rec.all_nodes);
     NumBytes.Text     := Conv (rec.all_bytes);
     NumUsed.Text      := Conv (rec.all_used);
End;

Procedure TTotalDialog.ChoosePage;
Var  no_fil, no_reg: boolean;
Begin
     { pripadne prepni stranky }
     no_fil := (rec.fil_dirs = 0) and (rec.fil_files = 0);
     no_reg := (rec.reg_keys = 0) and (rec.reg_values = 0);

     if no_fil and no_reg then
        PageControl.ActivePage := TotalSheet
     else begin
        if (PageControl.ActivePage = FileSheet) and no_fil then
           PageControl.ActivePage := RegistrySheet;

        if (PageControl.ActivePage = RegistrySheet) and no_reg then
           PageControl.ActivePage := FileSheet;
     end;
End;

Procedure TTotalDialog.CloseButtonClick (Sender: TObject);
Begin
     Close;
End;

END.
