UNIT VerFrm;

INTERFACE

Uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, Grids, Menus,
  ItemPro, VerPro, FilePro, SortPro, StopDlg;

Type
  TVerForm = class (TForm)
    Grid: TStringGrid;

    Menu: TMainMenu;
    MenuTable: TMenuItem;
    MenuSortByFileName: TMenuItem;
    MenuSortByProduct: TMenuItem;
    MenuSep1: TMenuItem;
    MenuClose: TMenuItem;

    procedure FormKeyDown (Sender: TObject;
                           var Key: Word;
                           Shift: TShiftState);

    procedure MenuSortByFileNameClick (Sender: TObject);
    procedure MenuSortByProductClick (Sender: TObject);
    procedure MenuCloseClick (Sender: TObject);

  private
    list: TList; { seznam souboru }

    procedure SetField (line, col: integer; txt: string);
    procedure Redraw;

  public
    procedure Init;
    procedure Add (item: TFileItem);
    procedure Run;
  end;

Var
  VerForm: TVerForm;

Procedure VersionTable (item: TFileItem);

IMPLEMENTATION

{$R *.DFM}

(******************************** COLUMNS *********************************)

Const tab_max = ord (high (TVerAttr)) - ord (low(TVerAttr)) + 1;
Type  TTabArr = array [1..tab_max] of TVerAttr;

Const tab: TTabArr =
      (siCompanyName,

       siProductName,
       siProductVersion,
       fiProductVersion,

       siFileDescription,
       siInternalName,
       siOriginalFilename,
       siFileVersion,
       fiFileVersion,

       siPrivateBuild,
       fiPrivateBuild,
       siSpecialBuild,
       fiSpecialBuild,

       fiDebug,
       fiPatched,
       fiPrerelease,
       fiInfoInferred,

       fiOperatingSystem,
       fiFileType,
       fiFileSubtype,
       fiFileDate,

       siComments,
       siLegalCopyright,
       siLegalTrademarks);

(****************************** REDRAW TABLE ******************************)

Procedure TVerForm.SetField (line, col: integer; txt: string);
Var  wid: integer;
Begin
     Grid.Cells [col, line] := txt;
     wid := Grid.Canvas.TextWidth (txt + '  ');
     if wid > Grid.ColWidths [col] then Grid.ColWidths [col] := wid;
End;

Procedure TVerForm.Redraw;
Var  cnt, line, col, inx: integer;
     item: TFileItem;
     ver: TVerInfo;
     atr: TVerAttr;
     used: array [TVerAttr] of boolean;
Begin
     for atr := low (TVerAttr) to high (TVerAttr) do
        used [atr] := false;

     for line := 1 to list.Count do begin
        item := TFileItem (list [line-1]);
        ver := item.ver;
        for atr := low (TVerAttr) to high (TVerAttr) do
           if ver.value [atr] <> '' then used [atr] := true;
     end;

     cnt := 0;
     for atr := low (TVerAttr) to high (TVerAttr) do
        if used [atr] then inc (cnt);

     Grid.ColCount := cnt+1;
     Grid.RowCount := list.Count+1;
     if Grid.RowCount = 1 then begin
        Grid.RowCount := 2;
        Grid.Rows[1].Clear;
        { nemenit - slouzi k vytvoreni jedne prazdne radky }
     end;
     Grid.FixedRows := 1;

     SetField (0, 0, 'File Name');
     col := 0;
     for inx := 1 to tab_max do begin
        atr := tab [inx];
        if used [atr] then begin
           inc (col);
           SetField (0, col, AttrName [atr]);
        end;
     end;

     for line := 1 to list.Count do begin
        item := TFileItem (list [line-1]);
        ver := item.ver;

        SetField (line, 0, item.name);
        col := 0;
        for inx := 1 to tab_max do begin
           atr := tab [inx];
           if used [atr] then begin
              inc (col);
              SetField (line, col, ver.value [atr]);
           end;
        end;
     end;
End;

(******************************** ADD DATA ********************************)

Procedure TVerForm.Init;
Begin
     list.Free;
     list := TList.Create;
End;

Procedure TVerForm.Add (item: TFileItem);
Begin
     if item.ver <> nil then
        list.Add (item);
End;

Procedure NoVerInfo;
Begin
     MessageDlg ('No Version Information', mtInformation, [mbOK], 0);
End;

Procedure TVerForm.Run;
Begin
     if list.Count > 0 then begin
        Redraw;
        WindowState := wsNormal;
        Show;
     end
     else NoVerInfo;
End;

Procedure VersionTable (item: TFileItem);
Var  p: TItem;
Begin
     StopOpen ('Scanning version information');
     StopInfo ('Version info');
     try
     
        p := item.sub.first;
        while (p <> nil) and not StopFlag do begin
           if not p.folder then
              (p as TFileItem).GetVer;
           p := p.next;
           StopTest;
        end;

        if not StopFlag then begin
           VerForm.Init;
           p := item.sub.first;
           while p <> nil do begin
              if not p.folder then
                 VerForm.Add (p as TFileItem);
              p := p.next;
           end;
           VerForm.Run;
        end;

     finally
        StopClose;
     end;
End;

(*************************** SORT BY FILE NAME ****************************)

Function CompareFileName (Item1, Item2: Pointer): Integer;
Var  a, b: TFileItem;
Begin
     a := TFileItem (Item1);
     b := TFileItem (Item2);
     result := CompByName (a, b);
End;

Procedure TVerForm.MenuSortByFileNameClick (Sender: TObject);
Begin
     MenuSortByFileName.Checked := true;
     list.Sort (CompareFileName);
     Redraw;
End;

(**************************** SORT BY PRODUCT *****************************)

Function CompareProduct (Item1, Item2: Pointer): Integer;
Const max = 9;
Var  a, b: TVerInfo;
     atr: TVerAttr;
     ans, i: integer;
Begin
     a := TFileItem (Item1).ver;
     b := TFileItem (Item2).ver;

     ans := 0;
     i := 1;
     while (ans=0) and (i<max) do begin
        atr := tab [i];
        ans := CompareText (a.value[atr], b.value[atr]);
        inc (i);
     end;
     result := ans;
End;

Procedure TVerForm.MenuSortByProductClick (Sender: TObject);
Begin
     MenuSortByProduct.Checked := true;
     list.Sort (CompareProduct);
     Redraw;
End;

(********************************* CLOSE **********************************)

Procedure TVerForm.FormKeyDown (Sender: TObject;
                                var Key: Word;
                                Shift: TShiftState);
Begin
     if (key = VK_ESCAPE) and (Shift = []) then Close;
End;

Procedure TVerForm.MenuCloseClick (Sender: TObject);
Begin
     Close;
End;

END.
