UNIT VerPro;

INTERFACE

Uses Windows, SysUtils;

Type TVerAttr = ( { fixed info }
                  fiFileVersion,
                  fiProductVersion,
                  fiDebug,
                  fiInfoInferred,
                  fiPatched,
                  fiPrerelease,
                  fiPrivateBuild,
                  fiSpecialBuild,
                  fiOperatingSystem,
                  fiFileType,
                  fiFileSubtype,
                  fiFileDate,

                  { string info }
                  siFileDescription,
                  siFileVersion,
                  siProductName,
                  siProductVersion,
                  siCompanyName,
                  siPrivateBuild,
                  siSpecialBuild,
                  siOriginalFilename,
                  siInternalName,
                  siComments,
                  siLegalCopyright,
                  siLegalTrademarks);

Const AttrName: array [TVerAttr] of string =
      ('File Version (fix)',
       'Product Version (fix)',
       'Debug (fix)',
       'Info Inferred (fix)',
       'Patched (fix)',
       'Prerelease (fix)',
       'Private Build (fix)',
       'Special Build (fix)',
       'Operating System (fix)',
       'File Type (fix)',
       'File Subtype (fix)',
       'File Date (fix)',

       'File Description (str)',
       'File Version (str)',
       'Product Name (str)',
       'Product Version (str)',
       'Company Name (str)',
       'Private Build (str)',
       'Special Build (str)',
       'Original Filename (str)',
       'Internal Name (str)',
       'Comments (str)',
       'Legal Copyright (str)',
       'Legal Trademarks (str)');

Type TVerInfo = class
     public
        value: array [TVerAttr] of string;

     private
        procedure PutStr (atr: TVerAttr; const txt: string);
        procedure PutVer (atr: TVerAttr; MS, LS: longword);
        procedure PutHex (atr: TVerAttr; MS, LS: longword);
        procedure PutFlag (flags: longword; atr: TVerAttr; opt: longword);

        procedure PutTab (atr: TVerAttr;
                          val: longword;
                          const key: array of longword;
                          const msg: array of string);

        procedure PutFixedInfo (block: pointer);
        procedure PutItem (block: pointer; path, key: string; atr: TVerAttr);
        procedure PutStringInfo (block: pointer);
     end;

Function ScanVerInfo (path: string): TVerInfo;

IMPLEMENTATION

(*************************** SCAN VERSION INFO ****************************)

Const os_mask = $FFFF0000; { high word }
      os_max = 5;
      os_key: array [1..os_max] of longword =
         ( VOS_UNKNOWN,
           VOS_DOS,
           VOS_OS216,
           VOS_OS232,
           VOS_NT );

      os_msg: array [1..os_max] of string =
         ( 'Unknown',
           'DOS',
           '16-bit OS/2',
           '32-bit OS/2',
           'Windows NT' );

Const gui_mask = $0000FFFF; { low word }
      gui_max = 5;
      gui_key: array [1..gui_max] of longword =
         ( VOS__BASE,
           VOS__WINDOWS16,
           VOS__PM16,
           VOS__PM32,
           VOS__WINDOWS32 );

      gui_msg: array [1..gui_max] of string =
         ( '', // prazdny retezec
           '16-bit Windows',
           '16-bit Presentation Manager',
           '32-bit Presentation Manager',
           '32-bit Windows' );

Const type_max = 7;
      type_key: array [1..type_max] of longword =
         ( VFT_UNKNOWN,
           VFT_APP,
           VFT_DLL,
           VFT_DRV,
           VFT_FONT,
           VFT_VXD,
           VFT_STATIC_LIB );

      type_msg: array [1..type_max] of string =
         ( 'Unknown',
           'Application',
           'Dynamic link library',
           'Device driver',
           'Font',
           'Virtual device',
           'Static library');

Const drv_max = 11;
      drv_key: array [1..drv_max] of longword =
         ( VFT2_UNKNOWN,
           VFT2_DRV_PRINTER,
           VFT2_DRV_KEYBOARD,
           VFT2_DRV_LANGUAGE,
           VFT2_DRV_DISPLAY,
           VFT2_DRV_MOUSE,
           VFT2_DRV_NETWORK,
           VFT2_DRV_SYSTEM,
           VFT2_DRV_INSTALLABLE,
           VFT2_DRV_SOUND,
           VFT2_DRV_COMM);

      drv_msg: array [1..drv_max] of string =
         ( 'Unknown',
           'Printer driver',
           'Keyboard driver',
           'Language driver',
           'Display driver',
           'Mouse driver',
           'Network driver',
           'System driver',
           'Installable driver',
           'Sound driver',
           'Communication driver');

Const font_max = 4;
      font_key: array [1..font_max] of longword =
         ( VFT2_UNKNOWN,
           VFT2_FONT_RASTER,
           VFT2_FONT_VECTOR,
           VFT2_FONT_TRUETYPE );

      font_msg: array [1..font_max] of string =
         ( 'Unknown',
           'Raster font',
           'Vector font',
           'TrueType font' );

{--------------------------------------------------------------------------}

Procedure TVerInfo.PutStr (atr: TVerAttr; const txt: string);
Begin
     value [atr] := txt;
End;

Procedure TVerInfo.PutVer (atr: TVerAttr; MS, LS: longword);
Begin
     if (MS<>0) or (LS<>0) then
        PutStr (atr, IntToStr (LongRec (MS).Hi) + '.' +
                     IntToStr (LongRec (MS).Lo) + '.' +
                     IntToStr (LongRec (LS).Hi) + '.' +
                     IntToStr (LongRec (LS).Lo) );
End;

Procedure TVerInfo.PutHex (atr: TVerAttr; MS, LS: longword);
Begin
     if (MS<>0) or (LS<>0) then
        PutStr (atr, IntToHex (MS, 8) + IntToHex (LS,8));
End;

Procedure TVerInfo.PutFlag (flags: longword; atr: TVerAttr; opt: longword);
Begin
     if flags and opt <> 0 then
        PutStr (atr, 'Yes');
End;

Function CnvTab (val: longword;
                 const key: array of longword;
                 const msg: array of string ): string;
Var  i: integer;
Begin
     i := low (key);
     while (i <= high (key)) and (key[i] <> val) do inc (i);

     if i <= high (msg)
        then result := msg[i]
        else result := '??? ' + IntToHex (val, 0) + ' ???';
End;

Procedure TVerInfo.PutTab (atr: TVerAttr;
                           val: longword;
                           const key: array of longword;
                           const msg: array of string);
Begin
     PutStr (atr, CnvTab (val, key, msg));
End;

Procedure TVerInfo.PutFixedInfo (block: pointer);
Var ver: PVSFixedFileInfo;
    len: longword;
    flags: longword;
    os, gui: string;
Begin
     if VerQueryValue (block, '\', pointer (ver), len) then
        with ver^ do begin

           PutVer (fiFileVersion, dwFileVersionMS, dwFileVersionLS);

           PutVer (fiProductVersion, dwProductVersionMS, dwProductVersionLS);

           flags := dwFileFlags;
           PutFlag (flags, fiDebug,        VS_FF_DEBUG);
           PutFlag (flags, fiInfoInferred, VS_FF_INFOINFERRED);
           PutFlag (flags, fiPatched,      VS_FF_PATCHED);
           PutFlag (flags, fiPrerelease,   VS_FF_PRERELEASE);
           PutFlag (flags, fiPrivateBuild, VS_FF_PRIVATEBUILD);
           PutFlag (flags, fiSpecialBuild, VS_FF_SPECIALBUILD);

           { OS }
           os := CnvTab (dwFileOS and os_mask, os_key, os_msg);
           gui := CnvTab (dwFileOS and gui_mask, gui_key, gui_msg);
           if gui <> '' then os := gui + ' running on ' + os;
           PutStr (fiOperatingSystem, os);

           PutTab (fiFileType, dwFileType, type_key, type_msg);

           if dwFileType=VFT_DRV then
              PutTab (fiFileSubtype,
                      dwFileSubtype, drv_key, drv_msg);

           if dwFileType=VFT_FONT then
              PutTab (fiFileSubtype,
                      dwFileSubtype, font_key, font_msg);

           PutHex (fiFileDate, dwFileDateMS, dwFileDateLS);
     end;
End;

{--------------------------------------------------------------------------}

Procedure TVerInfo.PutItem (block: pointer; path, key: string; atr:TVerAttr);
Var  buf: pointer;
     len: longword;
Begin
     if VerQueryValue (block, PChar (path+key), buf, len) then
        PutStr (atr, PChar (buf));
End;

Procedure TVerInfo.PutStringInfo (block: pointer);
Type TabArr = array [0..3] of byte;
     TabPtr = ^ TabArr;
Var  buf: pointer;
     len: longword;
     tab: TabPtr;
     path: string;
Begin
     if VerQueryValue (block, '\VarFileInfo\Translation', buf, len) then begin
        tab := TabPtr (buf);

        path := '\StringFileInfo\' +
                IntToHex (tab [1], 2) +
                IntToHex (tab [0], 2) +
                IntToHex (tab [3], 2) +
                IntToHex (tab [2], 2) +
                '\';

        PutItem (block, path, 'FileDescription',  siFileDescription);
        PutItem (block, path, 'FileVersion',      siFileVersion);
        PutItem (block, path, 'ProductName',      siProductName);
        PutItem (block, path, 'ProductVersion',   siProductVersion);
        PutItem (block, path, 'CompanyName',      siCompanyName);

        PutItem (block, path, 'PrivateBuild',     siPrivateBuild);
        PutItem (block, path, 'SpecialBuild',     siSpecialBuild);
        PutItem (block, path, 'OriginalFilename', siOriginalFilename);
        PutItem (block, path, 'InternalName',     siInternalName);

        PutItem (block, path, 'Comments',         siComments);
        PutItem (block, path, 'LegalCopyright',   siLegalCopyright);
        PutItem (block, path, 'LegalTrademarks',  siLegalTrademarks);
     end;
end;

{--------------------------------------------------------------------------}

Function ScanVerInfo (path: string): TVerInfo;
Var  info: TVerInfo;
     block: pointer;
     size, handle: longword;
Begin
     info := nil;

     size := GetFileVersionInfoSize (PChar (path), handle);
     if size>0 then begin

        block := AllocMem (size);
        if block<>nil then begin

           if GetFileVersionInfo (PChar (path), handle, size, block) then begin
              info := TVerInfo.Create;
              info.PutStringInfo (block);
              info.PutFixedInfo (block);
           end;

           FreeMem (block);
        end;
     end;

     result := info;
End;

END.

