UNIT ViewPro;

INTERFACE

Uses Windows,  {TPoint}
     SysUtils, {Exception}
     Classes,  {TShiftState}
     Controls, {TImageList}
     ComCtrls, {TTreeView}
     Menus,    {TMenuItem}
     ItemPro,  {TItem}
     SortPro,  {SortTable}
     ColDlg;   {ColumnsDialog}

{--------------------------------------------------------------------------}

Type TView = class
     public
        TreeView: TTreeView;
        ListView: TListView;
        StateImages: TImageList;

        MenuViewStateImages: TMenuItem;
        MenuViewColumns: TMenuItem;
        MenuViewSummary: TMenuItem;

        MenuSortByName: TMenuItem;
        MenuSortByExt: TMenuItem;
        MenuSortBySize: TMenuItem;
        MenuSortByTime: TMenuItem;
        MenuSortByValue: TMenuItem;
        MenuSortByColor: TMenuItem;
        MenuSortUnsorted: TMenuItem;

        MenuBlendNormal: TMenuItem;
        MenuBlendNew: TMenuItem;
        MenuBlendOld: TMenuItem;
        MenuBlendModified: TMenuItem;
        MenuBlendEqual: TMenuItem;
        MenuBlendChecked: TMenuItem;
        MenuBlendPartial: TMenuItem;
        MenuBlendUnchecked: TMenuItem;
        MenuBlendDirectories: TMenuItem;
        MenuBlendFiles: TMenuItem;

        ToolGoUp: TToolButton;
        ToolStateImages: TToolButton;
        ToolSortByColor: TToolButton;

        ToolBlendNormal: TToolButton;
        ToolBlendNew: TToolButton;
        ToolBlendOld: TToolButton;
        ToolBlendModified: TToolButton;
        ToolBlendEqual: TToolButton;
        ToolBlendChecked: TToolButton;
        ToolBlendPartial: TToolButton;
        ToolBlendUnchecked: TToolButton;
        ToolBlendDirectories: TToolButton;
        ToolBlendFiles: TToolButton;

     public
        root:   TItem;             { koren stromu }
        main:   TItem;             { prvek zobrazeny v seznamu }
        header: TBrwHeader;        { zahlavi seznamu }

        modes:      TModeSet;      { typy zobrazovanych polozek }
        states:     TStateSet;
        categories: TCategorySet;

        state_images: boolean;     { zobrazovat checkboxy }

        OnSelect: procedure (item: TItem) of object;

     private
        summary: boolean;          { zobrazovat soucty }

        sort:    TSort;            { zpusob trideni }
        SortSet: TSortSet;         { povolene zbusoby trideni }

        ShowTab: TableArr;         { tabulka zobrazovanych prvku }
        ShowCnt: integer;          { pocet zobrazovanych prvku }

        cleaning: boolean;         { true => prave se maze tree-view }

     private

        procedure CalcBranch (item: TItem; var rec: TCalcRec);
        procedure CalcSummary;         

        { show list view }
        procedure ChangeHeader;
        procedure SetupNode (node: TListItem; item: TItem);
        procedure UnlinkList;
        function  Filter (item: TItem): boolean;

        procedure ListViewSelectItem (sender: TObject; node: TListItem; selected: Boolean);

        { list view data }
        procedure ListViewData (sender: TObject; node: TListItem);

        procedure ListViewDataFind
          (Sender: TObject; Find: TItemFind; const FindString: String;
           const FindPosition: TPoint; FindData: Pointer; StartIndex: Integer;
           Direction: TSearchDirection; Wrap: Boolean; var Index: Integer);

        procedure ListViewDataHint
          (Sender: TObject; StartIndex, EndIndex: Integer);

        procedure ListViewDataStateChange
          (Sender: TObject; StartIndex, EndIndex: Integer;
           OldState, NewState: TItemStates);

        { show tree view }
        procedure AddToTree (item: TItem);
        procedure ShowSubBranch (item: TItem);
        procedure ShowBranch (item: TItem);

        procedure TreeViewExpanding
          (Sender: TObject; Node: TTreeNode; var AllowExpansion: Boolean);
        procedure TreeViewChange
          (Sender: TObject; Node: TTreeNode);

        procedure UpdateSubBranch (item: TItem);
        procedure UnlinkBranch (item: TItem);
        procedure SetSelection (item: TItem);

        { keybord commands }
        procedure GoUpClick (Sender: TObject);
        procedure ListViewDblClick (Sender: TObject);
        procedure ListViewKeyDown (Sender: TObject; var Key: Word; Shift: TShiftState);
        procedure TreeViewKeyDown (Sender: TObject; var Key: Word; Shift: TShiftState);

        { state }
        procedure SetItemState (item: TItem; val: TState);
        procedure SetBranchState (item: TItem; val: TState);
        procedure CheckState (item: TItem; go_up: boolean);
        procedure SetState (item: TItem; val: TState);
        procedure SelectTogleCmd;

        { view }
        procedure MenuViewStateImagesClick (Sender: TObject);
        procedure MenuViewColumnsClick (Sender: TObject);
        procedure MenuViewSummaryClick (Sender: TObject);

        { sort }
        procedure UpdateSortItem (MenuItem: TMenuItem; s: TSort);

        procedure SortByName    (Sender: TObject);
        procedure SortByExt     (Sender: TObject);
        procedure SortBySize    (Sender: TObject);
        procedure SortByTime    (Sender: TObject);
        procedure SortByValue   (Sender: TObject);
        procedure SortByColor   (Sender: TObject);
        procedure SortUnsorted  (Sender: TObject);

        procedure ListViewColumnClick (Sender: TObject; Column: TListColumn);

        { blend }
        procedure UpdateModes;
        procedure SwitchMode (v: TMode);

        procedure UpdateStates;
        procedure SwitchState (v: TState);

        procedure UpdateCategories;
        procedure SwitchCategory (v: TCategory);

        procedure BlendNormal      (Sender: TObject);
        procedure BlendNew         (Sender: TObject);
        procedure BlendOld         (Sender: TObject);
        procedure BlendModified    (Sender: TObject);
        procedure BlendEqual       (Sender: TObject);
        procedure BlendChecked     (Sender: TObject);
        Procedure BlendPartial     (Sender: TObject);
        procedure BlendUnchecked   (Sender: TObject);
        procedure BlendDirectories (Sender: TObject);
        procedure BlendFiles       (Sender: TObject);

        { toolbar }
        procedure ModeMouseUp
          (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
        procedure StateMouseUp
          (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
        procedure CategoryMouseUp
          (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

     public
        procedure Setup;

        { show list view }
        procedure StoreColumnsWidth;
        procedure SetupHeader;
        procedure ChangeList (start: TItem);

        { show tree view }
        procedure UpdateBranch (item: TItem);
        procedure AddBranch (item: TItem);
        procedure DeleteBranch (item: TItem);
        procedure ChangeTree (item: TItem);

        { current item }
        function GetListItem (node: TListItem): TItem;
        function GetTreeItem (node: TTreeNode): TItem;
        function CurrentListItem: TItem;
        function CurrentTreeItem: TItem;
        function CurrentItem: TItem;

        { select items }
        procedure SelectCmd;
        procedure UnselectCmd;
        procedure SelGroup (item: TItem; sel: TState; mask: string; sub: boolean);

        { view, sort, blend menu }
        procedure SetStateImages (b: boolean);
        procedure UpdateSortMenu;
        procedure UpdateBlendMenu;

        procedure SetSort (s: TSort);
        procedure SetModes (s: TModeSet);
        procedure SetStates (s: TStateSet);
        procedure SetCategories (s: TCategorySet);
     end;

IMPLEMENTATION

(********************************* ERROR **********************************)

Procedure Error (s: string);
Begin
     raise Exception.Create (s);
End;

(******************************** SUMMARY *********************************)

Procedure TView.CalcBranch (item: TItem;
                            var rec: TCalcRec);
Var  p: TItem;
Begin
     if (item.mode in modes) and (item.state in states) then begin
        item.GetSum (rec);
        if item.folder then inc (rec.all_folders)
                       else inc (rec.all_nodes);
     end;

     p := item.sub.first;
     while p <> nil do begin
        CalcBranch (p, rec);
        p := p.next;
     end;
End;

Procedure TView.CalcSummary;
Var  p: TItem;
     rec: TCalcRec;
Begin
     p := main.sub.first;
     while p <> nil do begin
        FillChar (rec, sizeof (rec), 0);
        CalcBranch (p, rec);
        rec.all_bytes := rec.fil_bytes + rec.reg_bytes;
        rec.all_used  := rec.fil_used + rec.reg_used;

        p.sum_leafs := rec.all_nodes;
        p.sum_folders := rec.all_folders;
        p.sum_bytes := rec.all_bytes;
        p.sum_used := rec.all_used;

        p := p.next;
     end;
End;

(*************************** LIST VIEW - HEADER ***************************)

Procedure TView.StoreColumnsWidth;
Var  i, k: integer;
Begin
     if header <> nil then begin
        k := 0;
        for i := 1 to header.cnt do
           if header.tab[i].c_enabled then begin
              header.tab[i].c_width := ListView.Column[k].Width;
              inc (k);
           end;
     end;
End;

Procedure TView.SetupHeader;
Var  i:  integer;
     cc: TListColumn;
Begin
     ListView.Items.BeginUpdate;
     ListView.Columns.BeginUpdate;
     ListView.Columns.Clear;

     if header <> nil then
        for i := 1 to header.cnt do
           with header.tab[i] do
              if c_enabled then begin
                 cc := ListView.Columns.Add;
                 cc.Caption := c_title;
                 cc.Width := c_width;
                 if (c_sort = BySize) or (c_sort >= ByLeafs) then
                    cc.Alignment := taRightJustify;
              end;

     ListView.Columns.EndUpdate;
     ListView.Items.EndUpdate;
End;

Procedure TView.ChangeHeader;
Var  hdr: TBrwHeader;
Begin
     hdr := nil;
     if main <> nil then
        if summary then hdr := main.GetSumHeader
                   else hdr := main.GetHeader;

     if hdr <> header then begin
        StoreColumnsWidth;
        header := hdr;
        SetupHeader;
     end;
End;

(**************************** LIST VIEW - SHOW ****************************)

Procedure TView.SetupNode (node: TListItem; item: TItem);
Var  i: integer;
     txt: TColumns;
Begin
     item.lst := node.Index;

     node.Caption := item.name;
     node.Data := item;
     node.ImageIndex := item.GetImageIndex (false);
     node.StateIndex := item.GetStateIndex;

     for i := 1 to header.cnt do txt [i] := '';
     if summary then item.ShowSumItem (txt)
                else item.ShowItem (txt);

     { txt[1] se nepouziva }
     for i := 2 to header.cnt do
        if header.tab[i].c_enabled then
           node.SubItems.Add (txt [i]);
End;

Procedure TView.UnlinkList;
Var  item: TItem;
Begin
     if main <> nil then begin
        item := main.sub.first;
        while item <> nil do begin
           item.lst := -1;
           item := item.next;
        end;
     end;
End;

Function TView.Filter (item: TItem): boolean;
Var  category: TCategory;
Begin
     if item.folder then category := dir_category
                    else category := file_category;

     result := (item.mode in modes) and
               (item.state in states) and
               (category in categories);
End;

Procedure TView.ChangeList (start: TItem);
Begin
     if start=nil then Error ('Empty list');

     { clear items }
     UnlinkList;
     ListView.Items.BeginUpdate;
     ListView.Items.count := 0;

     { set main }
     main := start;

     { change header }
     ChangeHeader;

     { summary }
     if summary then CalcSummary;

     { allocate table }
     FreeTable (ShowTab);
     AllocTable (start, Filter, ShowTab);
     ShowCnt := length (ShowTab);

     { check sort }
     if summary then
        SortSet := [ByName, ByLeafs, ByFolders, ByBytes, ByUsed]
     else begin
        SortSet := [ByName, ByColor, Unsorted];
        if main <> nil then SortSet := SortSet + main.GetSortSet;
     end;
     if not (sort in SortSet) then sort := ByName;

     { sort table }
     if main <> root then
        SortTable (SortFunc (sort), ShowTab);

     { fill ListView }
     ListView.Items.Count := ShowCnt;
     ListView.Items.EndUpdate;

     (*
     tmp := ListView.Items[0];
     if tmp <> nil then begin
        tmp.Selected := true;
        tmp.Focused := true;
     end;
     *)

     { notify detail view }
     if Assigned (OnSelect) then
        OnSelect (nil);
End;

Procedure TView.ListViewSelectItem (sender: TObject; node: TListItem; selected: Boolean);
Begin
     if node <> nil then
        if Assigned (OnSelect) then
           OnSelect (GetListItem (node));
End;

(**************************** LIST VIEW - DATA ****************************)

Procedure TView.ListViewData (sender: TObject; node: TListItem);
Var  inx: integer;
Begin
     inx := node.Index;
     if inx < ShowCnt then
        SetupNode (node, ShowTab [inx]);
End;

Procedure TView.ListViewDataFind
  (Sender: TObject;
   Find: TItemFind;
   const FindString: String;
   const FindPosition: TPoint;
   FindData: Pointer;
   StartIndex: Integer;
   Direction: TSearchDirection;
   Wrap: Boolean;
   var Index: Integer);
Begin
     index := -1; {!?}
End;

Procedure TView.ListViewDataHint
  (Sender: TObject; StartIndex, EndIndex: Integer);
Begin
     { nothing }
End;

Procedure TView.ListViewDataStateChange
  (Sender: TObject; StartIndex, EndIndex: Integer; OldState, NewState: TItemStates);
Begin
     { nothing }
End;

(**************************** TREE VIEW - SHOW ****************************)

Procedure TView.AddToTree (item: TItem);
Var  start, node: TTreeNode;
Begin
     if not item.folder then Error ('Folder required');
     if item.ref <> nil then Error ('Already in tree');

     if item.up = nil then start := nil
                      else start := item.up.ref;

     node := TreeView.Items.AddChildObject (start, item.name, item);
     item.ref := node;

     node.ImageIndex    := item.GetImageIndex (false);
     node.SelectedIndex := item.GetImageIndex (true);
     node.StateIndex    := item.GetStateIndex;
End;

Procedure TView.ShowSubBranch (item: TItem);
{ Zobraz zadany prvek a jednu vrstvu pod nim }
Var  p: TItem;
Begin
     if item.ref=nil then AddToTree (item);

     { nalezni podadresar }
     p := item.sub.first;
     while (p <> nil) and not p.folder do
        p := p.next;

     if (p<>nil) and p.folder and (p.ref = nil) then AddToTree (p);
End;

Procedure TView.ShowBranch (item: TItem);
{ Zobraz zadany prvek a dve vrstvy pod nim }
Var  p: TItem;
Begin
     if item.ref=nil then AddToTree (item);

     p := item.sub.first;
     while p <> nil do begin
        if p.folder then ShowSubBranch (p);
        p := p.next;
     end;
End;

Procedure TView.TreeViewExpanding
  (Sender: TObject; Node: TTreeNode; var AllowExpansion: Boolean);
Var  item: TItem;
Begin
     item := GetTreeItem (Node);
     ShowBranch (item);
End;

Procedure TView.TreeViewChange (Sender: TObject; Node: TTreeNode);
Begin
     if not cleaning then
        ChangeList (GetTreeItem (Node));
End;

{--------------------------------------------------------------------------}

Procedure TView.UpdateSubBranch (item: TItem);
{ Aktualizuj zobrazeni prvku }
Var  p: TItem;
     node: TTreeNode;
Begin
     node := item.ref;
     if node <> nil then begin

        node.Text          := item.name;
        node.ImageIndex    := item.GetImageIndex (false);
        node.SelectedIndex := item.GetImageIndex (true);
        node.StateIndex    := item.GetStateIndex;

        p := item.sub.first;
        while p <> nil do begin
           if p.folder then UpdateSubBranch (p);
           p := p.next;
        end;
     end;
End;

Procedure TView.UpdateBranch (item: TItem);
Begin
     TreeView.Items.BeginUpdate;
     UpdateSubBranch (item);
     TreeView.Items.EndUpdate;
End;

{--------------------------------------------------------------------------}

Procedure TView.UnlinkBranch (item: TItem);
{ Zrus propojeni mezi TItem a TTreeNode }
Var  p: TItem;
Begin
     if item <> nil then begin
        if item.ref <> nil then begin
           item.ref.Data := nil;
           item.ref := nil;
        end;

        p := item.sub.first;
        while p <> nil do begin
           if p.folder then UnlinkBranch (p);
           p := p.next;
        end;
     end;
End;

{--------------------------------------------------------------------------}

Procedure TView.SetSelection (item: TItem);
{ Vyber aktivniho prvku ve stromu }
Begin
     TreeView.Selected := item.ref;
End;

Procedure TView.AddBranch (item: TItem);
Begin
     if item <> nil then begin

        SortFolder (item, SortFunc (ByName), true);
        root.Add (item);

        TreeView.Items.BeginUpdate;
        ShowBranch (item);
        SetSelection (item);
        TreeView.Items.EndUpdate;
     end;
End;

Procedure TView.DeleteBranch (item: TItem);
Var  tmp:  TTreeNode;
Begin
     tmp := item.ref;
     UnlinkBranch (item);
     if tmp <> nil then TreeView.Items.Delete (tmp);
     item.Unlink;
     item.Free;

     { problemy s prekreslovanim - nemenit }
     SetSelection (root);
     ChangeList (root);
End;

Procedure TView.ChangeTree (item: TItem);
Begin
     if item = nil then Error ('Empty tree');

     TreeView.Items.BeginUpdate;

     { clear tree }
     UnlinkBranch (root);
     cleaning := true;
     TreeView.Items.Clear;
     cleaning := false;

     { change root }
     root.Free;
     root := item;
     SortFolder (root, SortFunc (ByName), true);

     { paint tree }
     ShowBranch (root);
     SetSelection (root);
     root.ref.Expand (false);

     TreeView.Items.EndUpdate;
End;

(****************************** CURRENT ITEM ******************************)

Function TView.GetListItem (node: TListItem): TItem;
Begin
     if node = nil then
        result := nil
     else begin
        result := ShowTab [node.Index];
        if result = nil then Error ('Invalid list item');
     end;
End;

Function TView.GetTreeItem (node: TTreeNode): TItem;
Begin
     if node = nil then
        result := nil
     else begin
        result := TObject (node.Data) as TItem;
        if (result = nil) or (result.ref <> node) then
           Error ('Invalid tree item');
     end;
End;

Function TView.CurrentListItem: TItem;
Begin
     result := GetListItem (ListView.Selected);
End;

Function TView.CurrentTreeItem: TItem;
Begin
     result := GetTreeItem (TreeView.Selected);
End;

Function TView.CurrentItem: TItem;
Begin
     if TreeView.Focused then result := CurrentTreeItem
     else if ListView.Focused then result := CurrentListItem
     else result := nil;
End;

(*************************** KEYBOARD COMMANDS ****************************)

Procedure TView.GoUpClick (Sender: TObject);
Var  item: TItem;
Begin
     item := CurrentTreeItem;
     if item <> nil then item := item.up;
     if item <> nil then SetSelection (item);
End;

Procedure TView.ListViewDblClick (Sender: TObject);
Var  item: TItem;
Begin
     item := CurrentListItem;
     if item <> nil then
        if item.folder
           then SetSelection (item)
End;

Procedure TView.ListViewKeyDown (Sender: TObject;
                                 var Key: Word;
                                 Shift: TShiftState);
Begin
     if Shift = [] then begin
             if Key = VK_RETURN then ListViewDblClick (nil)
        else if Key = VK_ESCAPE then GoUpClick (nil)
        else if Key = VK_INSERT then SelectTogleCmd
     end;
End;

Procedure TView.TreeViewKeyDown (Sender: TObject;
                                 var Key: Word;
                                 Shift: TShiftState);
Begin
     if Shift = [] then begin
             if Key = VK_RETURN then key := VK_RIGHT
        else if Key = VK_ESCAPE then GoUpClick (nil)
        else if Key = VK_INSERT then SelectTogleCmd
     end;
End;

(********************************* SELECT *********************************)

Procedure TView.SetItemState (item: TItem; val: TState);
Var  inx: integer;
Begin
     { set value }
     item.state := val;
     inx := item.GetStateIndex;

     { update tree view }
     if item.ref <> nil then
        item.ref.StateIndex := inx;

     { update list view }
     if item.lst >= 0 then
        ListView.UpdateItems (item.lst, item.lst);
End;

Procedure TView.SetBranchState (item: TItem; val: TState);
Var  p: TItem;
Begin
     if item = nil then RunError (255);
     if item.state <> val then begin
        SetItemState (item, val);

        p := item.sub.first;
        while p <> nil do begin
           SetBranchState (p, val);
           p := p.next;
        end;
     end;
End;

Procedure TView.CheckState (item: TItem; go_up: boolean);
{ Uprav nadrazene prvky }
Var  p: TItem;
     eq: boolean;
     ans: TState;
Begin
     if item <> nil then begin

        p := item.sub.first;
        if p <> nil then begin
           { item ma alespon jeden vnoreny prvek }

           { prvni prvek }
           ans := p.state;
           eq := true;
           p := p.next;

           { dalsi prvky }
           while (p <> nil) and eq do begin
              if p.state <> ans then eq := false;
              p := p.next;
           end;

           { vysledek }
           if not eq then ans := partial;
           if item.state <> ans then begin
              SetItemState (item, ans); { oprav stav }
              if go_up then CheckState (item.up, true); { pokracuj nahoru }
           end;

        end;
     end;
End;

Procedure TView.SetState (item: TItem; val: TState);
Begin
     if (item <> nil) and (item.state <> val) then begin
        SetBranchState (item, val);
        CheckState (item.up, true);
     end;
End;

{--------------------------------------------------------------------------}

Procedure TView.SelectTogleCmd;
Var  item: TItem;
     tmp: TListItem;
Begin
     item := CurrentItem;
     if item <> nil then begin
        if item.state = checked
           then SetState (item, unchecked)
           else SetState (item, checked);

        { posun na dalsi polozku }
        if item.lst >= 0 then begin
           tmp := ListView.Items [item.lst+1];

           if tmp <> nil then begin
              tmp.MakeVisible (false);
              tmp.Selected := true;
              tmp.Focused := true;
           end;
        end;

     end;
End;

Procedure TView.SelectCmd;
Begin
     SetState (CurrentItem, checked);
End;

Procedure TView.UnselectCmd;
Begin
     SetState (CurrentItem, unchecked);
End;

{--------------------------------------------------------------------------}

Function CompMask (name, mask: string; ni, mi: integer): boolean;
Var  nlen, mlen, cnt: integer;
     equal, asterisk: boolean;
     c: char;
Begin
     nlen := length (name);
     mlen := length (mask);

     { prvni cyklus:
       postupuje se dokud se znaky shoduji nebo dokud v masce jsou otazniky }

     equal := true;
     asterisk := false;

     while equal and (ni <= nlen) and (mi <= mlen) and not (asterisk) do begin
        case mask [mi] of
           '*': asterisk := true; { hvezdicka se nezpracovava timto cyklem }

           '?': begin
                   { postup o jeden znak }
                   inc (ni);
                   inc (mi);
                end;

           else if name [ni] = mask [mi] then begin
                   { postup o jeden znak }
                   inc (ni);
                   inc (mi);
                end
                else begin
                   { zastav porovnavani }
                   equal := false;
                end;
        end;
     end;

     { po ukonceni cyklu muze nastat:
       A) asterisk=true => porovnavani se zastavilo na hvezdicce
       B) asterisk=false =>
          a) equal=false => name a mask se neshoduji
          a) equal=true  => name [1..ni-1] a mask [1..mi-1] se shoduji
                            musi se zkontrolovat zbytky
                            v masce mohou zustat i hvezdicky }

     if asterisk then begin
        if (not equal) or
           (ni > nlen) or
           (mi > mlen) then Error ('Internal bug');

        { preskoc hvezdicky a otazniky }
        cnt := 0;
        while equal and (mi<=mlen) and (mask[mi] in ['?', '*']) do begin
           if mask[mi] = '?' then inc (cnt);
           inc (mi);
        end;

        { v 'name' musi byt alespon tolik znaku jako otazniku }
        inc (ni, cnt);

        if ni > nlen+1 then
           equal := false { malo znaku }
        else if ni = nlen+1 then
           equal := (mi > mlen) { vyhovuje pouze hvezdicka na konci masky }
        else begin
           if mi <= mlen then begin
              { hvezdicka neni na konci masky - porovnej zbytky }
              c := mask [mi];
              equal := false;
              while not equal and (ni <= nlen) do begin
                 if name[ni] = c then
                    equal := CompMask (name, mask, ni+1, mi+1); { zkus zbytek }
                 inc (ni);
              end;
           end;
        end;

        { vysledek je v promenne 'equal' - na 'ni' a 'mi' jiz nezalezi }
     end

     else begin { not asterisk }
        { preskoc zbyvajici hvezdicky }
        while (mi <= mlen) and (mask[mi] = '*') do inc (mi);

        { zkontroluj zbytky }
        if (ni <= nlen) or (mi <= mlen) then equal := false;
     end;

     result := equal;
End;

Function CompareMask (name, mask: string): boolean;
Var  i: integer;
Begin
     for i := 1 to length (name) do
        name [i] := UpCase (name [i]);

     for i := 1 to length (mask) do
        mask [i] := UpCase (mask [i]);

     result := CompMask (name, mask, 1, 1);
End;

Procedure TView.SelGroup (item: TItem; sel: TState; mask: string; sub: boolean);
Var  p: TItem;
Begin
     if item <> nil then begin
        p := item.sub.first;
        while p <> nil do begin
           if p.folder then begin
              if sub then begin
                 SelGroup (p, sel, mask, sub);
                 CheckState (p, false);
              end;
           end
           else begin
              if CompareMask (p.name, mask) then
                 SetItemState (p, sel);
           end;
           p := p.next;
        end;
        CheckState (item, true);
     end;
End;

(**************************** MENU and TOOLBAR ****************************)

Procedure CheckMenu (MenuItem: TMenuItem; val: boolean);
Begin
     if MenuItem <> nil then
        MenuItem.Checked := val;
End;

Procedure UpdateTool (ToolButton: TToolButton; val: boolean);
Begin
     if ToolButton <> nil then
        ToolButton.Down := val;
End;

(********************************** VIEW **********************************)

Procedure TView.SetStateImages (b: boolean);
Var  p: TImageList;
Begin
     state_images := b;

     if state_images then p := StateImages
                     else p := nil;

     TreeView.StateImages := p;
     ListView.StateImages := p;

     CheckMenu (MenuViewStateImages, state_images);
     UpdateTool (ToolStateImages, state_images);
End;

Procedure TView.MenuViewStateImagesClick (Sender: TObject);
Begin
     SetStateImages (not state_images);
End;

{--------------------------------------------------------------------------}

Procedure TView.MenuViewColumnsClick (Sender: TObject);
Begin
     StoreColumnsWidth; { uschovat sirku zahlavi drive nez se zmeni popis zahlavi }
     if ColumnsDialog.Execute (header) then begin
        SetupHeader; { zobraz nove zahlavi, pouze ChangeList nestaci }
        ChangeList (main); { zobrazit nove sloupce }
     end;
End;

{--------------------------------------------------------------------------}

Procedure TView.MenuViewSummaryClick (Sender: TObject);
Begin
     summary := not summary;
     MenuViewSummary.checked := summary;
     ChangeList (main); { zobrazit nova data }
End;

(********************************** SORT **********************************)

Procedure TView.UpdateSortItem (MenuItem: TMenuItem; s: TSort);
Begin
     if MenuItem <> nil then begin
        MenuItem.Enabled := s in SortSet;
        MenuItem.Checked := (s = sort);
     end;
End;

Procedure TView.UpdateSortMenu;
Begin
     UpdateSortItem (MenuSortByName,   ByName);
     UpdateSortItem (MenuSortByExt,    ByExt);
     UpdateSortItem (MenuSortBySize,   BySize);
     UpdateSortItem (MenuSortByTime,   ByTime);
     UpdateSortItem (MenuSortByValue,  ByValue);
     UpdateSortItem (MenuSortByColor,  ByColor);
     UpdateSortItem (MenuSortUnsorted, Unsorted);
End;

Procedure TView.SetSort (s: TSort);
Begin
     sort := s;
     ToolSortByColor.Down := (sort = ByColor);
     if main <> nil then ChangeList (main);
End;

{--------------------------------------------------------------------------}

Procedure TView.SortByName (Sender: TObject);
Begin
     SetSort (ByName);
End;

Procedure TView.SortByExt (Sender: TObject);
Begin
     SetSort (ByExt);
End;

Procedure TView.SortBySize (Sender: TObject);
Begin
     SetSort (BySize);
End;

Procedure TView.SortByTime (Sender: TObject);
Begin
     SetSort (ByTime);
End;

Procedure TView.SortByValue (Sender: TObject);
Begin
     SetSort (ByValue);
End;

Procedure TView.SortByColor (Sender: TObject);
Begin
     SetSort (ByColor);
End;

Procedure TView.SortUnsorted (Sender: TObject);
Begin
     SetSort (Unsorted);
End;

{--------------------------------------------------------------------------}

Procedure TView.ListViewColumnClick (Sender: TObject;
                                     Column: TListColumn);
Var  inx, i, n: integer;
     s: TSort;
Begin
     if header <> nil then begin
        inx := Column.Index + 1; { cislo hledaneho sloupce (od 1) }

        i := 0; { cislo sloupce v tabulce }
        n := 0; { pocet zobrazovanych sloupcu }

        while (i < header.cnt) and (n < inx) do begin
           inc (i);
           if header.tab[i].c_enabled then inc (n);
        end;

        if n = inx then begin
           s := header.tab[i].c_sort;
           { unsorted => sloupec nema definovan zpusob trideni }
           if s <> Unsorted then SetSort (s);
        end;
     end;
End;

(********************************* BLEND *********************************)

Procedure TView.UpdateBlendMenu;
Begin
     CheckMenu (MenuBlendNormal,   normal in modes);
     CheckMenu (MenuBlendNew,      new in modes);
     CheckMenu (MenuBlendOld,      old in modes);
     CheckMenu (MenuBlendModified, modified in modes);
     CheckMenu (MenuBlendEqual,    equal in modes);

     CheckMenu (MenuBlendChecked,   checked in states);
     CheckMenu (MenuBlendPartial,   partial in states);
     CheckMenu (MenuBlendUnchecked, unchecked in states);

     CheckMenu (MenuBlendDirectories, dir_category in categories);
     CheckMenu (MenuBlendFiles,       file_category in categories);
End;

{--------------------------------------------------------------------------}

Procedure TView.UpdateModes;
Begin
     UpdateTool (ToolBlendNormal, normal in modes);
     UpdateTool (ToolBlendNew, new in modes);
     UpdateTool (ToolBlendOld, old in modes);
     UpdateTool (ToolBlendModified, modified in modes);
     UpdateTool (ToolBlendEqual, equal in modes);
End;

Procedure TView.SetModes (s: TModeSet);
Begin
     modes := s;
     UpdateModes;
     if main <> nil then ChangeList (main);
End;

Procedure TView.SwitchMode (v: TMode);
Var  s: TModeSet;
Begin
     s := modes;
     if v in s then Exclude (s, v)
               else Include (s, v);
     SetModes (s);
End;

{--------------------------------------------------------------------------}

Procedure TView.UpdateStates;
Begin
     UpdateTool (ToolBlendChecked,   checked in states);
     UpdateTool (ToolBlendPartial,   partial in states);
     UpdateTool (ToolBlendUnchecked, unchecked in states);
End;

Procedure TView.SetStates (s: TStateSet);
Begin
     states := s;
     UpdateStates;
     if main <> nil then ChangeList (main);
End;

Procedure TView.SwitchState (v: TState);
Var  s: TStateSet;
Begin
     s := states;
     if v in s then Exclude (s, v)
               else Include (s, v);
     SetStates (s);
End;

{--------------------------------------------------------------------------}

Procedure TView.UpdateCategories;
Begin
     UpdateTool (ToolBlendDirectories, dir_category in categories);
     UpdateTool (ToolBlendFiles, file_category in categories);
End;

Procedure TView.SetCategories (s: TCategorySet);
Begin
     categories := s;
     UpdateCategories;
     if main <> nil then ChangeList (main);
End;

Procedure TView.SwitchCategory (v: TCategory);
Var  s: TCategorySet;
Begin
     s := categories;
     if v in s then Exclude (s, v)
               else Include (s, v);
     SetCategories (s);
End;

{--------------------------------------------------------------------------}

Procedure TView.BlendNormal (Sender: TObject);
Begin
     SwitchMode (normal);
End;

Procedure TView.BlendNew (Sender: TObject);
Begin
     SwitchMode (new);
End;

Procedure TView.BlendOld (Sender: TObject);
Begin
     SwitchMode (old);
End;

Procedure TView.BlendModified (Sender: TObject);
Begin
     SwitchMode (modified);
End;

Procedure TView.BlendEqual (Sender: TObject);
Begin
     SwitchMode (equal);
End;

Procedure TView.BlendChecked (Sender: TObject);
Begin
     SwitchState (checked);
End;

Procedure TView.BlendPartial (Sender: TObject);
Begin
     SwitchState (partial);
End;

Procedure TView.BlendUnchecked (Sender: TObject);
Begin
     SwitchState (unchecked);
End;

Procedure TView.BlendDirectories (Sender: TObject);
Begin
     SwitchCategory (dir_category);
End;

Procedure TView.BlendFiles (Sender: TObject);
Begin
     SwitchCategory (file_category);
End;

{--------------------------------------------------------------------------}

Procedure TView.ModeMouseUp
  (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
Const all = [normal, new, old, modified, equal];
Var   val: TMode;
Begin
     if Sender = ToolBlendNew then val := new
     else if Sender = ToolBlendOld then val := old
     else if Sender = ToolBlendModified then val := modified
     else if Sender = ToolBlendEqual then val := equal
     else val := normal;

     if Button = mbLeft then begin
        if Shift = [] then SwitchMode (val)
        else if Shift = [ssShift] then SetModes ([val])
        else if Shift = [ssCtrl] then SetModes (all-[val])
        else if Shift = [ssAlt] then SetModes (all)
        else UpdateModes;
     end;
End;

Procedure TView.StateMouseUp
  (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
Const all = [checked, partial, unchecked];
Var   val: TState;
Begin
     if Sender = ToolBlendChecked then val := checked
     else if Sender = ToolBlendPartial then val := partial
     else val := unchecked;

     if Button = mbLeft then begin
        if Shift = [] then SwitchState (val)
        else if Shift = [ssShift] then SetStates ([val])
        else if Shift = [ssCtrl] then SetStates (all-[val])
        else if Shift = [ssAlt] then SetStates (all)
        else UpdateStates;
     end;
End;

Procedure TView.CategoryMouseUp
  (Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
Const all = [dir_category, file_category];
Var  val: TCategory;
Begin
     if Sender = ToolBlendDirectories then val := dir_category
     else val := file_category;

     if Button = mbLeft then begin
        if Shift = [] then SwitchCategory (val)
        else if Shift = [ssShift] then SetCategories ([val])
        else if Shift = [ssCtrl] then SetCategories (all-[val])
        else if Shift = [ssAlt] then SetCategories (all)
        else UpdateCategories;
     end;
End;

(********************************** INIT **********************************)

Procedure SetupMenu (MenuItem: TMenuItem; MenuProc: TNotifyEvent);
Begin
     if MenuItem <> nil then
        MenuItem.OnClick := MenuProc;
End;

Procedure SetupSortMenu (MenuItem: TMenuItem; MenuProc: TNotifyEvent);
Begin
     if MenuItem <> nil then begin
        MenuItem.GroupIndex := 20; {!?}
        MenuItem.RadioItem := true;
        MenuItem.OnClick := MenuProc;
     end;
End;

Procedure SetupBlendMenu (MenuItem: TMenuItem; MenuProc: TNotifyEvent);
Begin
     if MenuItem <> nil then begin
        MenuItem.GroupIndex := 30; {!?}
        MenuItem.OnClick := MenuProc;
     end;
End;

Procedure SetupTool (ToolButton: TToolButton;
                     MenuProc: TNotifyEvent;
                     ButtonStyle: TToolButtonStyle = tbsButton);
Begin
     if ToolButton <> nil then begin
        ToolButton.Style := ButtonStyle;
        ToolButton.OnClick := MenuProc;
     end;
End;

Procedure SetupBlendTool (ToolButton: TToolButton; MouseProc: TMouseEvent);
Begin
     if ToolButton <> nil then begin
        ToolButton.AllowAllUp := true;
        ToolButton.Style := tbsCheck;
        ToolButton.OnMouseUp := MouseProc;
     end;
End;

Procedure TView.Setup;
Begin
     Assert (TreeView <> nil);
     Assert (ListView <> nil);

     summary := false;

     with TreeView do begin
        Indent := 19;
        ReadOnly := True;
        TabOrder := 2;
        OnChange := TreeViewChange;
        OnExpanding := TreeViewExpanding;
        OnKeyDown := TreeViewKeyDown;
     end;

     with ListView do begin
        AllocBy := 64;
        // Columns.Clear;
        OwnerData := true;
        ReadOnly := true;
        ViewStyle := vsReport;
        OnColumnClick := ListViewColumnClick;
        OnData := ListViewData;
        OnDataFind := ListViewDataFind;
        OnDataHint := ListViewDataHint;
        OnDataStateChange := ListViewDataStateChange;
        OnDblClick := ListViewDblClick;
        OnKeyDown := ListViewKeyDown;
        OnSelectItem := ListViewSelectItem;
     end;

     SetupMenu (MenuViewStateImages, MenuViewStateImagesClick);
     SetupMenu (MenuViewColumns, MenuViewColumnsClick);
     SetupMenu (MenuViewSummary, MenuViewSummaryClick);

     SetupSortMenu (MenuSortByName, SortByName);
     SetupSortMenu (MenuSortByExt, SortByExt);
     SetupSortMenu (MenuSortBySize, SortBySize);
     SetupSortMenu (MenuSortByTime, SortByTime);
     SetupSortMenu (MenuSortByValue, SortByValue);
     SetupSortMenu (MenuSortByColor, SortByColor);
     SetupSortMenu (MenuSortUnsorted, SortUnsorted);

     SetupBlendMenu (MenuBlendNormal, BlendNormal);
     SetupBlendMenu (MenuBlendNew, BlendNew);
     SetupBlendMenu (MenuBlendOld, BlendOld);
     SetupBlendMenu (MenuBlendModified, BlendModified);
     SetupBlendMenu (MenuBlendEqual, BlendEqual);
     SetupBlendMenu (MenuBlendChecked, BlendChecked);
     SetupBlendMenu (MenuBlendPartial, BlendPartial);
     SetupBlendMenu (MenuBlendUnchecked, BlendUnchecked);
     SetupBlendMenu (MenuBlendDirectories, BlendDirectories);
     SetupBlendMenu (MenuBlendFiles, BlendFiles);

     SetupTool (ToolGoUp, GoUpClick);
     SetupTool (ToolStateImages, MenuViewStateImagesClick, tbsCheck);
     SetupTool (ToolSortByColor, SortByColor, tbsCheck);

     SetupBlendTool (ToolBlendNormal, ModeMouseUp);
     SetupBlendTool (ToolBlendNew, ModeMouseUp);
     SetupBlendTool (ToolBlendOld, ModeMouseUp);
     SetupBlendTool (ToolBlendModified, ModeMouseUp);
     SetupBlendTool (ToolBlendEqual, ModeMouseUp);
     SetupBlendTool (ToolBlendChecked, StateMouseUp);
     SetupBlendTool (ToolBlendPartial, StateMouseUp);
     SetupBlendTool (ToolBlendUnchecked, StateMouseUp);
     SetupBlendTool (ToolBlendDirectories, CategoryMouseUp);
     SetupBlendTool (ToolBlendFiles, CategoryMouseUp);

     SetStateImages (false); 
     SetSort (ByName);

     SetModes ([normal, new, old, modified, equal]);
     SetStates ([checked, unchecked, partial]);
     SetCategories ([file_category, dir_category]);
End;

END.

